package org.jgroups.tests;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import junit.framework.Test;
import junit.framework.TestSuite;

import org.jgroups.Channel;
import org.jgroups.ChannelException;
import org.jgroups.Message;
import org.jgroups.ReceiverAdapter;
import org.jgroups.View;
import org.jgroups.util.Util;


/**
 * @author Bela Ban
 * @version $Id: JoinTest.java,v 1.9 2007/07/04 08:31:42 belaban Exp $
 */
public class JoinTest extends ChannelTestBase {
    Channel c1, c2;

    public JoinTest(String name) {
        super(name);
    }


    public void setUp() throws Exception {
        super.setUp();               
        c1=createChannel("A");
        c2=createChannel("A");
    }


    public void tearDown() throws Exception {        
        if(c2 != null)
            c2.close();
        if(c1 != null)
            c1.close();
        super.tearDown();
    }

    public void testSingleJoin() throws ChannelException {
        c1.connect("X");
        View v=c1.getView();
        assertNotNull(v);
        assertEquals(1, v.size());
    }


    /**
     * Tests that immediately after a connect(), a getView() returns the correct view
     * @throws ChannelException
     */
    public void testJoinsOnTwoChannels() throws ChannelException {
        c1.connect("X");
        c2.connect("X");
        
        //no blocking is used, let the view propagate
        Util.sleep(2000);
        
        View v1=c1.getView(), v2=c2.getView();
        System.out.println("v1=" + v1 + ", v2=" + v2);
        assertNotNull(v1);
        assertNotNull(v2);
        assertEquals(2, v1.size());
        assertEquals(2, v2.size());
        assertEquals(v1, v2);
    }


    public void testJoinsOnTwoChannelsAndSend() throws ChannelException {
        MyReceiver r1=new MyReceiver("c1");
        MyReceiver r2=new MyReceiver("c2");
        c1.setReceiver(r1);
        c2.setReceiver(r2);
        Message m1=new Message(null, null, "message-1"), m2=new Message(null, null, "message-2");
        c1.connect("X");
        View view=c1.getView();
        assertEquals("c1's view: " + view, 1, view.size());
        c2.connect("X");
        view=c2.getView();
        assertEquals("c2's view: " + view, 2, view.size());
        Util.sleep(200);
        view=c1.getView();
        assertEquals("c1's view: " + view, 2, view.size());

        c1.send(m1);
        c2.send(m2);

        Util.sleep(1500);
        List c1_list=r1.getMsgs(), c2_list=r2.getMsgs();
        System.out.println("c1: " + c1_list.size() + " msgs, c2: " + c2_list.size() + " msgs");
        assertNotNull(c1_list);
        assertNotNull(c2_list);
        assertEquals("cl_list: " + c1_list, 2, c1_list.size());
        assertEquals("c2_list: " + c2_list, 2, c2_list.size());
        assertTrue(c1_list.contains("message-1"));
        assertTrue(c2_list.contains("message-1"));
        assertTrue(c1_list.contains("message-2"));
        assertTrue(c2_list.contains("message-2"));
    }


    public static Test suite() {
        return new TestSuite(JoinTest.class);
    }

    public static void main(String[] args) {
        junit.textui.TestRunner.run(JoinTest.suite());
    }


    private static class MyReceiver extends ReceiverAdapter {
        private final String name;
        private final List<String> msgs;

        public MyReceiver(String name) {
            this.name=name;
            msgs = Collections.synchronizedList(new ArrayList<String>());
        }

        public List getMsgs() {
            return msgs;
        }

        public void clear() {msgs.clear();}

        public void receive(Message msg) {
            String s=(String)msg.getObject();
            msgs.add(s);
            System.out.println("[" + name + "] received " + s + " from " + msg.getSrc());
        }

        public void viewAccepted(View new_view) {
            System.out.println("[" + name + "] view: " + new_view);
        }
    }
}
