package org.jboss.cache.interceptors;

import org.jboss.cache.CacheException;
import org.jboss.cache.CacheSPI;
import org.jboss.cache.Fqn;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.NodeSPI;
import org.jboss.cache.loader.CacheLoader;
import org.jboss.cache.lock.NodeLock;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.marshall.MethodCallFactory;
import org.jboss.cache.marshall.MethodDeclarations;
import org.jboss.cache.transaction.GlobalTransaction;
import org.jboss.cache.transaction.TransactionEntry;
import org.jboss.cache.transaction.TransactionTable;

import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;

/**
 * Loads nodes that don't exist at the time of the call into memory from the CacheLoader
 *
 * @author Bela Ban
 * @version $Id: CacheLoaderInterceptor.java,v 1.92 2007/07/17 14:10:21 msurtani Exp $
 */
public class CacheLoaderInterceptor extends Interceptor implements CacheLoaderInterceptorMBean
{
   private long m_cacheLoads = 0;
   private long m_cacheMisses = 0;
   private TransactionTable txTable = null;
   protected boolean isActivation = false;
   protected CacheLoader loader;

   /**
    * True if CacheStoreInterceptor is in place.
    * This allows us to skip loading keys for remove(Fqn, key) and put(Fqn, key).
    * It also affects removal of node data and listing children.
    */
   protected boolean useCacheStore = true;

   public void setCache(CacheSPI cache)
   {
      super.setCache(cache);
      txTable = cache.getTransactionTable();
      this.loader = cache.getCacheLoaderManager().getCacheLoader();
   }

   /**
    * Makes sure a node is loaded into memory before a call executes (no-op if node is already loaded). If attributes
    * of a node are to be accessed by the method, the attributes are also loaded.
    *
    * @return
    * @throws Throwable
    */
   public Object invoke(InvocationContext ctx) throws Throwable
   {
      MethodCall m = ctx.getMethodCall();
      Fqn fqn = null, fqn2 = null;// if set, load the data.  fqn2 for 2nd fqn in move().

      Object[] args = m.getArgs();
      boolean acquireLock = false;// do we need to acquire a lock if we load this node from cloader?

      boolean initNode = false;// keep uninitialized
      Object key = null;
      TransactionEntry entry = null;
      GlobalTransaction gtx;
      boolean recursive = false;// do we also load children?
      boolean bypassLoadingData = false;


      if ((gtx = ctx.getGlobalTransaction()) != null)
      {
         entry = txTable.get(gtx);
      }

      if (log.isTraceEnabled())
      {
         log.trace("invoke " + m);
      }
      switch (m.getMethodId())
      {
         case MethodDeclarations.putDataEraseMethodLocal_id:
         case MethodDeclarations.putDataMethodLocal_id:
            fqn = (Fqn) args[1];
            initNode = true;
            break;
         case MethodDeclarations.putForExternalReadMethodLocal_id:
         case MethodDeclarations.putKeyValMethodLocal_id:
            fqn = (Fqn) args[1];
            key = args[2];
            if (useCacheStore)
            {
               initNode = true;
            }
            else
            {
               acquireLock = true;
            }
            break;
         case MethodDeclarations.moveMethodLocal_id:
            fqn = (Fqn) args[0];
            fqn2 = (Fqn) args[1];
            acquireLock = true;
            //initNode = true;
            recursive = true;
            break;
         case MethodDeclarations.addChildMethodLocal_id:
            fqn = (Fqn) args[1];
            break;
         case MethodDeclarations.getKeyValueMethodLocal_id:
            fqn = (Fqn) args[0];
            key = args[1];
            acquireLock = true;
            break;
         case MethodDeclarations.getNodeMethodLocal_id:
         case MethodDeclarations.getChildrenNamesMethodLocal_id:
            bypassLoadingData = true;
         case MethodDeclarations.getKeysMethodLocal_id:
         case MethodDeclarations.releaseAllLocksMethodLocal_id:
         case MethodDeclarations.printMethodLocal_id:
            fqn = (Fqn) args[0];
            acquireLock = true;
            break;
         case MethodDeclarations.rollbackMethod_id:
            // clean up nodesCreated map
            cleanupNodesCreated(entry);
            break;
         default:
            if (!useCacheStore)
            {
               if (m.getMethodId() == MethodDeclarations.removeKeyMethodLocal_id)
               {
                  fqn = (Fqn) args[1];
               }
               else if (m.getMethodId() == MethodDeclarations.removeDataMethodLocal_id)
               {
                  fqn = (Fqn) args[1];
                  initNode = true;
               }
            }
            break;
      }

      /* On the way in: load elements into cache from the CacheLoader if not yet in the cache. We need to synchronize
      this so only 1 thread attempts to load a given element */

      if (fqn != null)
      {
         if (fqn2 != null)
         {
            loadIfNeeded(ctx, fqn2, key, initNode, acquireLock, m, entry, false, m.getMethodId() == MethodDeclarations.moveMethodLocal_id, bypassLoadingData);
         }
         loadIfNeeded(ctx, fqn, key, initNode, acquireLock, m, entry, recursive, m.getMethodId() == MethodDeclarations.moveMethodLocal_id, bypassLoadingData);
      }

      return super.invoke(ctx);
   }


   private void loadIfNeeded(InvocationContext ctx, Fqn fqn, Object key, boolean initNode, boolean acquireLock, MethodCall m, TransactionEntry entry, boolean recursive, boolean isMove, boolean bypassLoadingData) throws Throwable
   {
      NodeSPI n = cache.peek(fqn, true);

      boolean mustLoad = mustLoad(n, key);
      if (log.isTraceEnabled())
      {
         log.trace("load element " + fqn + " mustLoad=" + mustLoad);
      }
      if (mustLoad)
      {
         if (initNode)
         {
            n = createTempNode(fqn, entry);
         }

         // Only attempt to acquire this lock if we need to - i.e., if
         // the lock hasn't already been acquired by the Lock
         // interceptor.  CRUD methods (put, remove) would have acquired
         // this lock - even if the node is not in memory and needs to be
         // loaded.  Non-CRUD methods (put) would NOT have acquired this
         // lock so if we are to load the node from cache loader, we need
         // to acquire a write lock here.  as a 'catch-all', DO NOT
         // attempt to acquire a lock here *anyway*, even for CRUD
         // methods - this leads to a deadlock when you have threads
         // simultaneously trying to create a node.  See
         // org.jboss.cache.loader.deadlock.ConcurrentCreationDeadlockTest
         // - Manik Surtani (21 March 2006)
         if (acquireLock)
         {
            lock(fqn, NodeLock.LockType.WRITE, false);// non-recursive for now
         }

//         if (!initNode && !wasRemovedInTx(fqn, ctx.getGlobalTransaction()))
         if (!wasRemovedInTx(fqn, ctx.getGlobalTransaction()))
         {
            if (bypassLoadingData)
            {
               if (n == null && loader.exists(fqn))
               {
                  // just create a dummy node in memory
                  n = createTempNode(fqn, entry);
               }
            }
            else
            {
               n = loadNode(ctx, fqn, n, entry);
            }
         }
      }

      // The complete list of children aren't known without loading them
      if (recursive || m.getMethodId() == MethodDeclarations.getChildrenNamesMethodLocal_id)
      {
         loadChildren(fqn, n, recursive, isMove);
      }
   }

   /**
    * Load the children.
    *
    * @param node may be null if the node was not found.
    */
   private void loadChildren(Fqn fqn, NodeSPI node, boolean recursive, boolean isMove) throws Throwable
   {

      if (node != null && node.isChildrenLoaded())
      {
         return;
      }
      Set children_names = loader.getChildrenNames(fqn);

      if (log.isTraceEnabled())
      {
         log.trace("load children " + fqn + " children=" + children_names);
      }

      // For getChildrenNames null means no children
      if (children_names == null)
      {
         if (node != null)
         {
            if (useCacheStore)
            {
               node.removeChildrenDirect();//getChildrenMapDirect().clear();
            }
            node.setChildrenLoaded(true);
         }
         return;
      }

      // Create if node had not been created already
      if (node == null)
      {
         node = createNodes(fqn, null);// dont care about local transactions
      }

      // Create one DataNode per child, mark as UNINITIALIZED
      for (Object name : children_names)
      {
         Fqn child_fqn = new Fqn(name);// this is a RELATIVE Fqn!!

         // create child if it didn't exist
         NodeSPI child = node.addChildDirect(child_fqn);
         if ((isMove || isActivation) && recursive)
         {
            // load data for children as well!
            child.putAllDirect(loader.get(child.getFqn()));
            child.setDataLoaded(true);
         }

         // why are we doing this?!??
//         else
//         {
//            child.setDataLoaded(false);
//         }

         if (recursive)
         {
            loadChildren(child.getFqn(), child, true, isMove);
         }
      }
      lock(fqn, recursive ? NodeLock.LockType.WRITE : NodeLock.LockType.READ, true);// recursive=true: lock entire subtree
      node.setChildrenLoaded(true);
   }

   private boolean mustLoad(NodeSPI n, Object key)
   {
      if (n == null)
      {
         log.trace("must load, node null");
         return true;
      }
      // if we are not looking for a specific key don't bother loading!
      if (key == null)
      {
         log.trace("don't load, key requested is null");
         return false;
      }
      if (n.getKeysDirect().contains(key))
      {
         log.trace("don't load, already have necessary key in memory");
         return false;
      }
      if (!n.isDataLoaded())
      {
         log.trace("must Load, uninitialized");
         return true;
      }
      return false;
   }

   public long getCacheLoaderLoads()
   {
      return m_cacheLoads;
   }

   public long getCacheLoaderMisses()
   {
      return m_cacheMisses;
   }

   public void resetStatistics()
   {
      m_cacheLoads = 0;
      m_cacheMisses = 0;
   }

   public Map<String, Object> dumpStatistics()
   {
      Map<String, Object> retval = new HashMap<String, Object>();
      retval.put("CacheLoaderLoads", m_cacheLoads);
      retval.put("CacheLoaderMisses", m_cacheMisses);
      return retval;
   }

   protected void lock(Fqn fqn, NodeLock.LockType lock_type, boolean recursive) throws Throwable
   {

      if (configuration.isNodeLockingOptimistic()) return;

      MethodCall m = MethodCallFactory.create(MethodDeclarations.lockMethodLocal,
              fqn, lock_type, recursive);

      // hacky
      cache.getInterceptorChain().get(0).invoke(InvocationContext.fromMethodCall(m));
//      super.invoke(m);
   }

   /**
    * Retrieves a node from memory; doesn't access the cache loader
    *
    * @param fqn
    */
   protected NodeSPI getNode(Fqn fqn)
   {
      return cache.peek(fqn, true);
      //      int treeNodeSize = fqn.size();
      //
      //      // root node
      //      Node n = cache.getRoot();
      //      Node child_node;
      //      Object child_name;
      //      for (int i = 0; i < treeNodeSize && n != null; i++)
      //      {
      //         child_name = fqn.get(i);
      //         cache.getInvocationContext().getOptionOverrides().setBypassInterceptorChain(true);
      //         child_node = n.getChild(new Fqn(child_name));
      //         n = child_node;
      //      }
      //      return n;
   }

   /**
    * Returns true if the FQN or parent was removed during the current
    * transaction.
    * This is O(N) WRT to the number of modifications so far.
    */
   private boolean wasRemovedInTx(Fqn fqn, GlobalTransaction t)
   {
      if (t == null)
      {
         return false;
      }
      TransactionEntry entry = txTable.get(t);
      for (MethodCall m : entry.getCacheLoaderModifications())
      {
         if (m.getMethodId() == MethodDeclarations.removeNodeMethodLocal_id
                 && fqn.isChildOrEquals((Fqn) m.getArgs()[1]))
         {
            return true;
         }
      }
      return false;
   }

   /**
    * Loads a node from disk; if it exists creates parent TreeNodes.
    * If it doesn't exist on disk but in memory, clears the
    * uninitialized flag, otherwise returns null.
    */
   private NodeSPI loadNode(InvocationContext ctx, Fqn fqn, NodeSPI n, TransactionEntry entry) throws Exception
   {
      if (log.isTraceEnabled()) log.trace("loadNode " + fqn);
      Map nodeData = loadData(fqn);
      if (nodeData != null)
      {
         log.trace("Node data is not null, loading");

         cache.getNotifier().notifyNodeLoaded(fqn, true, Collections.emptyMap(), ctx);
         if (isActivation)
         {
            cache.getNotifier().notifyNodeActivated(fqn, true, Collections.emptyMap(), ctx);
         }

         n = createNodes(fqn, entry);
//         n.clearDataDirect();
         n.putAllDirect(nodeData);

         cache.getNotifier().notifyNodeLoaded(fqn, false, nodeData, ctx);
         if (isActivation)
         {
            cache.getNotifier().notifyNodeActivated(fqn, false, nodeData, ctx);
         }
      }

      if (n != null && !n.isDataLoaded())
      {
         log.trace("Setting dataLoaded to true");
         n.setDataLoaded(true);
      }
      return n;
   }

   /**
    * Creates a new memory node in preparation for storage.
    */
   private NodeSPI createTempNode(Fqn fqn, TransactionEntry entry) throws Exception
   {
      NodeSPI n = createNodes(fqn, entry);
      n.setDataLoaded(false);
      if (log.isTraceEnabled())
      {
         log.trace("createTempNode n " + n);
      }
      return n;
   }


   private NodeSPI createNodes(Fqn fqn, TransactionEntry entry) throws Exception
   {
      Fqn tmp_fqn = Fqn.ROOT;

      int size = fqn.size();

      // root node
      NodeSPI n = cache.getRoot();
      for (int i = 0; i < size; i++)
      {
         Object child_name = fqn.get(i);
         tmp_fqn = new Fqn(tmp_fqn, child_name);

         NodeSPI child_node = findChild(n, child_name);
         boolean last = (i == size - 1);

         if (child_node == null)
         {
            if (last)
            {
               child_node = n.addChildDirect(new Fqn(child_name));
               child_node.setDataLoaded(true);
            }
            else
            {
               child_node = n.addChildDirect(new Fqn(child_name));
               child_node.setDataLoaded(false);
            }

            if (entry != null)
            {
               entry.loadUninitialisedNode(tmp_fqn);
            }
         }

         n = child_node;
      }

      return n;
   }

   private NodeSPI findChild(NodeSPI node, Object child_name)
   {
      Map children = node.getChildrenMapDirect();
      if (children == null) return null;
      return (NodeSPI) children.get(child_name);
   }

   private void cleanupNodesCreated(TransactionEntry entry)
   {
      boolean traceEnabled = log.isTraceEnabled();
      log.trace("Removing temporarily created nodes from treecache");

      // this needs to be done in reverse order.
      List list = entry.getDummyNodesCreatedByCacheLoader();
      if (list != null && list.size() > 0)
      {
         ListIterator i = list.listIterator(list.size());
         while (i.hasPrevious())
         {
            Fqn fqn = (Fqn) i.previous();
            try
            {
               cache.evict(fqn, false);
            }
            catch (CacheException e)
            {
               if (traceEnabled) log.trace("Unable to evict node " + fqn, e);
            }
         }
      }
   }


   private Map loadData(Fqn fqn) throws Exception
   {

      Map nodeData = loader.get(fqn);
      boolean nodeExists = (nodeData != null);
      if (log.isTraceEnabled()) log.trace("nodeExists " + nodeExists);

      if (configuration.getExposeManagementStatistics() && getStatisticsEnabled())
      {
         if (nodeExists)
         {
            m_cacheLoads++;
         }
         else
         {
            m_cacheMisses++;
         }
      }

      // BES Jan-4-2007 Stop doing this; it's annoying and people
      // should have converted by now
      //      if (!nodeExists && isCustomCacheLoader)
      //      {
      //         warnCustom();
      //      }

      // BES Jan-21-2007 Do the notifications in loadNode, before and after
      // we create nodes
//      if (nodeExists)
//      {
//         cache.getNotifier().notifyNodeLoaded(fqn, true, Collections.emptyMap(), true);
//         cache.getNotifier().notifyNodeLoaded(fqn, false, nodeData, true);
//
//         if (isActivation)
//         {
//            cache.getNotifier().notifyNodeActivated(fqn, true, true);
//            cache.getNotifier().notifyNodeActivated(fqn, false, true);
//         }
//      }

      return nodeData;
   }

   private void warnCustom()
   {
      log.warn("CacheLoader.get(Fqn) returned a null; assuming the node nodes not exist.");
      log.warn("The CacheLoader interface has changed since JBossCache 1.3.x");
      log.warn("Please see http://jira.jboss.com/jira/browse/JBCACHE-118");
      log.warn("CacheLoader.get() should return an empty Map if the node does exist but doesn't have any attributes.");
   }

}