package org.jboss.cache.interceptors;

import org.jboss.cache.InvocationContext;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.config.Option;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.marshall.MethodDeclarations;
import org.jboss.cache.transaction.GlobalTransaction;

/**
 * Takes care of replicating modifications to other nodes in a cluster. Also
 * listens for prepare(), commit() and rollback() messages which are received
 * 'side-ways' (see docs/design/Refactoring.txt).
 *
 * @author Bela Ban
 * @version $Id: ReplicationInterceptor.java,v 1.44 2007/05/23 15:22:03 msurtani Exp $
 */
public class ReplicationInterceptor extends BaseRpcInterceptor
{

   public Object invoke(InvocationContext ctx) throws Throwable
   {
      MethodCall m = ctx.getMethodCall();
      GlobalTransaction gtx = ctx.getGlobalTransaction();

      // bypass for buddy group org metod calls.
      if (MethodDeclarations.isBuddyGroupOrganisationMethod(m.getMethodId())) return super.invoke(ctx);

      boolean isLocalCommitOrRollback = gtx != null && !gtx.isRemote() && (m.getMethodId() == MethodDeclarations.commitMethod_id || m.getMethodId() == MethodDeclarations.rollbackMethod_id);

      if (log.isTraceEnabled()) log.trace("isLocalCommitOrRollback? " + isLocalCommitOrRollback + "; gtx = " + gtx);

      // pass up the chain if not a local commit or rollback (in which case replicate first)
      Object o = isLocalCommitOrRollback ? null : super.invoke(ctx);
//       ctx = cache.getInvocationContext();

      Option optionOverride = ctx.getOptionOverrides();

      if (optionOverride != null && optionOverride.isCacheModeLocal() && ctx.getTransaction() == null)
      {
         log.trace("skip replication");
         return isLocalCommitOrRollback ? super.invoke(ctx) : o;
      }

      // could be potentially TRANSACTIONAL. If so, we register for transaction completion callbacks (if we
      // have not yet done so
      if (ctx.getTransaction() != null)
      {
         if (gtx != null && !gtx.isRemote())
         {
            // lets see what sort of method we've got.
            switch (m.getMethodId())
            {
               case MethodDeclarations.commitMethod_id:
                  // REPL_ASYNC will result in only a prepare() method - 1 phase commit.
                  if (containsModifications(ctx)) replicateCall(m, configuration.isSyncCommitPhase());
                  // now pass up the chain
                  o = super.invoke(ctx);
                  break;
               case MethodDeclarations.prepareMethod_id:
                  if (containsModifications(ctx))
                  {
                     // this is a prepare method
                     runPreparePhase(m, gtx);
                  }
                  break;
               case MethodDeclarations.rollbackMethod_id:
                  // REPL_ASYNC will result in only a prepare() method - 1 phase commit.
                  if (containsModifications(ctx) && !ctx.isLocalRollbackOnly())
                  {
                     replicateCall(m, configuration.isSyncRollbackPhase());
                  }
                  // now pass up the chain
                  o = super.invoke(ctx);
                  break;
               case MethodDeclarations.putForExternalReadMethodLocal_id:
                  cache.getTransactionTable().get(gtx).setForceAsyncReplication(true);
            }
         }
      }
      else if (MethodDeclarations.isCrudMethod(m.getMethodId()))
      {
         // NON-TRANSACTIONAL and CRUD method
         if (log.isTraceEnabled()) log.trace("Non-tx crud meth");
         if (ctx.isOriginLocal())
         {
            // don't re-broadcast if we've received this from anotehr cache in the cluster.
            handleReplicatedMethod(m, configuration.getCacheMode());
         }
      }
      else
      {
         if (log.isTraceEnabled()) log.trace("Non-tx and non crud meth");
      }

      return o;
   }

   void handleReplicatedMethod(MethodCall m, Configuration.CacheMode mode) throws Throwable
   {
      if (log.isTraceEnabled())
      {
         log.trace("invoking method " + m + ", members=" + cache.getMembers() + ", mode=" +
                 configuration.getCacheMode() + ", exclude_self=" + true + ", timeout=" +
                 configuration.getSyncReplTimeout());
      }
      if (mode == Configuration.CacheMode.REPL_ASYNC || m.getMethodId() == MethodDeclarations.putForExternalReadMethodLocal_id)
      {
         // 2. Replicate change to all *other* members (exclude self !)
         replicateCall(m, false);
      }
      else
      {
         // REVISIT Needs to exclude itself and apply the local change manually.
         // This is needed such that transient field is modified properly in-VM.
         replicateCall(m, true);
      }
   }

   /**
    * Calls prepare(GlobalTransaction,List,org.jgroups.Address,boolean)) in all members except self.
    * Waits for all responses. If one of the members failed to prepare, its return value
    * will be an exception. If there is one exception we rethrow it. This will mark the
    * current transaction as rolled back, which will cause the
    * afterCompletion(int) callback to have a status
    * of <tt>MARKED_ROLLBACK</tt>. When we get that call, we simply roll back the
    * transaction.<br/>
    * If everything runs okay, the afterCompletion(int)
    * callback will trigger the @link #runCommitPhase(GlobalTransaction)).
    * <br/>
    *
    * @throws Exception
    */
   protected void runPreparePhase(MethodCall prepareMethod, GlobalTransaction gtx) throws Throwable
   {
      boolean async = configuration.getCacheMode() == Configuration.CacheMode.REPL_ASYNC;
      if (log.isTraceEnabled())
      {
         log.trace("(" + cache.getLocalAddress() + "): running remote prepare for global tx " + gtx + " with async mode=" + async);
      }

      // this method will return immediately if we're the only member (because exclude_self=true)
      replicateCall(prepareMethod, !async);
   }
}
