package performancetests;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.CacheImpl;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.config.Configuration.CacheMode;
import org.jboss.cache.factories.UnitTestCacheConfigurationFactory;
import org.jboss.cache.transaction.DummyTransactionManager;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.transaction.UserTransaction;
import java.util.Properties;

/**
 * A test created to mesure performance in normal replication, it can be triggered from console to allow
 * several caches to be intantiated.
 *  
 * modified by dnielben
 * @author Marian Nikolov
 * @author $Author: vblagojevic $
 * @version $Date: 2007/04/16 17:37:13 $
 */
public class ReplicationPerformanceTest extends TestCase
{

   // Number of worker threads
   private static final int NUM_WORKERS = 2;

   // The number of test runs to perform.
   private static final int NUM_RUNS = 1;
   
   // The number of puts and gets that are going to be performed on each worker
   private static final int NUM_PUTS = 1000;
   

   // useful to increase this to get thread dumps, etc.  Typically should be set to 10,000 though.
   private static final long LOCK_ACQUISITION_TIMEOUT = 100000;

   /**
    * The initial context factory properties.
    */
   private static final Properties PROPERTIES;
   /**
    * The context factory to be used for the test.
    */
   private static final String CONTEXT_FACTORY =
           "org.jboss.cache.transaction.DummyContextFactory";
   /**
    * The original context factory to be restored after the test.
    */
   private String m_contextFactory = null;

   /**
    * Exception recorded if any of the worker threads fails.
    */
   private static volatile Exception exception = null;

   /**
    * The source cache where we put modifications.
    */
   private CacheImpl srcCache = null;
   /**
    * The target cache where we replicate modifications.
    */
   private CacheImpl dstCache = null;

   private Log log = LogFactory.getLog(ReplicationPerformanceTest.class);

   static
   {
      PROPERTIES = new Properties();
      PROPERTIES.put(Context.INITIAL_CONTEXT_FACTORY,
              "org.jboss.cache.transaction.DummyContextFactory");
   }

   /**
    * Constructor.
    *
    * @param name The test name.
    */
   public ReplicationPerformanceTest(String name)
   {
      super(name);
   }

   /**
    * {@inheritDoc}
    */
   protected void setUp() throws Exception
   {
      super.setUp();
      exception = null;
      m_contextFactory = System.getProperty(Context.INITIAL_CONTEXT_FACTORY);
      System.setProperty(Context.INITIAL_CONTEXT_FACTORY, CONTEXT_FACTORY);
      DummyTransactionManager.getInstance();
      // setup and start the source cache
      srcCache = (CacheImpl) DefaultCacheFactory.getInstance().createCache(false);
      srcCache.setConfiguration(UnitTestCacheConfigurationFactory.createConfiguration(CacheMode.REPL_SYNC));
      srcCache.getConfiguration().setTransactionManagerLookupClass(
              "org.jboss.cache.transaction.DummyTransactionManagerLookup");
      srcCache.getConfiguration().setCacheMode(Configuration.CacheMode.REPL_SYNC);

      srcCache.getConfiguration().setSyncCommitPhase(true);
      srcCache.getConfiguration().setLockAcquisitionTimeout(LOCK_ACQUISITION_TIMEOUT);
      srcCache.create();
      srcCache.start();
      // setup and start the destination cache
      dstCache = (CacheImpl) DefaultCacheFactory.getInstance().createCache(false);
      dstCache.setConfiguration(UnitTestCacheConfigurationFactory.createConfiguration(CacheMode.REPL_SYNC));
      dstCache.getConfiguration().setTransactionManagerLookupClass(
              "org.jboss.cache.transaction.DummyTransactionManagerLookup");
      dstCache.getConfiguration().setCacheMode(Configuration.CacheMode.REPL_SYNC);

      dstCache.getConfiguration().setSyncCommitPhase(true);
      dstCache.getConfiguration().setLockAcquisitionTimeout(LOCK_ACQUISITION_TIMEOUT);
      dstCache.create();
      dstCache.start();
   }

   /**
    * {@inheritDoc}
    */
   protected void tearDown() throws Exception
   {
      super.tearDown();
      DummyTransactionManager.destroy();
      srcCache.stop();
      srcCache = null;
      dstCache.stop();
      dstCache = null;
      if (m_contextFactory != null)
      {
         System.setProperty(Context.INITIAL_CONTEXT_FACTORY,
                 m_contextFactory);
         m_contextFactory = null;
      }
   }

   /**
    * Test for a synchronously replicated cache with concurrent transactions on
    * the same node.<br>
    * This test fails very often with a TimeoutException.
    *
    * @throws Exception Any exception if thrown by the cache.
    */
   public void testConcurrentReplicatedTransaction() throws Exception
   {
      performTest();
   }

   /**
    * Perform a single test, using the pre-configured cache.
    *
    * @throws Exception Any exception if thrown by the cache.
    */
   private void performTest() throws Exception
   {
      // repeat the test several times since it's not always reproducible
      for (int i = 0; i < NUM_RUNS; i++)
      {
         if (exception != null)
         {
            // terminate the test on the first failed worker
            fail("Due to an exception: " + exception);
         }
         // start several worker threads to work with the same FQN
         Worker[] t = new Worker[NUM_WORKERS];
         for (int j = 0; j < t.length; j++)
         {
            t[j] = new Worker("worker " + i + ":" + j);
            t[j].start();
         }
         // wait for all workers to complete before repeating the test
         for (Worker aT : t) aT.join();
      }
   }

   /**
    * Returns a user transaction to be associated with the calling thread.
    *
    * @return A user transaction.
    * @throws Exception Any exception thrown by the context lookup.
    */
   private UserTransaction getTransaction() throws Exception
   {
      return (UserTransaction) new InitialContext(PROPERTIES)
              .lookup("UserTransaction");
   }

   /**
    * A worker thread that applies the concurrent modifications.
    *
    * @author Marian Nikolov
    * @author $Author: vblagojevic $
    * @version $Date: 2007/04/16 17:37:13 $
    */
   private class Worker extends Thread
   {
      /**
       * Constructor.
       */
      public Worker(String name)
      {
         super(name);
      }

      /**
       * {@inheritDoc}
       */
      public void run()
      {
         try
         {
        	//Thread.yield();
        	for(int i=0; i < NUM_PUTS; i++){ 
            //UserTransaction tx = getTransaction();
            //log.warn("begin");
            //tx.begin();
            //log.warn("put");
            srcCache.put(new Fqn("Node" + i), Boolean.FALSE, Boolean.TRUE);
            srcCache.get(new Fqn("Node" + i));
            //Thread.yield();
            //log.warn("commit");
            //tx.commit();
            //log.warn("leave");
        	}
         }
         catch (Exception e)
         {
            log.error("caught exception " + e, e);
            exception = e;
         }
      }
   }

   public static Test suite()
   {
      return new TestSuite(ReplicationPerformanceTest.class);
   }

   public static void main(String[] args) throws Exception
   {
      junit.textui.TestRunner.run(suite());
   }
}
