package org.jboss.cache;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.lock.IsolationLevel;
import org.jboss.cache.notifications.annotation.CacheListener;
import org.jboss.cache.notifications.annotation.NodeCreated;
import org.jboss.cache.notifications.event.Event;
import org.jboss.cache.transaction.DummyTransactionManager;

import javax.transaction.NotSupportedException;
import javax.transaction.Transaction;

/**
 * Tests whether modifications within callbacks (TreeCacheListener) are handled correctly
 *
 * @author Bela Ban
 * @version $Id: CallbackTest.java,v 1.20 2007/07/03 09:05:26 msurtani Exp $
 */
public class CallbackTest extends TestCase
{
   CacheImpl cache = null, cache2;
   Transaction tx = null;
   final Fqn FQN_A = Fqn.fromString("/a");
   final Fqn FQN_B = Fqn.fromString("/b");
   final String KEY = "key";
   final String VALUE = "value";

   protected void setUp() throws Exception
   {
      super.setUp();
   }

   protected void tearDown() throws Exception
   {
      super.tearDown();
      if (cache != null)
      {
         cache.stop();
         cache.destroy();
         cache = null;
      }
      if (tx != null)
      {
         tx.commit();
         tx = null;
      }
   }


   public void testLocalPutCallbackWithoutTransaction() throws Exception, NotSupportedException
   {
      cache = createCache(Configuration.CacheMode.LOCAL, IsolationLevel.SERIALIZABLE);
      cache.addCacheListener(new PutListener(cache));

      cache.put(FQN_A, null);
      assertTrue(cache.exists(FQN_A));
      assertTrue(cache.exists(FQN_B));//created by callback
      assertEquals(cache.getLockTable().size(), 0);
      System.out.println("cache locks:\n" + cache.printLockInfo());
      assertEquals(0, cache.getNumberOfLocksHeld());
   }

   public void testLocalGetCallbackSameFqnWithoutTransaction() throws Exception, NotSupportedException
   {
      cache = createCache(Configuration.CacheMode.LOCAL, IsolationLevel.SERIALIZABLE);
      cache.getNotifier().addCacheListener(new GetListener(cache, FQN_A));

      cache.put(FQN_A, null);
      assertTrue(cache.exists(FQN_A));
      assertEquals(cache.getLockTable().size(), 0);
      System.out.println("cache locks:\n" + cache.printLockInfo());
      assertEquals(0, cache.getNumberOfLocksHeld());
   }

   public void testLocalGetCallbackDifferentFqnWithoutTransaction() throws Exception, NotSupportedException
   {
      cache = createCache(Configuration.CacheMode.LOCAL, IsolationLevel.SERIALIZABLE);
      cache.put(FQN_B, null);
      cache.getNotifier().addCacheListener(new GetListener(cache, FQN_B));

      cache.put("/a", null);
      assertTrue(cache.exists(FQN_A));
      assertTrue(cache.exists(FQN_B));
      assertEquals(cache.getLockTable().size(), 0);
      System.out.println("cache locks:\n" + cache.printLockInfo());
      assertEquals(0, cache.getNumberOfLocksHeld());
   }


   public void testLocalCallbackWithTransaction() throws Exception, NotSupportedException
   {
      cache = createCache(Configuration.CacheMode.LOCAL, IsolationLevel.SERIALIZABLE);
      cache.getNotifier().addCacheListener(new PutListener(cache));
      tx = startTransaction();
      cache.put(FQN_A, null);
      tx.commit();
      assertTrue(cache.exists(FQN_A));
      assertEquals(0, cache.getNumberOfLocksHeld());
   }


   public void testLocalCallbackWithException() throws Exception, NotSupportedException
   {
      cache = createCache(Configuration.CacheMode.LOCAL, IsolationLevel.SERIALIZABLE);
      cache.getNotifier().addCacheListener(new ExceptionListener());
      tx = startTransaction();
      try
      {
         cache.put(FQN_A, null);
         tx.rollback();
      }
      catch (RuntimeException ex)
      {
         tx.rollback();
      }
      assertFalse(cache.exists(FQN_A));
      assertEquals(0, cache.getNumberOfLocksHeld());
   }

   CacheImpl createCache(Configuration.CacheMode mode, IsolationLevel level) throws Exception
   {
      Configuration c = new Configuration();
      c.setCacheMode(mode);
      c.setIsolationLevel(level);
      c.setTransactionManagerLookupClass("org.jboss.cache.transaction.DummyTransactionManagerLookup");
      return (CacheImpl) DefaultCacheFactory.getInstance().createCache(c);
   }

   Transaction startTransaction()
   {
      DummyTransactionManager mgr = DummyTransactionManager.getInstance();
      try
      {
         mgr.begin();
         return mgr.getTransaction();
      }
      catch (Throwable t)
      {
         return null;
      }
   }

   @CacheListener
   public class ExceptionListener
   {
      @NodeCreated
      public void nodeCreated(Event e)
      {
         if (e.isPre()) throw new RuntimeException("this will cause the TX to rollback");
      }
   }


   @CacheListener
   public class GetListener
   {
      CacheImpl c;
      Fqn my_fqn;

      public GetListener(CacheImpl c, Fqn my_fqn)
      {
         this.c = c;
         this.my_fqn = my_fqn;
      }

      @NodeCreated
      public void nodeCreated(Event e)
      {
         if (!e.isPre())
         {
            try
            {
               Node n = c.get(this.my_fqn);
               assertNotNull(n);
            }
            catch (CacheException ex)
            {
               fail("listener was unable to do a get(" + my_fqn + ") during callback: " + ex);
            }
         }
      }

   }

   @CacheListener
   public class PutListener
   {
      CacheImpl c;

      public PutListener(CacheImpl c)
      {
         this.c = c;
      }

      @NodeCreated
      public void nodeCreated(Event e)
      {
         if (!e.isPre())
         {
            try
            {
               if (!c.exists(FQN_B))
               {
                  System.out.println("PutListener: creating node " + FQN_B);
                  c.put(FQN_B, KEY, VALUE);
                  System.out.println("PutListener: created node " + FQN_B);
               }
            }
            catch (CacheException ex)
            {
               fail("listener was unable to update cache during callback: " + ex);
            }
         }
      }

   }

   public static Test suite()
   {
      return new TestSuite(CallbackTest.class);
   }

}
