package org.jboss.cache;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.notifications.annotation.CacheListener;
import org.jboss.cache.notifications.annotation.CacheStarted;
import org.jboss.cache.notifications.annotation.CacheStopped;
import org.jboss.cache.notifications.event.Event;
import org.jboss.cache.transaction.DummyTransactionManager;

import javax.transaction.NotSupportedException;
import javax.transaction.RollbackException;
import javax.transaction.SystemException;
import javax.transaction.Transaction;

/**
 * Tests restart (stop-destroy-create-start) of CacheImpl
 *
 * @author Bela Ban
 * @version $Id: LifeCycleTest.java,v 1.15 2007/07/17 14:59:10 msurtani Exp $
 */
public class LifeCycleTest extends TestCase
{

   private static Log log = LogFactory.getLog(LifeCycleTest.class);

   public void testLocalRestartNoTransactions() throws Exception
   {
      CacheImpl cache = createCache(Configuration.CacheMode.LOCAL);
      cache.create();
      cache.start();

      cache.put("/a/b/c", null);
      assertTrue(cache.getNumberOfNodes() > 0);
      assertEquals(0, cache.getNumberOfLocksHeld());

      System.out.println("cache locks before restart:\n" + cache.printLockInfo());
      restartCache(cache);
      System.out.println("cache locks after restart:\n" + cache.printLockInfo());

      assertEquals(0, cache.getNumberOfNodes());
      assertEquals(0, cache.getNumberOfLocksHeld());
   }


   public void testLocalRestartWithTransactions() throws Exception
   {
      CacheImpl cache = createCache(Configuration.CacheMode.LOCAL);
      cache.create();
      cache.start();

      Transaction tx = beginTransaction();

      cache.put("/a/b/c", null);
      log.debug("cache locks before restart:\n" + cache.printLockInfo());
      assertTrue(cache.getNumberOfNodes() > 0);
      assertEquals(4, cache.getNumberOfLocksHeld());

      restartCache(cache);
      log.debug("cache locks after restart:\n" + cache.printLockInfo());

      //assertEquals(4, cache.getNumberOfLocksHeld());
      assertEquals(0, cache.getNumberOfNodes());

      tx.rollback();
      assertEquals(0, cache.getNumberOfLocksHeld());
   }

   public void testStartNoCreate() throws Exception
   {
      CacheImpl cache = createCache(Configuration.CacheMode.LOCAL);
      cache.start();

      cache.put("/a/b/c", null);
      assertTrue(cache.getNumberOfNodes() > 0);
      assertEquals(0, cache.getNumberOfLocksHeld());

      System.out.println("cache locks before restart:\n" + cache.printLockInfo());
      restartCache(cache);
      System.out.println("cache locks after restart:\n" + cache.printLockInfo());

      assertEquals(0, cache.getNumberOfNodes());
      assertEquals(0, cache.getNumberOfLocksHeld());
   }

   public void testReStartNoCreate() throws Exception
   {
      CacheImpl cache = createCache(Configuration.CacheMode.LOCAL);
      cache.start();
      cache.stop();
      cache.start();

      cache.put("/a/b/c", null);
      assertTrue(cache.getNumberOfNodes() > 0);
      assertEquals(0, cache.getNumberOfLocksHeld());

      System.out.println("cache locks before restart:\n" + cache.printLockInfo());
      restartCache(cache);
      System.out.println("cache locks after restart:\n" + cache.printLockInfo());

      assertEquals(0, cache.getNumberOfNodes());
      assertEquals(0, cache.getNumberOfLocksHeld());
   }

   public void testDuplicateInvocation() throws Exception
   {
      CacheImpl cache = createCache(Configuration.CacheMode.LOCAL);
      cache.create();
      cache.start();
      cache.create();
      cache.start();

      cache.put("/a/b/c", null);
      assertTrue(cache.getNumberOfNodes() > 0);
      assertEquals(0, cache.getNumberOfLocksHeld());

      System.out.println("cache locks before restart:\n" + cache.printLockInfo());
      restartCache(cache);
      System.out.println("cache locks after restart:\n" + cache.printLockInfo());

      assertEquals(0, cache.getNumberOfNodes());
      assertEquals(0, cache.getNumberOfLocksHeld());

      cache.stop();
      cache.destroy();
      cache.stop();
      cache.destroy();
   }

   public void testFailedStart() throws Exception
   {

      CacheImpl cache = createCache(Configuration.CacheMode.LOCAL);
      assertEquals("Correct state", CacheStatus.INSTANTIATED, cache.getCacheStatus());

      DisruptLifecycleListener listener = new DisruptLifecycleListener();
      cache.addCacheListener(listener);

      cache.create();

      listener.disrupt = true;

      assertEquals("Correct state", CacheStatus.CREATED, cache.getCacheStatus());
      try
      {
         cache.start();
         fail("Listener did not prevent start");
      }
      catch (CacheException good)
      {
      }

      assertEquals("Correct state", CacheStatus.FAILED, cache.getCacheStatus());

      cache.addCacheListener(listener);
      listener.disrupt = false;

      cache.start();

      assertEquals("Correct state", CacheStatus.STARTED, cache.getCacheStatus());

      cache.put("/a/b/c", null);
      assertTrue(cache.getNumberOfNodes() > 0);
      assertEquals(0, cache.getNumberOfLocksHeld());

      listener.disrupt = true;
      cache.addCacheListener(listener);

      try
      {
         cache.stop();
         fail("Listener did not prevent stop");
      }
      catch (CacheException good)
      {
      }

      assertEquals("Correct state", CacheStatus.FAILED, cache.getCacheStatus());

      listener.disrupt = false;

      cache.stop();
      assertEquals("Correct state", CacheStatus.STOPPED, cache.getCacheStatus());
      cache.destroy();
      assertEquals("Correct state", CacheStatus.DESTROYED, cache.getCacheStatus());
   }

   public void testInvalidStateInvocations() throws Exception
   {
      Cache c = createCache(Configuration.CacheMode.LOCAL);
      try
      {
         try
         {
            c.get(Fqn.ROOT, "k");
            fail("Cache isn't ready!");
         }
         catch (IllegalStateException good)
         {
         }

         c.create();
         try
         {
            c.get(Fqn.ROOT, "k");
            fail("Cache isn't ready!");
         }
         catch (IllegalStateException good)
         {
         }

         c.start();
         c.get(Fqn.ROOT, "k"); // should work

         c.stop();

         try
         {
            c.get(Fqn.ROOT, "k");
            fail("Cache isn't ready!");
         }
         catch (IllegalStateException good)
         {
         }

         c.destroy();
         try
         {
            c.get(Fqn.ROOT, "k");
            fail("Cache isn't ready!");
         }
         catch (IllegalStateException good)
         {
         }
      }
      finally
      {
         c.stop();
      }
   }

   public void testRemoteInvalidStateInvocations() throws Exception
   {
      CacheImpl c1 = createCache(Configuration.CacheMode.REPL_SYNC);
      CacheImpl c2 = createCache(Configuration.CacheMode.REPL_SYNC);
      try
      {
         // need to start them both first to ensure they see each other
         c1.start();
         c2.start();

         // now DIRECTLY change the status of c2.
         // emulate the race condition where the remote cache is stopping but hasn't disconnected from the channel.
         c2.cacheStatus = CacheStatus.STOPPING;

         try
         {
            c1.put(Fqn.ROOT, "k", "v");
            fail("Cache isn't ready!");
         }
         catch (IllegalStateException good)
         {
         }
      }
      finally
      {
         c1.stop();
         c2.stop();
      }
   }


   public void testInvalidStateTxCommit() throws Exception
   {
      CacheImpl c = createCache(Configuration.CacheMode.LOCAL);
      try
      {
         c.start();

         c.getTransactionManager().begin();
         c.put(Fqn.ROOT, "k1", "v1");
         c.put(Fqn.ROOT, "k2", "v2");

         // now DIRECTLY change the status of c.
         c.cacheStatus = CacheStatus.STOPPING;

         try
         {
            c.getTransactionManager().commit();
            fail("Cache isn't STARTED!");
         }
         catch (RollbackException good)
         {
         }
      }
      finally
      {
         c.stop();
      }
   }

   public void testInvalidStateTxRollback() throws Exception
   {
      CacheImpl c = createCache(Configuration.CacheMode.LOCAL);
      try
      {
         c.start();

         c.getTransactionManager().begin();
         c.put(Fqn.ROOT, "k1", "v1");
         c.put(Fqn.ROOT, "k2", "v2");

         // now DIRECTLY change the status of c.
         c.cacheStatus = CacheStatus.STOPPING;

         // rollbacks should just log a message
         c.getTransactionManager().rollback();
      }
      finally
      {
         c.stop();
      }
   }


   CacheImpl createCache(Configuration.CacheMode cache_mode) throws Exception
   {
      CacheImpl retval = (CacheImpl) DefaultCacheFactory.getInstance().createCache(false);
      retval.getConfiguration().setCacheMode(cache_mode);
      retval.getConfiguration().setTransactionManagerLookupClass("org.jboss.cache.transaction.DummyTransactionManagerLookup");
      return retval;
   }


   Transaction beginTransaction() throws SystemException, NotSupportedException
   {
      DummyTransactionManager mgr = DummyTransactionManager.getInstance();
      mgr.begin();
      Transaction tx = mgr.getTransaction();
      return tx;
   }


   void startCache(CacheImpl c) throws Exception
   {
      c.create();
      c.start();
   }

   void stopCache(CacheImpl c)
   {
      c.stop();
      c.destroy();
   }

   void restartCache(CacheImpl c) throws Exception
   {
      stopCache(c);
      startCache(c);
   }

   public static Test suite()
   {
      return new TestSuite(LifeCycleTest.class);
   }

   public static void main(String[] args)
   {
      junit.textui.TestRunner.run(suite());
   }

   @CacheListener
   public class DisruptLifecycleListener
   {
      private boolean disrupt;

      @CacheStarted
      public void cacheStarted(Event e)
      {
         if (disrupt) throw new IllegalStateException("I don't want to start");
      }

      @CacheStopped
      public void cacheStopped(Event e)
      {
         if (disrupt) throw new IllegalStateException("I don't want to stop");
      }

      public void setDisrupt(boolean disrupt)
      {
         this.disrupt = disrupt;
      }
   }
}
