package org.jboss.cache.api;

import junit.framework.TestCase;
import org.jboss.cache.CacheImpl;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.Node;
import org.jboss.cache.config.Configuration;

import javax.transaction.TransactionManager;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * Tests {@link org.jboss.cache.Node}-centric operations
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani</a>
 * @since 2.0.0
 */
public class NodeAPITest extends TestCase
{
   private Node rootNode;

   private CacheImpl cache;

   private TransactionManager tm;

   private static final Fqn A = Fqn.fromString("/a"), B = Fqn.fromString("/b"), C = Fqn.fromString("/c"), D = Fqn
           .fromString("/d");

   protected boolean optimistic = false;

   protected void setUp() throws Exception
   {
      // start a single cache instance
      cache = (CacheImpl) DefaultCacheFactory.getInstance().createCache("META-INF/local-tx-service.xml", false);
      cache.getConfiguration().setNodeLockingScheme(optimistic ? Configuration.NodeLockingScheme.OPTIMISTIC : Configuration.NodeLockingScheme.PESSIMISTIC);
      cache.start();
      rootNode = cache.getRoot();
      tm = cache.getTransactionManager();
   }

   protected void tearDown()
   {
      if (cache != null)
      {
         cache.stop();
      }
      if (rootNode != null)
      {
         rootNode = null;
      }
   }

   public void testAddingData()
   {
      Node nodeA = rootNode.addChild(A);
      nodeA.put("key", "value");

      assertEquals("value", nodeA.get("key"));
   }

   public void testAddingDataTx() throws Exception
   {
      tm.begin();
      Node nodeA = rootNode.addChild(A);
      nodeA.put("key", "value");

      assertEquals("value", nodeA.get("key"));
      tm.commit();
   }


   /**
    * Remember, Fqns are relative!!
    */
   public void testParentsAndChildren()
   {
      Node nodeA = rootNode.addChild(A);
      Node nodeB = nodeA.addChild(B);
      Node nodeC = nodeA.addChild(C);
      Node nodeD = rootNode.addChild(D);

      assertEquals(rootNode, nodeA.getParent());
      assertEquals(nodeA, nodeB.getParent());
      assertEquals(nodeA, nodeC.getParent());
      assertEquals(rootNode, nodeD.getParent());

      assertTrue(rootNode.hasChild(A));
      assertFalse(rootNode.hasChild(B));
      assertFalse(rootNode.hasChild(C));
      assertTrue(rootNode.hasChild(D));

      assertTrue(nodeA.hasChild(B));
      assertTrue(nodeA.hasChild(C));

      assertEquals(nodeA, rootNode.getChild(A));
      assertEquals(nodeD, rootNode.getChild(D));
      assertEquals(nodeB, nodeA.getChild(B));
      assertEquals(nodeC, nodeA.getChild(C));

      assertTrue(nodeA.getChildren().contains(nodeB));
      assertTrue(nodeA.getChildren().contains(nodeC));
      assertEquals(2, nodeA.getChildren().size());

      assertTrue(rootNode.getChildren().contains(nodeA));
      assertTrue(rootNode.getChildren().contains(nodeD));
      assertEquals(2, rootNode.getChildren().size());

      assertEquals(true, rootNode.removeChild(A));
      assertFalse(rootNode.getChildren().contains(nodeA));
      assertTrue(rootNode.getChildren().contains(nodeD));
      assertEquals(1, rootNode.getChildren().size());

      assertEquals("double remove", false, rootNode.removeChild(A));
      assertEquals("double remove", false, rootNode.removeChild(A.getLastElement()));
   }

   public void testLocking() throws Exception
   {
      tm.begin();
      Node nodeA = rootNode.addChild(A);
      Node nodeB = nodeA.addChild(B);
      Node nodeC = nodeB.addChild(C);

      if (!optimistic)
      {
         assertEquals(3, cache.getNumberOfNodes());
         assertEquals(4, cache.getNumberOfLocksHeld());
      }
      tm.commit();

      tm.begin();
      assertEquals(0, cache.getNumberOfLocksHeld());
      nodeC.put("key", "value");
      if (!optimistic) assertEquals(4, cache.getNumberOfLocksHeld());
      tm.commit();
   }

   public void testImmutabilityOfData()
   {
      rootNode.put("key", "value");
      Map m = rootNode.getData();
      try
      {
         m.put("x", "y");
         fail("Map should be immutable!!");
      }
      catch (Exception e)
      {
         // expected
      }

      try
      {
         rootNode.getKeys().add(new Object());
         fail("Key set should be immutable");
      }
      catch (Exception e)
      {
         // expected
      }
   }

   public void testImmutabilityOfChildren()
   {
      rootNode.addChild(A);

      try
      {
         rootNode.getChildren().clear();
         fail("Collection of child nodes returned in getChildrenDirect() should be immutable");
      }
      catch (Exception e)
      {
         // expected
      }
   }

   public void testGetChildrenUnderTx() throws Exception
   {
      tm.begin();
      cache.put(new Fqn(A, B), "1", "1");
      cache.put(new Fqn(A, C), "2", "2");

      if (!optimistic)
      {
         assertEquals(3, cache.getNumberOfNodes());
         assertEquals(4, cache.getNumberOfLocksHeld());
      }
      assertEquals("Number of child", 2, cache.getRoot().getChild(A).getChildren().size());
      tm.commit();
   }

   public void testGetChildAPI()
   {
      // creates a node with fqn /a/b/c
      rootNode.addChild(A).addChild(B).addChild(C);

      rootNode.getChild(A).put("key", "value");
      rootNode.getChild(A).getChild(B).put("key", "value");
      rootNode.getChild(A).getChild(B).getChild(C).put("key", "value");

      assertEquals("value", rootNode.getChild(A).get("key"));
      assertEquals("value", rootNode.getChild(A).getChild(B).get("key"));
      assertEquals("value", rootNode.getChild(A).getChild(B).getChild(C).get("key"));

      assertNull(rootNode.getChild(new Fqn("nonexistent")));
   }

   public void testClearingData()
   {
      rootNode.put("k", "v");
      rootNode.put("k2", "v2");
      assertEquals(2, rootNode.getKeys().size());
      rootNode.clearData();
      assertEquals(0, rootNode.getKeys().size());
      assertTrue(rootNode.getData().isEmpty());
   }

   public void testClearingDataTx() throws Exception
   {
      tm.begin();
      rootNode.put("k", "v");
      rootNode.put("k2", "v2");
      assertEquals(2, rootNode.getKeys().size());
      rootNode.clearData();
      assertEquals(0, rootNode.getKeys().size());
      assertTrue(rootNode.getData().isEmpty());
      tm.commit();
      assertTrue(rootNode.getData().isEmpty());
   }

   public void testPutData()
   {
      assertTrue(rootNode.getData().isEmpty());

      Map map = new HashMap();
      map.put("k1", "v1");
      map.put("k2", "v2");

      rootNode.putAll(map);

      assertEquals(2, rootNode.getData().size());
      assertEquals("v1", rootNode.get("k1"));
      assertEquals("v2", rootNode.get("k2"));

      map.clear();
      map.put("k3", "v3");

      rootNode.putAll(map);
      assertEquals(3, rootNode.getData().size());
      assertEquals("v1", rootNode.get("k1"));
      assertEquals("v2", rootNode.get("k2"));
      assertEquals("v3", rootNode.get("k3"));

      map.clear();
      map.put("k4", "v4");
      map.put("k5", "v5");

      rootNode.replaceAll(map);
      assertEquals(2, rootNode.getData().size());
      assertEquals("v4", rootNode.get("k4"));
      assertEquals("v5", rootNode.get("k5"));
   }

   public void testGetChildrenNames() throws Exception
   {
      rootNode.addChild(A).put("k", "v");
      rootNode.addChild(B).put("k", "v");

      Set childrenNames = new HashSet();
      childrenNames.add(A.getLastElement());
      childrenNames.add(B.getLastElement());

      assertEquals(childrenNames, rootNode.getChildrenNames());

      // now delete a child, within a tx
      tm.begin();
      rootNode.removeChild(B);
      assertFalse(rootNode.hasChild(B));
      childrenNames.remove(B.getLastElement());
      assertEquals(childrenNames, rootNode.getChildrenNames());
      tm.commit();
      assertEquals(childrenNames, rootNode.getChildrenNames());
   }

}
