/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.api;

import junit.framework.TestCase;
import org.jboss.cache.CacheSPI;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.Node;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.config.Configuration.CacheMode;
import org.jboss.cache.factories.UnitTestCacheConfigurationFactory;

import javax.transaction.TransactionManager;

public class NodeReplicatedMoveTest extends TestCase
{
   protected Node rootNode, nodeA, nodeB, nodeC, nodeD, nodeE;
   protected CacheSPI[] cache;
   protected TransactionManager tm;
   protected static final Fqn A = Fqn.fromString("/a"), B = Fqn.fromString("/b"), C = Fqn.fromString("/c"), D = Fqn.fromString("/d"), E = Fqn.fromString("/e");
   protected Object k = "key", vA = "valueA", vB = "valueB", vC = "valueC", vD = "valueD", vE = "valueE";

   protected boolean optimistic = false;


   protected void setUp() throws Exception
   {
      cache = new CacheSPI[2];

      // start a single cache instance      
      cache[0] = (CacheSPI) DefaultCacheFactory.getInstance().createCache(UnitTestCacheConfigurationFactory.createConfiguration(CacheMode.REPL_SYNC), false);
      cache[0].getConfiguration().setSyncCommitPhase(true);
      cache[0].getConfiguration().setSyncRollbackPhase(true);
      cache[0].getConfiguration().setNodeLockingScheme(optimistic ? Configuration.NodeLockingScheme.OPTIMISTIC : Configuration.NodeLockingScheme.PESSIMISTIC);
      cache[0].start();
      rootNode = cache[0].getRoot();
      tm = cache[0].getTransactionManager();

      //  start second instance
      cache[1] = (CacheSPI) DefaultCacheFactory.getInstance().createCache(UnitTestCacheConfigurationFactory.createConfiguration(CacheMode.REPL_SYNC), false);
      cache[1].getConfiguration().setSyncCommitPhase(true);
      cache[1].getConfiguration().setSyncRollbackPhase(true);
      cache[1].getConfiguration().setNodeLockingScheme(optimistic ? Configuration.NodeLockingScheme.OPTIMISTIC : Configuration.NodeLockingScheme.PESSIMISTIC);
      cache[1].start();
   }

   protected void tearDown()
   {
      if (cache != null)
      {
         if (cache[0] != null) cache[0].stop();
         if (cache[1] != null) cache[1].stop();
         cache = null;
      }
      if (rootNode != null) rootNode = null;
   }

   public void testReplicatability()
   {
      nodeA = rootNode.addChild(A);
      nodeB = nodeA.addChild(B);

      nodeA.put(k, vA);
      nodeB.put(k, vB);

      assertEquals(vA, cache[0].getRoot().getChild(A).get(k));
      assertEquals(vB, cache[0].getRoot().getChild(A).getChild(B).get(k));

      assertEquals(vA, cache[1].getRoot().getChild(A).get(k));
      assertEquals(vB, cache[1].getRoot().getChild(A).getChild(B).get(k));

      // now move...
      cache[0].move(nodeB.getFqn(), Fqn.ROOT);

      assertEquals(vA, cache[0].getRoot().getChild(A).get(k));
      assertEquals(vB, cache[0].getRoot().getChild(B).get(k));

      assertEquals(vA, cache[1].getRoot().getChild(A).get(k));
      assertEquals(vB, cache[1].getRoot().getChild(B).get(k));
   }

   public void testInvalidations() throws Exception
   {
      cache[0].stop();
      cache[1].stop();
      cache[0].destroy();
      cache[1].destroy();
      cache[0].getConfiguration().setCacheMode(Configuration.CacheMode.INVALIDATION_SYNC);
      cache[1].getConfiguration().setCacheMode(Configuration.CacheMode.INVALIDATION_SYNC);
      cache[0].start();
      cache[1].start();

      nodeA = rootNode.addChild(A);
      nodeB = nodeA.addChild(B);

      nodeA.put(k, vA);
      nodeB.put(k, vB);

      assertEquals(vA, cache[0].getRoot().getChild(A).get(k));
      assertEquals(vB, cache[0].getRoot().getChild(A).getChild(B).get(k));

      assertNull(cache[1].getRoot().getChild(A));
      assertNull(cache[1].getRoot().getChild(B));

      // now move...
      cache[0].move(nodeB.getFqn(), Fqn.ROOT);

      assertEquals(vA, cache[0].getRoot().getChild(A).get(k));
      assertEquals(vB, cache[0].getRoot().getChild(B).get(k));

      assertNull(cache[1].getRoot().getChild(A));
      assertNull(cache[1].getRoot().getChild(B));

      // now make sure a node exists on cache 2
      cache[1].getRoot().addChild(A);

      try
      {
         cache[0].move(cache[0].getRoot().getChild(B).getFqn(), cache[0].getRoot().getChild(A).getFqn());// should throw an NPE
         fail("Expecting an exception");
      }
      catch (Exception e)
      {
         // expected
      }
   }

   public void testReplTxCommit() throws Exception
   {
      Fqn A_B = new Fqn(A, B);
      nodeA = rootNode.addChild(A);
      nodeB = nodeA.addChild(B);

      nodeA.put(k, vA);
      nodeB.put(k, vB);

      assertEquals(vA, cache[0].getRoot().getChild(A).get(k));
      assertEquals(vB, cache[0].getRoot().getChild(A).getChild(B).get(k));

      assertEquals(vA, cache[1].getRoot().getChild(A).get(k));
      assertEquals(vB, cache[1].getRoot().getChild(A).getChild(B).get(k));

      // now move...
      tm.begin();
      cache[0].move(nodeB.getFqn(), Fqn.ROOT);

      assertEquals(vA, cache[0].get(A, k));
      assertNull(cache[0].get(A_B, k));
      assertEquals(vB, cache[0].get(B, k));
      tm.commit();

      assertEquals(vA, cache[0].getRoot().getChild(A).get(k));
      assertEquals(vB, cache[0].getRoot().getChild(B).get(k));
      assertEquals(vA, cache[1].getRoot().getChild(A).get(k));
      assertEquals(vB, cache[1].getRoot().getChild(B).get(k));

   }

   public void testReplTxRollback() throws Exception
   {
      nodeA = rootNode.addChild(A);
      nodeB = nodeA.addChild(B);

      nodeA.put(k, vA);
      nodeB.put(k, vB);

      assertEquals(vA, cache[0].getRoot().getChild(A).get(k));
      assertEquals(vB, cache[0].getRoot().getChild(A).getChild(B).get(k));
      assertEquals(vA, cache[1].getRoot().getChild(A).get(k));
      assertEquals(vB, cache[1].getRoot().getChild(A).getChild(B).get(k));

      // now move...
      tm.begin();
      cache[0].move(nodeB.getFqn(), Fqn.ROOT);

      assertEquals(vA, cache[0].get(A, k));
      assertEquals(vB, cache[0].get(B, k));

      tm.rollback();

      assertEquals(vA, cache[0].getRoot().getChild(A).get(k));
      assertEquals(vB, cache[0].getRoot().getChild(A).getChild(B).get(k));
      assertEquals(vA, cache[1].getRoot().getChild(A).get(k));
      assertEquals(vB, cache[1].getRoot().getChild(A).getChild(B).get(k));
   }

   public void testReplConcurrency()
   {
      //fail("Implement me");
      // TODO: Implement this test
   }

}
