/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */

package org.jboss.cache.api;

import junit.framework.TestCase;
import org.jboss.cache.Cache;
import org.jboss.cache.CacheSPI;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.Node;
import org.jboss.cache.config.Configuration.CacheMode;
import org.jboss.cache.config.Option;
import org.jboss.cache.factories.UnitTestCacheConfigurationFactory;
import org.jboss.cache.misc.TestingUtil;
import org.jboss.cache.transaction.DummyTransactionManager;

import javax.transaction.HeuristicMixedException;
import javax.transaction.HeuristicRollbackException;
import javax.transaction.NotSupportedException;
import javax.transaction.RollbackException;
import javax.transaction.SystemException;
import javax.transaction.Transaction;

/**
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
public class SyncReplTxTest extends TestCase
{
   private CacheSPI[] caches;

   protected void setUp()
   {
      System.out.println("*** In setUp()");
      caches = new CacheSPI[2];
      caches[0] = (CacheSPI) DefaultCacheFactory.getInstance().createCache(UnitTestCacheConfigurationFactory.createConfiguration(CacheMode.REPL_SYNC));
      caches[1] = (CacheSPI) DefaultCacheFactory.getInstance().createCache(UnitTestCacheConfigurationFactory.createConfiguration(CacheMode.REPL_SYNC));

      TestingUtil.blockUntilViewsReceived(caches, 5000);
      System.out.println("*** Finished setUp()");
   }

   protected void tearDown()
   {
      System.out.println("*** In tearDown()");
      if (caches != null)
      {
         for (CacheSPI c : caches)
         {
            if (c != null)
            {
               Transaction t;

               try
               {
                  if ((t = c.getTransactionManager().getTransaction()) != null)
                  {
                     try
                     {
                        t.rollback();
                     }
                     catch (SystemException e)
                     {
                        // do nothing
                     }
                  }
               }
               catch (SystemException e)
               {
                  // do nothing
               }
               c.stop();
            }
         }
         caches = null;
      }
      System.out.println("*** Finished tearDown()");
   }

   Transaction beginTransaction() throws SystemException, NotSupportedException
   {
      DummyTransactionManager mgr = DummyTransactionManager.getInstance();
      mgr.begin();
      return mgr.getTransaction();
   }

   public void testBasicOperation() throws SystemException, NotSupportedException, HeuristicMixedException, HeuristicRollbackException, RollbackException
   {
      assertClusterSize("Should only be 2  caches in the cluster!!!", 2);
      assertInvocationContextInitState();

      Fqn f = Fqn.fromString("/test/data");
      String k = "key", v = "value";

      assertNull("Should be null", caches[0].getRoot().getChild(f));
      assertNull("Should be null", caches[1].getRoot().getChild(f));

      Node node = caches[0].getRoot().addChild(f);

      assertNotNull("Should not be null", node);

      Transaction tx = beginTransaction();
      node.put(k, v);
      Transaction tx1 = caches[0].getInvocationContext().getTransaction();
      assertNotNull("Transaction can't be null ", tx1);
      tx.commit();

      assertEquals(v, node.get(k));
      assertEquals(v, caches[0].get(f, k));
      assertEquals("Should have replicated", v, caches[1].get(f, k));
   }

   private void assertClusterSize(String message, int size)
   {
      for (Cache c : caches)
      {
         assertClusterSize(message, size, c);
      }
   }

   private void assertClusterSize(String message, int size, Cache c)
   {
      assertEquals(message, size, c.getMembers().size());
   }

   private void assertInvocationContextInitState()
   {
      for (Cache c : caches)
      {
         assertInvocationContextInitState(c);
      }
   }

   private void assertInvocationContextInitState(Cache c)
   {
      InvocationContext ctx = c.getInvocationContext();
      InvocationContext control = null;
      try
      {
         control = ctx.clone();
      }
      catch (CloneNotSupportedException e)
      {
         fail("Unable to clone InvocationCOntext");
      }

      control.reset();
      control.setOptionOverrides(new Option());

      assertEquals("Should be equal", control, ctx);
   }


}
