/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.buddyreplication;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.CacheImpl;
import org.jboss.cache.Fqn;
import org.jboss.cache.misc.TestingUtil;

/**
 * Teststhe transfer of content under *normal* operation
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
public class BuddyReplicationContentTest extends BuddyReplicationTestsBase
{
   private String key = "key";
   private String value = "value";
   private Log log = LogFactory.getLog(BuddyGroupAssignmentTest.class);


   private void assertNoStaleLocks(CacheImpl[] caches)
   {
      for (CacheImpl cache : caches) assertNoStaleLocks(cache);
   }

   private void assertNoStaleLocks(CacheImpl cache)
   {
      assertEquals("Number of locks in cache instance " + cache.toString(true) + " should be 0", 0, cache.getNumberOfLocksHeld());
   }

   protected void setUp() throws Exception
   {
      log.debug("Starting setUp()");
      super.setUp();
      log.debug("Finishing setUp()");
   }

   public void testSimplePut() throws Exception
   {
      log.debug("Running testSimplePut");
      caches = createCaches(3, false);

      String fqn = "/test";
      String backupFqn = "/" + BuddyManager.BUDDY_BACKUP_SUBTREE + "/" + BuddyManager.getGroupNameFromAddress(caches[0].getLocalAddress()) + fqn;

      assertNoStaleLocks(caches);

      // put something in cache 1
      caches[0].put(fqn, key, value);

      assertNoStaleLocks(caches);

      // this should be in neither of the other caches' "main" trees
      assertEquals(value, caches[0].get(fqn, key));
      assertNull("Should be null", caches[1].get(fqn, key));
      assertNull("Should be null", caches[2].get(fqn, key));

      // check the backup trees

      assertEquals("Buddy should have data in backup tree", value, caches[1].get(backupFqn, key));
      assertNull("Should be null", caches[2].get(backupFqn, key));

      assertNoStaleLocks(caches);
   }

   public void testPutAndRemove() throws Exception
   {
      log.debug("Running testPutAndRemove");
      caches = createCaches(3, false);

      String fqn = "/test";
      String backupFqn = "/" + BuddyManager.BUDDY_BACKUP_SUBTREE + "/" + BuddyManager.getGroupNameFromAddress(caches[0].getLocalAddress()) + fqn;

      assertNoStaleLocks(caches);

      // put something in cache 1
      caches[0].put(fqn, key, value);

      assertNoStaleLocks(caches);

      // this should be in neither of the other caches' "main" trees
      assertEquals(value, caches[0].get(fqn, key));
      assertNull("Should be null", caches[1].get(fqn, key));
      assertNull("Should be null", caches[2].get(fqn, key));

      // check the backup trees

      assertEquals("Buddy should have data in backup tree", value, caches[1].get(backupFqn, key));
      assertNull("Should be null", caches[2].get(backupFqn, key));

      assertNoStaleLocks(caches);

      // now remove
      caches[0].remove(fqn);
      assertNoStaleLocks(caches);


      assertNull("Should be null", caches[0].get(fqn, key));
      assertNull("Should be null", caches[1].get(fqn, key));
      assertNull("Should be null", caches[2].get(fqn, key));

      // check the backup trees
      assertNull("Should be null", caches[0].get(backupFqn, key));
      assertNull("Should be null", caches[1].get(backupFqn, key));
      assertNull("Should be null", caches[2].get(backupFqn, key));

      assertNoStaleLocks(caches);
   }

   public void testPutAndRemove2() throws Exception
   {
      log.debug("Running testPutAndRemove2");
      caches = createCaches(2, 4, false);

      String fqn = "/test";
      String backupFqn = "/" + BuddyManager.BUDDY_BACKUP_SUBTREE + "/" + BuddyManager.getGroupNameFromAddress(caches[0].getLocalAddress()) + fqn;

      // put something in cache 1
      assertNoStaleLocks(caches);

      caches[0].put(fqn, key, value);

      assertNoStaleLocks(caches);

      // this should be in neither of the other caches' "main" trees
      assertEquals(value, caches[0].get(fqn, key));
      assertNull("Should be null", caches[1].get(fqn, key));
      assertNull("Should be null", caches[2].get(fqn, key));
      assertNull("Should be null", caches[3].get(fqn, key));

      // check the backup trees
      assertEquals("Buddy should have data in backup tree", value, caches[1].get(backupFqn, key));
      assertEquals("Buddy should have data in backup tree", value, caches[2].get(backupFqn, key));
      assertNull("Should be null", caches[3].get(backupFqn, key));

      assertNoStaleLocks(caches);

      // now remove
      caches[0].remove(fqn);
      assertNoStaleLocks(caches);

      assertNull("Should be null", caches[0].get(fqn, key));
      assertNull("Should be null", caches[1].get(fqn, key));
      assertNull("Should be null", caches[2].get(fqn, key));
      assertNull("Should be null", caches[3].get(fqn, key));

      // check the backup trees
      assertNull("Should be null", caches[0].get(backupFqn, key));
      assertNull("Should be null", caches[1].get(backupFqn, key));
      assertNull("Should be null", caches[2].get(backupFqn, key));
      assertNull("Should be null", caches[3].get(backupFqn, key));

      assertNoStaleLocks(caches);
   }

   public void testBuddyJoin() throws Exception
   {
      log.debug("Running testBuddyJoin");
      caches = createCaches(2, false);
      CacheImpl cache2 = null;

      try
      {
         Fqn fqn = Fqn.fromString("/test");
         Fqn backupFqn = BuddyManager.getBackupFqn(caches[1].getLocalAddress(), fqn);

         assertNoStaleLocks(caches);

         // put something in cache 1
         caches[1].put(fqn, key, value);

         assertNoStaleLocks(caches);

         // this should be in neither of the other caches' "main" trees
         assertEquals(value, caches[1].get(fqn, key));
         assertFalse("Should be false", caches[0].exists(fqn));

         // check the backup trees
         assertEquals("Buddy should have data in backup tree", value, caches[0].get(backupFqn, key));

         assertNoStaleLocks(caches);

         // now add a new cache to the cluster
         cache2 = createCache(1, null);

         // allow this cache a few msecs to join
         TestingUtil.blockUntilViewsReceived(3000, caches[0], caches[1], cache2);

         TestingUtil.sleepThread(2000); // allow buddy group reorg

         dumpCacheContents(caches[0], caches[1], cache2);

         // now caches[1]'s buddy should be cache2, not cache[0]
         assertIsBuddy(caches[1], cache2, true);
         // this should still be the same
         assertIsBuddy(caches[0], caches[1], true);
         // and cache2's buddy should be cache[0]
         assertIsBuddy(cache2, caches[0], true);

         // so now the backup data we saw on cache[0] should have been removed.
         assertFalse("This backup data should have been removed", caches[0].exists(backupFqn));

         // while cache2 should now posess this backup (due to a state transfer)
         assertEquals("Backup state should have been transferred to this new cache instance", value, cache2.get(backupFqn, key));

         caches[1].remove(fqn);
         assertNoStaleLocks(caches);


         assertFalse("Should be null", caches[0].exists(fqn));
         assertFalse("Should be null", caches[1].exists(fqn));
         assertFalse("Should be null", cache2.exists(fqn));

         // check the backup trees
         assertFalse("Should be null", caches[0].exists(backupFqn));
         assertFalse("Should be null", caches[1].exists(backupFqn));
         assertFalse("Should be null", cache2.exists(backupFqn));

         assertNoStaleLocks(caches);

      }
      finally
      {
         if (cache2 != null) cache2.stop();
      }
   }

   public void testCompleteStateSurvival() throws Exception
   {
      log.debug("Running testCompleteStateSurvival");
      caches = null;

      caches = createCaches(3, false, true);
      caches[0].put("/0", "key", "value");
      caches[1].put("/1", "key", "value");
      caches[2].put("/2", "key", "value");

      TestingUtil.sleepThread(getSleepTimeout());

      log.info("stopping 2");
      caches[2].stop();

      log.info("0 ** " + caches[0].printLockInfo());
      log.info("1 ** " + caches[1].printLockInfo());
      log.info("2 ** " + caches[2].printLockInfo());

      TestingUtil.sleepThread(getSleepTimeout());

      assertEquals("value", caches[0].get("/2", "key"));

      log.info("0 ** " + caches[0].printLockInfo());
      log.info("1 ** " + caches[1].printLockInfo());
      log.info("2 ** " + caches[2].printLockInfo());

      TestingUtil.sleepThread(getSleepTimeout());

      caches[1].stop();
      log.info("0 ** " + caches[0].printLockInfo());
      log.info("1 ** " + caches[1].printLockInfo());
      log.info("2 ** " + caches[2].printLockInfo());

      // cache[0] is all thats left!!

      assertEquals("value", caches[0].get("/0", "key"));

      try
      {
         assertEquals("value", caches[0].get("/1", "key"));
      }
      catch (Exception e)
      {
         // may barf the first time around since we are unable to contact our buddy and store this data.
         if (e instanceof RuntimeException)
         {
            assertEquals(IllegalArgumentException.class, e.getCause().getClass());
         }
         else throw e;
      }

      // now try the assertion again since the local gravitation would have worked.
      assertEquals("value", caches[0].get("/1", "key"));
      assertEquals("value", caches[0].get("/2", "key"));
   }
}
