/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.buddyreplication;

import org.jboss.cache.CacheImpl;
import org.jboss.cache.Fqn;
import org.jboss.cache.misc.TestingUtil;

/**
 * Tests behaviour when data owners fail - essentially this tests data gravitation
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
public class BuddyReplicationFailoverTest extends BuddyReplicationTestsBase
{
   protected boolean optimisticLocks = false;
   private String key = "key";
   private String value = "value";

   public void testDataGravitationKillOwner() throws Exception
   {
      testDataGravitation(true);
   }

   public void testDataGravitationDontKillOwner() throws Exception
   {
      testDataGravitation(false);
   }

   private void testDataGravitation(boolean killOwner) throws Exception
   {
      caches = createCaches(3, false, true, optimisticLocks);

      Fqn fqn = Fqn.fromString("/test");
      Fqn backupFqn = BuddyManager.getBackupFqn(caches[0].getLocalAddress(), fqn);

      dumpCacheContents(caches);

      caches[0].put(fqn, key, value);

      dumpCacheContents(caches);

      assertEquals("Value should exist", value, caches[0].get(fqn, key));

      dumpCacheContents(caches);

      // use exists instead of get() to prevent going up the interceptor stack
      assertTrue("Should be false", !caches[1].exists(fqn));
      assertTrue("Should be false", !caches[2].exists(fqn));

      assertFalse("Should be false", caches[0].exists(backupFqn));
      assertTrue("Value be true", caches[1].exists(backupFqn));
      assertFalse("Should be false", caches[2].exists(backupFqn));

      if (killOwner)
      {
         System.out.println("***** About to kill original data owner (" + caches[0].getLocalAddress() + ").  *****");
         // make sure the JGroups channel shutdown is forced.
//         System.setProperty("org.jboss.cache.shutdown.force", "true");
         caches[0].stop();
         caches[0] = null;
         // wait up to 5 mins?  :S
         TestingUtil.blockUntilViewsReceived(5 * 60 * 1000, false, caches[1], caches[2]);

//         TestingUtil.sleepThread(600000);
//         TestingUtil.sleepThread(1000);
      }

      System.out.println("***** Killed original data owner, about to call a get on a different cache instance.  *****");

      // according to data gravitation, a call to *any* cache should retrieve the data, and move the data to the new cache.
      assertEquals("Value should have gravitated", value, caches[2].get(fqn, key));

      TestingUtil.sleepThread(500);

      dumpCacheContents(caches);

      // now lets test the eviction part of gravitation
      Fqn newBackupFqn = BuddyManager.getBackupFqn(caches[2].getLocalAddress(), fqn);

      // use exists instead of get() to prevent going up the interceptor stack
      if (!killOwner) assertTrue("Should be false", !caches[0].exists(fqn));
      assertTrue("Should be false", !caches[1].exists(fqn));

      // the old backup should no longer exist
      if (!killOwner) assertFalse("Should be null", caches[0].exists(backupFqn));
      assertFalse("Should be null", caches[1].exists(backupFqn));
      assertFalse("Should be null", caches[2].exists(backupFqn));

      // and the backup should now exist in caches[2]'s buddy which is caches[0]
      if (killOwner)
      {
         assertEquals("Value should exist", value, caches[1].get(newBackupFqn, key));
      }
      else
      {
         assertEquals("Value should exist", value, caches[0].get(newBackupFqn, key));
         assertFalse("Should be null", caches[1].exists(newBackupFqn));
      }
      assertFalse("Should be null", caches[2].exists(newBackupFqn));
   }

   public void testDataReplicationSuppression() throws Exception
   {
      caches = createCaches(3, false, false, optimisticLocks);

      Fqn fqn = Fqn.fromString("/test");
      Fqn backupFqn = BuddyManager.getBackupFqn(caches[0].getLocalAddress(), fqn);
      caches[0].put(fqn, key, value);

      dumpCacheContents(caches);

      assertEquals("value", caches[0].get(fqn, key));
      assertFalse(caches[0].exists(backupFqn));
      assertEquals("value", caches[1].get(backupFqn, key));
      assertFalse(caches[1].exists(fqn));
      assertFalse(caches[2].exists(fqn));
      assertFalse(caches[2].exists(backupFqn));

      assertNoLocks(caches);

      backupFqn = BuddyManager.getBackupFqn(caches[1].getLocalAddress(), fqn);

      System.out.println("*** Calling get() on cache[1] with force option");

      caches[1].getInvocationContext().getOptionOverrides().setForceDataGravitation(true);
      assertEquals("value", caches[1].get(fqn, key));

      dumpCacheContents(caches);

      assertFalse(caches[1].exists(backupFqn));
      assertEquals("value", caches[2].get(backupFqn, key));
      assertFalse(caches[2].exists(fqn));
      assertFalse(caches[0].exists(fqn));
      assertFalse(caches[0].exists(backupFqn));
   }

   public void testSubtreeRetrieval() throws Exception
   {
      caches = createCaches(3, false, true, optimisticLocks);

      Fqn fqn = Fqn.fromString("/test");
      Fqn fqn2 = Fqn.fromString("/test/subtree");

      Fqn backupFqn = Fqn.fromString("/" + BuddyManager.BUDDY_BACKUP_SUBTREE + "/" + BuddyManager.getGroupNameFromAddress(caches[0].getLocalAddress()) + "/test");
      Fqn backupFqn2 = Fqn.fromString("/" + BuddyManager.BUDDY_BACKUP_SUBTREE + "/" + BuddyManager.getGroupNameFromAddress(caches[0].getLocalAddress()) + "/test/subtree");


      caches[0].put(fqn, key, value);
      caches[0].put(fqn2, key, value);

      // test backup replication to buddy
      assertEquals(value, caches[0].get(fqn, key));
      assertEquals(value, caches[0].get(fqn2, key));
      assertEquals(value, caches[1].get(backupFqn, key));
      assertEquals(value, caches[1].get(backupFqn2, key));

      assertTrue(!caches[0].exists(backupFqn));
      assertTrue(!caches[0].exists(backupFqn2));
      assertTrue(!caches[1].exists(fqn));
      assertTrue(!caches[1].exists(fqn2));
      assertTrue(!caches[2].exists(fqn));
      assertTrue(!caches[2].exists(fqn2));
      assertTrue(!caches[2].exists(backupFqn));
      assertTrue(!caches[2].exists(backupFqn2));

      assertNoLocks(caches);

      // gravitate to 2:
      caches[2].get(fqn);  // expect entire subtree to gravitate.

      Fqn newBackupFqn = Fqn.fromString("/" + BuddyManager.BUDDY_BACKUP_SUBTREE + "/" + BuddyManager.getGroupNameFromAddress(caches[2].getLocalAddress()) + "/test");
      Fqn newBackupFqn2 = Fqn.fromString("/" + BuddyManager.BUDDY_BACKUP_SUBTREE + "/" + BuddyManager.getGroupNameFromAddress(caches[2].getLocalAddress()) + "/test/subtree");

      assertEquals(value, caches[2].get(fqn, key));
      assertTrue(caches[2].exists(fqn2));
      assertEquals(value, caches[0].get(newBackupFqn, key));
      assertTrue(caches[0].exists(newBackupFqn2));

      assertTrue(!caches[2].exists(newBackupFqn));
      assertTrue(!caches[2].exists(newBackupFqn2));
      assertTrue(!caches[0].exists(fqn));
      assertTrue(!caches[0].exists(fqn2));
      assertTrue(!caches[1].exists(fqn));
      assertTrue(!caches[1].exists(fqn2));
      assertTrue(!caches[1].exists(newBackupFqn));
      assertTrue(!caches[1].exists(newBackupFqn2));

      for (CacheImpl cache : caches)
      {
         assertTrue(!cache.exists(backupFqn));
         assertTrue(!cache.exists(backupFqn2));
      }

      assertNoLocks(caches);
   }
}
