/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */

package org.jboss.cache.eviction;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import org.jboss.cache.CacheImpl;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.misc.TestingUtil;

import java.util.Properties;

/**
 * Tests cache behavior in the presence of concurrent passivation.
 *
 * @author Brian Stansberry
 * @version $Revision: 1.12 $
 */
public class ConcurrentEvictionTest extends TestCase
{
   private CacheImpl cache_;
   private int wakeupIntervalMillis_ = 0;
   private String tmpDir = System.getProperty("java.io.tmpdir", "/tmp");
   private String cacheLoaderDir = "/JBossCacheFileCacheLoader";

   public ConcurrentEvictionTest(String s)
   {
      super(s);
   }

   protected void setUp() throws Exception
   {
      super.setUp();
      initCaches();
      wakeupIntervalMillis_ = cache_.getConfiguration().getEvictionConfig().getWakeupIntervalSeconds() * 1000;
      if (wakeupIntervalMillis_ < 0)
      {
         fail("testEviction(): eviction thread wake up interval is illegal " + wakeupIntervalMillis_);
      }

   }

   void initCaches() throws Exception
   {
      TestingUtil.recursiveFileRemove(tmpDir + cacheLoaderDir);
      cache_ = (CacheImpl) DefaultCacheFactory.getInstance().createCache("META-INF/local-eviction-cacheloader-service.xml", false);// read in generic local xml
      cache_.getConfiguration().setTransactionManagerLookupClass("org.jboss.cache.transaction.DummyTransactionManagerLookup");
      Properties p = new Properties();
      p.put("location", tmpDir + cacheLoaderDir);
      cache_.getConfiguration().getCacheLoaderConfig().getFirstCacheLoaderConfig().setProperties(p);
      cache_.start();
   }

   public void tearDown() throws Exception
   {
      super.tearDown();
      TestingUtil.recursiveFileRemove(tmpDir + cacheLoaderDir);
      cache_.stop();
      cache_ = null;
   }

   public void testConcurrentEviction() throws Exception
   {
      Fqn base = Fqn.fromString("/org/jboss/test/data/concurrent/eviction");

      // Create a bunch of nodes; more than the /org/jboss/test/data
      // region's maxNodes so we know eviction will kick in
      for (int i = 0; i < 1000; i++)
      {
         cache_.put(new Fqn(base, i / 100), i, "value");
      }

      // Loop for long enough to have 5 runs of the eviction thread
      long loopDone = System.currentTimeMillis() + (5 * wakeupIntervalMillis_);
      while (System.currentTimeMillis() < loopDone)
      {
         // If any get returns null, that's a failure
         for (int i = 0; i < 1000; i++)
         {
            Fqn fqn = new Fqn(base, i / 100);
            Integer key = i;
            assertNotNull("found value under Fqn " + fqn + " and key " + key,
                    cache_.get(fqn, key));
         }
      }
   }

   public static Test suite()
   {
      return new TestSuite(ConcurrentEvictionTest.class);
   }
}
