/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.loader;

import junit.framework.Assert;
import junit.framework.Test;
import junit.framework.TestSuite;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.Cache;
import org.jboss.cache.CacheSPI;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.misc.TestingUtil;

import javax.transaction.TransactionManager;

/**
 * Tests using cache loaders with replicating data
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
public class CacheLoaderWithReplicationTest extends AbstractCacheLoaderTestBase
{
   private Cache cache1, cache2;
   private Fqn fqn = Fqn.fromString("/a");
   private String key = "key";

   private static final Log log = LogFactory.getLog(CacheLoaderWithReplicationTest.class);
   private CacheLoader loader1, loader2;
   private TransactionManager mgr1, mgr2;


   public void setUp() throws Exception
   {
      cache1 = DefaultCacheFactory.getInstance().createCache(false);
      cache1.getConfiguration().setCacheLoaderConfig(getSingleCacheLoaderConfig("", "org.jboss.cache.loader.DummyInMemoryCacheLoader", null, false, true, false));
      cache1.getConfiguration().setTransactionManagerLookupClass("org.jboss.cache.transaction.DummyTransactionManagerLookup");

      cache2 = DefaultCacheFactory.getInstance().createCache(false);
      cache2.getConfiguration().setCacheLoaderConfig(getSingleCacheLoaderConfig("", "org.jboss.cache.loader.DummyInMemoryCacheLoader", null, false, true, false));
      cache2.getConfiguration().setTransactionManagerLookupClass("org.jboss.cache.transaction.DummyTransactionManagerLookup");

      loader1 = loader2 = null;
      mgr1 = mgr2 = null;
   }

   public void tearDown() throws Exception
   {
      if (cache1 != null)
      {
         try
         {
            try
            {
               if (mgr1 != null) mgr1.rollback();
            }
            catch (Exception e)
            {

            }

            loader1.remove(fqn);
            cache1.removeNode(fqn);
            cache1.stop();
            cache1.destroy();
         }
         finally
         {
            cache1 = null;
            mgr1 = null;
            loader1 = null;
         }
      }

      if (cache2 != null)
      {
         try
         {
            try
            {
               if (mgr2 != null) mgr2.rollback();
            }
            catch (Exception e)
            {

            }
            loader2.remove(fqn);
            cache2.removeNode(fqn);
            cache2.stop();
            cache2.destroy();
         }
         finally
         {
            cache2 = null;
            mgr2 = null;
            loader2 = null;
         }

      }
   }

   private void createCaches(boolean sync, boolean optimistic) throws Exception
   {
      cache1.getConfiguration().setCacheMode(sync ? Configuration.CacheMode.REPL_SYNC : Configuration.CacheMode.REPL_ASYNC);
      cache2.getConfiguration().setCacheMode(sync ? Configuration.CacheMode.REPL_SYNC : Configuration.CacheMode.REPL_ASYNC);
      if (sync)
      {
         cache1.getConfiguration().setSyncCommitPhase(true);
         cache2.getConfiguration().setSyncCommitPhase(true);
         cache1.getConfiguration().setSyncRollbackPhase(true);
         cache2.getConfiguration().setSyncRollbackPhase(true);
      }

      if (optimistic)
      {
         cache1.getConfiguration().setNodeLockingScheme(Configuration.NodeLockingScheme.OPTIMISTIC);
         cache2.getConfiguration().setNodeLockingScheme(Configuration.NodeLockingScheme.OPTIMISTIC);
      }

      cache1.start();
      cache2.start();

      TestingUtil.blockUntilViewsReceived(10000, cache1, cache2);

      loader1 = ((CacheSPI) cache1).getCacheLoaderManager().getCacheLoader();
      loader2 = ((CacheSPI) cache2).getCacheLoaderManager().getCacheLoader();

      // make sure everything is empty...
      Assert.assertNull(loader1.get(fqn));
      Assert.assertNull(loader2.get(fqn));
      Assert.assertNull(cache1.getRoot().getChild(fqn));
      Assert.assertNull(cache2.getRoot().getChild(fqn));

      mgr1 = cache1.getConfiguration().getRuntimeConfig().getTransactionManager();
      mgr2 = cache2.getConfiguration().getRuntimeConfig().getTransactionManager();
   }

   public void testPessSyncRepl() throws Exception
   {
      createCaches(true, false);

      mgr1.begin();
      cache1.put(fqn, key, "value");

      Assert.assertEquals("value", cache1.get(fqn, key));
      Assert.assertNull(cache2.get(fqn, key));
      Assert.assertNull(loader1.get(fqn));
      Assert.assertNull(loader2.get(fqn));
      mgr1.commit();

      Assert.assertEquals("value", cache1.get(fqn, key));
      Assert.assertEquals("value", cache2.get(fqn, key));
      Assert.assertEquals("value", loader1.get(fqn).get(key));
      Assert.assertEquals("value", loader2.get(fqn).get(key));

      mgr1.begin();
      cache1.put(fqn, key, "value2");

      Assert.assertEquals("value2", cache1.get(fqn, key));
      Assert.assertEquals("value", cache2.get(fqn, key));
      Assert.assertEquals("value", loader1.get(fqn).get(key));
      Assert.assertEquals("value", loader2.get(fqn).get(key));

      mgr1.rollback();

      Assert.assertEquals("value", cache1.get(fqn, key));
      Assert.assertEquals("value", cache2.get(fqn, key));
      Assert.assertEquals("value", loader1.get(fqn).get(key));
      Assert.assertEquals("value", loader2.get(fqn).get(key));

   }

   public void testPessAsyncRepl() throws Exception
   {
      createCaches(false, false);

      mgr1.begin();
      cache1.put(fqn, key, "value");

      Assert.assertEquals("value", cache1.get(fqn, key));
      Assert.assertNull(cache2.get(fqn, key));
      Assert.assertNull(loader1.get(fqn));
      Assert.assertNull(loader2.get(fqn));
      mgr1.commit();

      TestingUtil.sleepThread(500);

      Assert.assertEquals("value", cache1.get(fqn, key));
      Assert.assertEquals("value", cache2.get(fqn, key));
      Assert.assertEquals("value", loader1.get(fqn).get(key));
      Assert.assertEquals("value", loader2.get(fqn).get(key));

      mgr1.begin();
      cache1.put(fqn, key, "value2");

      Assert.assertEquals("value2", cache1.get(fqn, key));
      Assert.assertEquals("value", cache2.get(fqn, key));
      Assert.assertEquals("value", loader1.get(fqn).get(key));
      Assert.assertEquals("value", loader2.get(fqn).get(key));

      mgr1.rollback();

      TestingUtil.sleepThread(500);

      Assert.assertEquals("value", cache1.get(fqn, key));
      Assert.assertEquals("value", cache2.get(fqn, key));
      Assert.assertEquals("value", loader1.get(fqn).get(key));
      Assert.assertEquals("value", loader2.get(fqn).get(key));
   }

   public void testOptSyncRepl() throws Exception
   {
      createCaches(true, true);
      mgr1.begin();
      cache1.put(fqn, key, "value");

      Assert.assertEquals("value", cache1.get(fqn, key));
      Assert.assertNull(cache2.get(fqn, key));
      Assert.assertNull(loader1.get(fqn));
      Assert.assertNull(loader2.get(fqn));
      mgr1.commit();

      Assert.assertEquals("value", cache1.get(fqn, key));
      Assert.assertEquals("value", cache2.get(fqn, key));
      Assert.assertEquals("value", loader1.get(fqn).get(key));
      Assert.assertEquals("value", loader2.get(fqn).get(key));

      mgr1.begin();
      cache1.put(fqn, key, "value2");

      Assert.assertEquals("value2", cache1.get(fqn, key));
      Assert.assertEquals("value", cache2.get(fqn, key));
      Assert.assertEquals("value", loader1.get(fqn).get(key));
      Assert.assertEquals("value", loader2.get(fqn).get(key));

      mgr1.rollback();

      Assert.assertEquals("value", cache1.get(fqn, key));
      Assert.assertEquals("value", cache2.get(fqn, key));
      Assert.assertEquals("value", loader1.get(fqn).get(key));
      Assert.assertEquals("value", loader2.get(fqn).get(key));
   }

   public void testOptAsyncRepl() throws Exception
   {
      createCaches(false, true);

      mgr1.begin();
      cache1.put(fqn, key, "value");

      Assert.assertEquals("value", cache1.get(fqn, key));
      Assert.assertNull(cache2.get(fqn, key));
      Assert.assertNull(loader1.get(fqn));
      Assert.assertNull(loader2.get(fqn));
      mgr1.commit();

      TestingUtil.sleepThread(500);

      Assert.assertEquals("value", cache1.get(fqn, key));
      Assert.assertEquals("value", cache2.get(fqn, key));
      Assert.assertEquals("value", loader1.get(fqn).get(key));
      Assert.assertEquals("value", loader2.get(fqn).get(key));

      mgr1.begin();
      cache1.put(fqn, key, "value2");

      Assert.assertEquals("value2", cache1.get(fqn, key));
      Assert.assertEquals("value", cache2.get(fqn, key));
      Assert.assertEquals("value", loader1.get(fqn).get(key));
      Assert.assertEquals("value", loader2.get(fqn).get(key));

      mgr1.rollback();
      TestingUtil.sleepThread(500);

      Assert.assertEquals("value", cache1.get(fqn, key));
      Assert.assertEquals("value", cache2.get(fqn, key));
      Assert.assertEquals("value", loader1.get(fqn).get(key));
      Assert.assertEquals("value", loader2.get(fqn).get(key));
   }

   public static Test suite()
   {
      return new TestSuite(CacheLoaderWithReplicationTest.class);
   }

   public static void main(String[] args)
   {
      junit.textui.TestRunner.run(suite());
   }

}