/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.loader;

import net.jcip.annotations.ThreadSafe;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.Modification;
import org.jboss.cache.config.CacheLoaderConfig.IndividualCacheLoaderConfig;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Dummy cache loader that stores data in memory
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
@ThreadSafe
public class DummyInMemoryCacheLoader extends AbstractCacheLoader
{
   protected Map<Fqn, DummyNode> nodes = new ConcurrentHashMap<Fqn, DummyNode>();
   protected Log log = LogFactory.getLog(DummyInMemoryCacheLoader.class);
   protected Map<Object, List<Modification>> transactions = new ConcurrentHashMap<Object, List<Modification>>();
   protected boolean debug; // whether to dump System.out messages as well as log messages or not
   protected final Object NULL = new Object()
   {
      public String toString()
      {
         return "NULL placeholder";
      }
   };

   public void setConfig(IndividualCacheLoaderConfig config)
   {
      if (config != null && config.getProperties() != null)
      {
         debug = Boolean.parseBoolean(config.getProperties().getProperty("debug", "false"));
      }
   }

   public IndividualCacheLoaderConfig getConfig()
   {
      return null;
   }

   public Set getChildrenNames(Fqn fqn) throws Exception
   {
      if (log.isDebugEnabled()) log.debug("Calling getChildrenNames on Fqn " + fqn + ".  Data map = " + nodes);
      debugMessage("Calling getChildrenNames on Fqn " + fqn + ".  Data map = " + nodes);
      if (!nodes.containsKey(fqn))
      {
         log.debug("node not in loader");
         debugMessage("node not in loader");
         return null;
      }

      Set children = findChildren(fqn);
      log.debug("Fqn " + fqn + " has children " + children);
      debugMessage("Fqn " + fqn + " has children " + children);
      // to keep in line with the CacheLoader interface contract for this method.
      return children.size() == 0 ? null : children;
   }

   private Set findChildren(Fqn p)
   {
      Set c = new HashSet();
      for (Fqn f : nodes.keySet())
      {
         if (!f.isRoot() && f.getParent().equals(p))
         {
            c.add(f.getLastElement());
         }
      }
      return c;
   }

   public Map get(Fqn name) throws Exception
   {
      DummyNode dn = nodes.get(name);
      Map<Object, Object> d = dn != null ? dn.data : null;

      if (log.isDebugEnabled()) log.debug("Getting data for fqn " + name + " = " + d);
      debugMessage("Getting data for fqn " + name + " = " + d);
      return stripNULLs(d);
   }

   private Map stripNULLs(Map data)
   {
      if (data == null) return null;
      // otherwise make sure we replace NULL placeholders with nulls.
      Map<Object, Object> d = new HashMap(data);
      if (d.containsKey(NULL))
      {
         Object v = d.remove(NULL);
         d.put(null, v);
      }
      Set keys = new HashSet();
      for (Map.Entry<?, ?> e : d.entrySet())
      {
         if (e.getValue() == NULL)
         {
            keys.add(e.getKey());
         }
      }
      for (Object k : keys)
      {
         d.put(k, null);
      }
      return d;
   }

   private Map injectNULLs(Map data)
   {
      if (data == null) return null;
      // otherwise make sure we replace NULL placeholders with nulls.
      Map<Object, Object> d = new HashMap(data);
      if (d.containsKey(null))
      {
         Object v = d.remove(null);
         d.put(NULL, v);
      }
      Set keys = new HashSet();
      for (Map.Entry<?, ?> e : d.entrySet())
      {
         if (e.getValue() == null)
         {
            keys.add(e.getKey());
         }
      }
      for (Object k : keys)
      {
         d.put(k, NULL);
      }
      return d;
   }


   public boolean exists(Fqn name) throws Exception
   {
      debugMessage("Performing exists() on " + name);
      return nodes.containsKey(name);
   }

   public Object put(Fqn name, Object key, Object value) throws Exception
   {
      DummyNode n = nodes.get(name);
      if (n == null)
      {
         n = new DummyNode(name);
      }
      Object k = key == null ? NULL : key;
      Object v = value == null ? NULL : value;
      Object old = n.data.put(k, v);

      nodes.put(name, n);
      // we need to make sure parents get put in as well.
      recursivelyPutParentsIfNeeded(name);
      if (log.isDebugEnabled()) log.debug("Did a put on " + name + ", data is " + n.data);
      debugMessage("Did a put on " + name + ", data is " + n.data);
      return old == NULL ? null : old;
   }

   public void put(Fqn name, Map attributes) throws Exception
   {
      DummyNode n = nodes.get(name);
      if (n == null)
      {
         n = new DummyNode(name);
      }
      if (attributes != null) n.data.putAll(injectNULLs(attributes));
      nodes.put(name, n);
      // we need to make sure parents get put in as well.
      recursivelyPutParentsIfNeeded(name);
      if (log.isDebugEnabled()) log.debug("Did a put on " + name + ", data is " + n.data);
      debugMessage("Did a put on " + name + ", data is " + n.data);
   }

   private void recursivelyPutParentsIfNeeded(Fqn node)
   {
      Fqn parent = node.getParent();
      if (nodes.containsKey(parent)) return; // nothing to do.

      // else put the parent in.
      nodes.put(parent, new DummyNode(parent));
      recursivelyPutParentsIfNeeded(parent);
   }

   public Object remove(Fqn fqn, Object key) throws Exception
   {
      log.debug("Removing data from " + fqn);
      debugMessage("Removing data from " + fqn);
      DummyNode n = nodes.get(fqn);
      if (n == null) n = new DummyNode(fqn);
      Object old = n.data.remove(key == null ? NULL : key);
      nodes.put(fqn, n);
      return old == NULL ? null : old;
   }

   public void remove(Fqn fqn) throws Exception
   {
      log.debug("Removing fqn " + fqn);
      debugMessage("Removing fqn " + fqn);
      nodes.remove(fqn);
      // remove children.
      recursivelyRemoveChildren(fqn);
   }

   private void recursivelyRemoveChildren(Fqn removedParent)
   {
      for (Fqn f : nodes.keySet())
      {
         if (f.getParent().equals(removedParent))
         {
            // remove the child node too
            nodes.remove(f);
            // and it's children.  Depth first.
            recursivelyRemoveChildren(f);
         }
      }
   }

   public void removeData(Fqn fqn) throws Exception
   {
      log.debug("Removing data from " + fqn);
      debugMessage("Removing data from " + fqn);
      DummyNode n = nodes.get(fqn);
      if (n == null) n = new DummyNode(fqn);
      n.data.clear();
      nodes.put(fqn, n);
   }

   public class DummyNode
   {
      Map data = new ConcurrentHashMap();
      Fqn fqn;

      public DummyNode(Fqn fqn)
      {
         this.fqn = fqn;
      }

      public String toString()
      {
         return "Node{" +
                 "data=" + data +
                 ", fqn=" + fqn +
                 '}';
      }
   }


   public String toString()
   {
      return "DummyInMemoryCacheLoader{" +
              "nodes=" + nodes +
              '}';
   }

   protected void debugMessage(String msg)
   {
      if (debug) System.out.println(" DummyInMemoryCacheLoader debug: " + msg);
   }
}
