package org.jboss.cache.loader;

import org.jboss.cache.CacheSPI;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.Node;
import org.jboss.cache.config.CacheLoaderConfig;
import org.jboss.cache.factories.InterceptorChainFactory;
import org.jboss.cache.interceptors.Interceptor;
import org.jmock.Mock;
import org.jmock.MockObjectTestCase;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
 * //TODO: MANIK: Javadoc this class
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani</a>
 * @since 2.0.0
 */
public class UnnecessaryLoadingTest extends MockObjectTestCase
{
   private CacheSPI cache;
   private Fqn parent = Fqn.fromString("/parent");
   private Fqn child = Fqn.fromString("/parent/child");
   private String k = "k", v = "v";
   private Mock mockCacheLoader;

   protected void setUp()
   {
      cache = (CacheSPI) DefaultCacheFactory.getInstance().createCache(false);
      CacheLoaderConfig clc = new CacheLoaderConfig();
      CacheLoaderConfig.IndividualCacheLoaderConfig iclc = new CacheLoaderConfig.IndividualCacheLoaderConfig();
      iclc.setClassName(DummyInMemoryCacheLoader.class.getName());
      clc.addIndividualCacheLoaderConfig(iclc);
      cache.getConfiguration().setCacheLoaderConfig(clc);
      cache.start();
      mockCacheLoader = mock(CacheLoader.class);

      cache.getCacheLoaderManager().setCacheLoader((CacheLoader) mockCacheLoader.proxy());
      // re-set the cache so that the mock CL is registered with the inerceptors.
      InterceptorChainFactory.getInstance().initialiseInterceptors((Interceptor) cache.getInterceptorChain().get(0), cache);

      // lifecycle stuff
      mockCacheLoader.expects(atMostOnce()).method("stop").withNoArguments();
      mockCacheLoader.expects(atMostOnce()).method("destroy").withNoArguments();
   }

   protected void tearDown()
   {
      cache.stop();
   }

   protected void assertDataLoaded(Fqn f)
   {
      assertTrue("Data should be loaded for node " + f, cache.peek(f, false).isDataLoaded());
   }

   protected void assertDataNotLoaded(Fqn f)
   {
      assertFalse("Data should not be loaded for node " + f, cache.peek(f, false).isDataLoaded());
   }


   public void testNoLoading()
   {
      // we expect these nodes to be stored.
      mockCacheLoader.expects(never()).method("get").withAnyArguments();
      mockCacheLoader.expects(never()).method("getChildrenNames").withAnyArguments();
      mockCacheLoader.expects(never()).method("exists").withAnyArguments();
      mockCacheLoader.expects(never()).method("remove").withAnyArguments();
      mockCacheLoader.expects(never()).method("removeData").withAnyArguments();

      mockCacheLoader.expects(once()).method("put").with(eq(parent), eq(k), eq(v));
      mockCacheLoader.expects(once()).method("put").with(eq(child), eq(k), eq(v));

      // create parent and child with data
      // new nodes being created, will result in loading them from the cache loader first
      mockCacheLoader.expects(once()).method("get").with(eq(parent));
      mockCacheLoader.expects(once()).method("get").with(eq(child));

      cache.put(parent, k, v);
      cache.put(child, k, v);

      // should be NO cache loading involved whatsoever
      // so no exceptions should be thrown by the mock CL
   }

   public void testLoadChild()
   {
      // we expect these nodes to be stored.
      mockCacheLoader.expects(never()).method("get").withAnyArguments();
      mockCacheLoader.expects(never()).method("getChildrenNames").withAnyArguments();
      mockCacheLoader.expects(never()).method("exists").withAnyArguments();
      mockCacheLoader.expects(never()).method("remove").withAnyArguments();
      mockCacheLoader.expects(never()).method("removeData").withAnyArguments();

      mockCacheLoader.expects(once()).method("put").with(eq(parent), eq(k), eq(v));
      mockCacheLoader.expects(once()).method("put").with(eq(child), eq(k), eq(v));

      // create parent and child with data
      // new nodes being created, will result in loading them from the cache loader first
      mockCacheLoader.expects(once()).method("get").with(eq(parent));
      mockCacheLoader.expects(once()).method("get").with(eq(child));

      cache.put(parent, k, v);
      cache.put(child, k, v);

      mockCacheLoader.expects(once()).method("get").with(eq(child));
      cache.evict(child, false);

      // there is no REAL cache loader so this return value should not be tested
      cache.get(child, k);

      // should be NO cache loading involved whatsoever
      // so no exceptions should be thrown by the mock CL
   }

   public void testDontLoadChild()
   {
      // we expect these nodes to be stored.
      mockCacheLoader.expects(never()).method("get").withAnyArguments();
      mockCacheLoader.expects(never()).method("getChildrenNames").withAnyArguments();
      mockCacheLoader.expects(never()).method("exists").withAnyArguments();
      mockCacheLoader.expects(never()).method("remove").withAnyArguments();
      mockCacheLoader.expects(never()).method("removeData").withAnyArguments();

      mockCacheLoader.expects(once()).method("put").with(eq(parent), eq(k), eq(v));
      mockCacheLoader.expects(once()).method("put").with(eq(child), eq(k), eq(v));

      // create parent and child with data
      // new nodes being created, will result in loading them from the cache loader first
      mockCacheLoader.expects(once()).method("get").with(eq(parent));
      mockCacheLoader.expects(once()).method("get").with(eq(child));

      cache.put(parent, k, v);
      cache.put(child, k, v);

      // nodes should be marked as having data loaded.
      assertDataLoaded(parent);
      assertDataLoaded(child);

      // now evict the parent
      cache.evict(parent, false);

      // only the evicted parent should have isDataLoaded() set to false.  Not the child.
      assertDataNotLoaded(parent);
      assertDataLoaded(child);
      assertNotNull(cache.peek(child, false));

      // there is no REAL cache loader so this return value should not be tested
      cache.get(child, k);

      // should be NO cache loading involved whatsoever
   }

   public void testUnnecessaryMultipleLoading()
   {
      mockCacheLoader.expects(never()).method("get").withAnyArguments();
      mockCacheLoader.expects(never()).method("getChildrenNames").withAnyArguments();
      mockCacheLoader.expects(never()).method("exists").withAnyArguments();
      mockCacheLoader.expects(never()).method("remove").withAnyArguments();
      mockCacheLoader.expects(never()).method("removeData").withAnyArguments();
      mockCacheLoader.expects(never()).method("put").withAnyArguments();

      Map m = new HashMap();
      m.put("foo", "bar");

      // expecting a put on parent and child
//      mockCacheLoader.expects(once()).method("get").with(eq(parent));
      mockCacheLoader.expects(once()).method("put").with(eq(child), eq(m));

      cache.put(child, m);
      assertDataNotLoaded(child);

      // should load child data
      mockCacheLoader.expects(once()).method("get").with(eq(child));
      cache.get(child, "foo");
      assertDataNotLoaded(child);
      cache.get(child, "foo");
      assertDataNotLoaded(child);
      cache.get(child, "foo2"); // does not exist, will trigger a load
      assertDataLoaded(child);

      // should not load
      Node node = cache.getRoot().getChild(parent);
      assertDataLoaded(child);
      assertDataNotLoaded(parent);

      // needs to load children at this stage in case there are other children that have been evicted.
      mockCacheLoader.expects(once()).method("getChildrenNames").with(eq(parent)).will(returnValue(Collections.singleton(child.getLastElementAsString())));

      Set children = node.getChildren(); //getchildrennames /parent
      assertEquals(1, children.size());
      assertDataLoaded(child);
      cache.get(child, "foo"); //get /parent/child
      assertDataLoaded(child);
   }

   public void testDontLoadDataWhenGettingNode()
   {
      mockCacheLoader.expects(never()).method("get").withAnyArguments();
      mockCacheLoader.expects(never()).method("getChildrenNames").withAnyArguments();
      mockCacheLoader.expects(never()).method("exists").withAnyArguments();
      mockCacheLoader.expects(never()).method("remove").withAnyArguments();
      mockCacheLoader.expects(never()).method("removeData").withAnyArguments();
      mockCacheLoader.expects(never()).method("put").withAnyArguments();

      mockCacheLoader.expects(once()).method("put").with(eq(parent), eq(k), eq(v));

      // create parent and child with data
      // new nodes being created, will result in loading them from the cache loader first
      mockCacheLoader.expects(once()).method("get").with(eq(parent));

      cache.put(parent, k, v);
      assertDataLoaded(parent);
      // evict the parent
      cache.evict(parent, false);
      assertNull(cache.peek(parent, false));

      mockCacheLoader.expects(once()).method("exists").with(eq(parent)).will(returnValue(true));
      // now get node.
      Node n = cache.getRoot().getChild(parent);
      assertNotNull(n);
      assertDataNotLoaded(parent);
   }

   public void testDontLoadDataWhenClearingNode()
   {
      mockCacheLoader.expects(never()).method("get").withAnyArguments();
      mockCacheLoader.expects(never()).method("getChildrenNames").withAnyArguments();
      mockCacheLoader.expects(never()).method("exists").withAnyArguments();
      mockCacheLoader.expects(never()).method("remove").withAnyArguments();
      mockCacheLoader.expects(never()).method("removeData").withAnyArguments();
      mockCacheLoader.expects(never()).method("put").withAnyArguments();

      mockCacheLoader.expects(once()).method("put").with(eq(parent), eq(k), eq(v));

      // create parent and child with data
      // new nodes being created, will result in loading them from the cache loader first
      mockCacheLoader.expects(once()).method("get").with(eq(parent));

      cache.put(parent, k, v);
      assertDataLoaded(parent);
      // evict the parent
      cache.evict(parent, false);
      assertNull(cache.peek(parent, false));

      mockCacheLoader.expects(once()).method("exists").with(eq(parent)).will(returnValue(true));
      // now get node.
      Node n = cache.getRoot().getChild(parent);
      assertNotNull(n);
      assertDataNotLoaded(parent);

      // should not load node but should change isDataLoaded to true
      // will trigger a removedata() though
      mockCacheLoader.expects(once()).method("removeData").with(eq(parent));
      n.clearData();

      assertDataLoaded(parent);
   }

   public void testDontLoadDataWhenReplacingNode()
   {
      mockCacheLoader.expects(never()).method("get").withAnyArguments();
      mockCacheLoader.expects(never()).method("getChildrenNames").withAnyArguments();
      mockCacheLoader.expects(never()).method("exists").withAnyArguments();
      mockCacheLoader.expects(never()).method("remove").withAnyArguments();
      mockCacheLoader.expects(never()).method("removeData").withAnyArguments();
      mockCacheLoader.expects(never()).method("put").withAnyArguments();

      mockCacheLoader.expects(once()).method("put").with(eq(parent), eq(k), eq(v));

      // create parent and child with data
      // new nodes being created, will result in loading them from the cache loader first
      mockCacheLoader.expects(once()).method("get").with(eq(parent));

      cache.put(parent, k, v);
      assertTrue(cache.peek(parent, false).isDataLoaded());
      // evict the parent
      cache.evict(parent, false);
      assertNull(cache.peek(parent, false));

      mockCacheLoader.expects(once()).method("exists").with(eq(parent)).will(returnValue(true));
      // now get node.
      Node n = cache.getRoot().getChild(parent);
      assertNotNull(n);
      assertDataNotLoaded(parent);

      // should not load node but should change isDataLoaded to true
      // will trigger a put() though
      // for the moment this does a get as well - which while unnecessary is the best we can do for now until we bring in
      // an AOP framework to work on nodes directly.
      mockCacheLoader.expects(once()).method("removeData").with(eq(parent));
      mockCacheLoader.expects(once()).method("put").with(eq(parent), eq(Collections.singletonMap("hello", "world")));
      n.replaceAll(Collections.singletonMap("hello", "world"));

      assertDataLoaded(parent);
   }

   public void testLazyLoadDataWhenWorkingWithNode()
   {
      mockCacheLoader.expects(never()).method("get").withAnyArguments();
      mockCacheLoader.expects(never()).method("getChildrenNames").withAnyArguments();
      mockCacheLoader.expects(never()).method("exists").withAnyArguments();
      mockCacheLoader.expects(never()).method("remove").withAnyArguments();
      mockCacheLoader.expects(never()).method("removeData").withAnyArguments();
      mockCacheLoader.expects(never()).method("put").withAnyArguments();

      mockCacheLoader.expects(once()).method("put").with(eq(parent), eq(k), eq(v));

      // create parent and child with data
      // new nodes being created, will result in loading them from the cache loader first
      mockCacheLoader.expects(once()).method("get").with(eq(parent));

      cache.put(parent, k, v);
      assertDataLoaded(parent);
      // evict the parent
      cache.evict(parent, false);
      assertNull(cache.peek(parent, false));

      mockCacheLoader.expects(once()).method("exists").with(eq(parent)).will(returnValue(true));
      // now get node.
      Node n = cache.getRoot().getChild(parent);
      assertNotNull(n);
      assertDataNotLoaded(parent);

      // will trigger a load
      mockCacheLoader.expects(once()).method("get").with(eq(parent)).will(returnValue(Collections.singletonMap(k, v)));
      assertEquals(v, n.get(k));
      // should change isDataLoaded to true
      assertDataLoaded(parent);
   }

   public void testDontLoadWhenKeyInMemory()
   {
      mockCacheLoader.expects(never()).method("get").withAnyArguments();
      mockCacheLoader.expects(never()).method("getChildrenNames").withAnyArguments();
      mockCacheLoader.expects(never()).method("exists").withAnyArguments();
      mockCacheLoader.expects(never()).method("remove").withAnyArguments();
      mockCacheLoader.expects(never()).method("removeData").withAnyArguments();
      mockCacheLoader.expects(never()).method("put").withAnyArguments();

      mockCacheLoader.expects(once()).method("get").with(eq(parent));
      mockCacheLoader.expects(once()).method("put").with(eq(parent), eq(k), eq(v));

      cache.put(parent, k, v);
      assertDataLoaded(parent);

      // now evict
      cache.evict(parent, false);

      // now to a put(Map)
      Map m = new HashMap();
      m.put("k2", "v2");

      // should not load
      mockCacheLoader.expects(once()).method("put").with(eq(parent), eq(m));
      cache.put(parent, m);

      assertDataNotLoaded(parent);

      // now a get for an existing key should not trigger a load!
      assertEquals("v2", cache.get(parent, "k2"));

      assertDataNotLoaded(parent);

      // but going a get for a nonexistent key should!
      mockCacheLoader.expects(once()).method("get").with(eq(parent)).will(returnValue(Collections.singletonMap(k, v)));
      assertEquals(v, cache.get(parent, k));

      // should not have overwritten in-memory data
      assertEquals("v2", cache.get(parent, "k2"));
   }
}