/*
 *
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */

package org.jboss.cache.marshall;


import junit.framework.Test;
import junit.framework.TestSuite;
import org.jboss.cache.CacheSPI;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.Region;
import org.jboss.cache.config.Configuration.CacheMode;
import org.jboss.cache.factories.UnitTestCacheConfigurationFactory;
import org.jboss.cache.marshall.data.Address;
import org.jboss.cache.marshall.data.Person;
import org.jboss.cache.misc.TestingUtil;
import org.jboss.cache.transaction.DummyTransactionManager;

import javax.transaction.NotSupportedException;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import java.lang.reflect.Method;
import java.util.HashMap;

/**
 * Test case for marshalling using Sync mode.
 *
 * @author Ben Wang
 * @version $Revision: 1.18 $
 */
public class SyncReplTest extends RegionBasedMarshallingTestBase
{
   CacheSPI cache1, cache2;
   String props = null;
   Person ben_;
   Address addr_;
   Throwable ex_;
   private Fqn aop = Fqn.fromString("/aop");

   public void setUp() throws Exception
   {
      super.setUp();

      log("creating cache1");
      cache1 = createCache("TestCache");

      log("creating cache2");

      cache2 = createCache("TestCache");
      addr_ = new Address();
      addr_.setCity("San Jose");
      ben_ = new Person();
      ben_.setName("Ben");
      ben_.setAddress(addr_);

      // Pause to give caches time to see each other
      TestingUtil.blockUntilViewsReceived(new CacheSPI[]{cache1, cache2}, 60000);
   }

   private CacheSPI createCache(String name) throws Exception
   {
      CacheSPI cache = (CacheSPI) DefaultCacheFactory.getInstance().createCache(UnitTestCacheConfigurationFactory.createConfiguration(CacheMode.REPL_SYNC), false);
      cache.getConfiguration().setClusterName(name);
      // Use marshaller
      cache.getConfiguration().setUseRegionBasedMarshalling(true);
      cache.create();
      cache.start();
      return cache;
   }

   public void tearDown() throws Exception
   {
      super.tearDown();
      cache1.removeNode(Fqn.ROOT);
      if (cache1 != null)
      {
         log("stopping cache1");
         cache1.stop();
      }

      if (cache2 != null)
      {
         log("stopping cache2");
         cache2.stop();
      }
   }

   public void testPlainPut() throws Exception
   {
      cache1.put(aop, "person", ben_);
      Person ben2 = (Person) cache2.get(aop, "person");
      assertNotNull("Person from 2nd cache should not be null ", ben2);
      assertEquals(ben_.toString(), ben2.toString());
   }

   public void testCCE() throws Exception
   {
      ClassLoader cl = getClassLoader();
      Region r1 = cache1.getRegion(aop, false) == null ? cache1.getRegion(aop, true) : cache1.getRegion(aop, false);
      r1.registerContextClassLoader(cl);

      cl = getClassLoader();
      Region r2 = cache2.getRegion(aop, false) == null ? cache2.getRegion(aop, true) : cache2.getRegion(aop, false);
      r2.registerContextClassLoader(cl);

      cache1.put(aop, "person", ben_);
      try
      {
         Person ben2 = (Person) cache2.get(aop, "person");
      }
      catch (ClassCastException ex)
      {
         // That's ok.
         return;
      }
      fail("Should have thrown an exception");
   }

   public void testPut() throws Exception
   {
      ClassLoader cla = getClassLoader();
      Region r1 = cache1.getRegion(aop, false) == null ? cache1.getRegion(aop, true) : cache1.getRegion(aop, false);
      r1.registerContextClassLoader(cla);
      ClassLoader clb = getClassLoader();
      Region r2 = cache2.getRegion(aop, false) == null ? cache2.getRegion(aop, true) : cache2.getRegion(aop, false);
      r2.registerContextClassLoader(clb);

      cache1.put(aop, "person", ben_);

      Object ben2;
      // Can't cast it to Person. CCE will resutl.
      ben2 = cache2.get(aop, "person");
      assertEquals(ben_.toString(), ben2.toString());
   }

   public void testCLSet() throws Exception
   {
      ClassLoader cla = getClassLoader();
      Region r1 = cache1.getRegion(aop, false) == null ? cache1.getRegion(aop, true) : cache1.getRegion(aop, false);
      r1.registerContextClassLoader(cla);
      ClassLoader clb = getClassLoader();
      Region r2 = cache2.getRegion(aop, false) == null ? cache2.getRegion(aop, true) : cache2.getRegion(aop, false);
      r2.registerContextClassLoader(clb);

      cache1.put(aop, "person", ben_);

      Object ben2;
      // Can't cast it to Person. CCE will resutl.
      ben2 = cache2.get(aop, "person");
      assertEquals(ben_.toString(), ben2.toString());

      Class claz = clb.loadClass(ADDRESS_CLASSNAME);
      Object add = claz.newInstance();
      {
         Class[] types = {String.class};
         Method setValue = claz.getMethod("setCity", types);
         Object[] margs = {"Sunnyvale"};
         setValue.invoke(add, margs);
      }

      {
         Class clasz1 = clb.loadClass(PERSON_CLASSNAME);
         Class[] types = {claz};
         Method setValue = clasz1.getMethod("setAddress", types);
         Object[] margs = {add};
         setValue.invoke(ben2, margs);
      }
   }

   /**
    * Test replication with classloaders.
    *
    * @throws Exception
    */
   public void testCLSet2() throws Exception
   {
      ClassLoader cla = getClassLoader();
      Region r1 = cache1.getRegion(aop, false) == null ? cache1.getRegion(aop, true) : cache1.getRegion(aop, false);
      r1.registerContextClassLoader(cla);
      ClassLoader clb = getClassLoader();
      Region r2 = cache2.getRegion(aop, false) == null ? cache2.getRegion(aop, true) : cache2.getRegion(aop, false);
      r2.registerContextClassLoader(clb);

      cache1.put(aop, "person", ben_);

      Object ben2;
      // Can't cast it to Person. CCE will resutl.
      ben2 = cache2.get(aop, "person");
      assertEquals(ben_.toString(), ben2.toString());

      Class claz = clb.loadClass(ADDRESS_CLASSNAME);
      Object add = claz.newInstance();
      {
         Class[] types = {String.class};
         Method setValue = claz.getMethod("setCity", types);
         Object[] margs = {"Sunnyvale"};
         setValue.invoke(add, margs);
      }

      {
         Class clasz1 = clb.loadClass(PERSON_CLASSNAME);
         Class[] types = {claz};
         Method setValue = clasz1.getMethod("setAddress", types);
         Object[] margs = {add};
         setValue.invoke(ben2, margs);
      }

      // Set it back to the cache
      // Can't cast it to Person. CCE will resutl.
      cache2.put(aop, "person", ben2);
      Object ben3 = cache1.get(aop, "person");
      assertEquals(ben2.toString(), ben3.toString());

   }

   public void testPuts() throws Exception
   {
      ClassLoader cl = getClassLoader();
      Region r1 = cache1.getRegion(aop, false) == null ? cache1.getRegion(aop, true) : cache1.getRegion(aop, false);
      r1.registerContextClassLoader(cl);

      // Create an empty Person loaded by this classloader
      Object scopedBen1 = getPersonFromClassloader(cl);

      cl = getClassLoader();
      Region r2 = cache2.getRegion(aop, false) == null ? cache2.getRegion(aop, true) : cache2.getRegion(aop, false);
      r2.registerContextClassLoader(cl);

      // Create another empty Person loaded by this classloader
      Object scopedBen2 = getPersonFromClassloader(cl);

      cache1.put(Fqn.fromString("/aop/1"), "person", ben_);
      cache1.put(Fqn.fromString("/aop/2"), "person", scopedBen1);

      Object ben2 = null;
      try
      {
         // Can't cast it to Person. CCE will resutl.
         ben2 = cache2.get(Fqn.fromString("/aop/1"), "person");
         assertEquals(ben_.toString(), ben2.toString());

         ben2 = cache2.get(Fqn.fromString("/aop/2"), "person");
         assertFalse("cache2 deserialized with scoped classloader", ben2 instanceof Person);
         assertFalse("cache2 deserialized with cache2 classloader", scopedBen1.equals(ben2));
         assertEquals("scopedBen deserialized properly", scopedBen2, ben2);
      }
      catch (Exception ex)
      {
         fail("Test fails with exception " + ex);
      }

   }

   public void testMethodCall() throws Exception
   {
      ClassLoader cl = getClassLoader();
      Region r1 = cache1.getRegion(aop, false) == null ? cache1.getRegion(aop, true) : cache1.getRegion(aop, false);
      r1.registerContextClassLoader(cl);
      cl = getClassLoader();
      Region r2 = cache2.getRegion(aop, false) == null ? cache2.getRegion(aop, true) : cache2.getRegion(aop, false);
      r2.registerContextClassLoader(cl);

      cache1.put(Fqn.fromString("/aop/1"), "person", ben_);
      cache1.remove(Fqn.fromString("/aop/1"), "person");
      HashMap map = new HashMap();
      map.put("1", "1");
      map.put("2", "2");
      cache1.put(Fqn.fromString("/aop/2"), map);
      cache1.removeNode(Fqn.fromString("/aop/2"));

      TestingUtil.sleepThread(1000);
   }

   public void testTxMethodCall() throws Exception
   {
      ClassLoader cl = getClassLoader();
      Region r1 = cache1.getRegion(aop, false) == null ? cache1.getRegion(aop, true) : cache1.getRegion(aop, false);
      r1.registerContextClassLoader(cl);
      cl = getClassLoader();
      Region r2 = cache2.getRegion(aop, false) == null ? cache2.getRegion(aop, true) : cache2.getRegion(aop, false);
      r2.registerContextClassLoader(cl);

      Transaction tx = beginTransaction();
      cache1.put(Fqn.fromString("/aop/1"), "person", ben_);
      cache1.remove(Fqn.fromString("/aop/1"), "person");
      HashMap map = new HashMap();
      map.put("1", "1");
      map.put("2", "2");
      cache1.put(Fqn.fromString("/aop/2"), map);
      cache1.removeNode(Fqn.fromString("/aop/2"));
      tx.commit();

      TestingUtil.sleepThread(1000);
   }

   public void testTxPut() throws Exception
   {
      Transaction tx = beginTransaction();
      cache1.put(aop, "person", ben_);
      cache1.put(aop, "person1", ben_);
      cache1.removeNode(aop);
      cache1.put(aop, "person", ben_);
      tx.commit();
      Person ben2 = (Person) cache2.get(aop, "person");
      assertNotNull("Person from 2nd cache should not be null ", ben2);
      assertEquals(ben_.toString(), ben2.toString());
   }

   public void testTxRollback() throws Exception
   {
      Transaction tx = beginTransaction();
      cache1.put(aop, "person", ben_);
      cache1.put(aop, "person1", ben_);
      tx.rollback();
      Person ben2 = (Person) cache2.get(aop, "person");
      assertNull("Person from 2nd cache should be null ", ben2);
   }

   public void testTxCLSet2() throws Exception
   {
      ClassLoader cla = getClassLoader();
      Region r1 = cache1.getRegion(aop, false) == null ? cache1.getRegion(aop, true) : cache1.getRegion(aop, false);
      r1.registerContextClassLoader(cla);
      ClassLoader clb = getClassLoader();
      Region r2 = cache2.getRegion(aop, false) == null ? cache2.getRegion(aop, true) : cache2.getRegion(aop, false);
      r2.registerContextClassLoader(clb);

      Transaction tx = beginTransaction();
      cache1.put(aop, "person", ben_);
      tx.commit();

      Object ben2;
      // Can't cast it to Person. CCE will resutl.
      ben2 = cache2.get(aop, "person");
      assertEquals(ben_.toString(), ben2.toString());

      Class claz = clb.loadClass(ADDRESS_CLASSNAME);
      Object add = claz.newInstance();
      {
         Class[] types = {String.class};
         Method setValue = claz.getMethod("setCity", types);
         Object[] margs = {"Sunnyvale"};
         setValue.invoke(add, margs);
      }

      {
         Class clasz1 = clb.loadClass(PERSON_CLASSNAME);
         Class[] types = {claz};
         Method setValue = clasz1.getMethod("setAddress", types);
         Object[] margs = {add};
         setValue.invoke(ben2, margs);
      }

      // Set it back to the cache
      // Can't cast it to Person. CCE will resutl.
      cache2.put(aop, "person", ben2);
      Object ben3 = cache1.get(aop, "person");
      assertEquals(ben2.toString(), ben3.toString());
   }

   public void testStateTransfer() throws Exception
   {
      // Need to test out if app is not registered with beforehand??
   }

   public void testCustomFqn() throws Exception
   {
      FooClassLoader cl1 = new FooClassLoader(Thread.currentThread().getContextClassLoader());
      Region r1 = cache1.getRegion(aop, false) == null ? cache1.getRegion(aop, true) : cache1.getRegion(aop, false);
      r1.registerContextClassLoader(cl1);
      FooClassLoader cl2 = new FooClassLoader(Thread.currentThread().getContextClassLoader());
      Region r2 = cache2.getRegion(aop, false) == null ? cache2.getRegion(aop, true) : cache2.getRegion(aop, false);
      r2.registerContextClassLoader(cl2);

      Class clazz = cl1.loadFoo();
      Object custom1 = clazz.newInstance();

      clazz = cl2.loadFoo();
      Object custom2 = clazz.newInstance();

      Fqn base = aop;
      cache1.put(new Fqn(base, custom1), "key", "value");

      try
      {
         Object val = cache2.get(new Fqn(base, custom2), "key");
         assertEquals("value", val);
      }
      catch (Exception ex)
      {
         fail("Test fails with exception " + ex);
      }
   }

   Transaction beginTransaction() throws SystemException, NotSupportedException
   {
      DummyTransactionManager mgr = DummyTransactionManager.getInstance();
      mgr.begin();
      return mgr.getTransaction();
   }

   protected Object getPersonFromClassloader(ClassLoader cl) throws Exception
   {
      Class clazz = cl.loadClass(PERSON_CLASSNAME);
      return clazz.newInstance();
   }

   void log(String msg)
   {
      System.out.println("-- [" + Thread.currentThread() + "]: " + msg);
   }


   public static Test suite()
   {
      return new TestSuite(SyncReplTest.class);
   }

   public static void main(String[] args)
   {
      junit.textui.TestRunner.run(suite());
   }

}
