/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */

package org.jboss.cache.passivation;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import org.jboss.cache.CacheImpl;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.factories.XmlConfigurationParser;
import org.jboss.cache.loader.DummyInMemoryCacheLoader;

/**
 * Tests cache behavior in the presence of concurrent passivation.
 *
 * @author Brian Stansberry
 * @version $Revision: 1.15 $
 */
public class ConcurrentPassivationTest extends TestCase
{
   private CacheImpl cache_;
   private int wakeupIntervalMillis_ = 0;

   public ConcurrentPassivationTest(String s)
   {
      super(s);
   }

   public void setUp() throws Exception
   {
      super.setUp();
      initCaches();
      wakeupIntervalMillis_ = cache_.getConfiguration().getEvictionConfig().getWakeupIntervalSeconds() * 1000;
      if (wakeupIntervalMillis_ < 0)
      {
         fail("testEviction(): eviction thread wake up interval is illegal " + wakeupIntervalMillis_);
      }

   }

   void initCaches() throws Exception
   {
      cache_ = (CacheImpl) DefaultCacheFactory.getInstance().createCache(false);
      cache_.setConfiguration(new XmlConfigurationParser().parseFile("META-INF/local-passivation-service.xml"));// read in generic local xml
      cache_.getConfiguration().setTransactionManagerLookupClass("org.jboss.cache.transaction.DummyTransactionManagerLookup");
      cache_.getConfiguration().getCacheLoaderConfig().getFirstCacheLoaderConfig().setClassName(DummyInMemoryCacheLoader.class.getName());
      cache_.start();
   }

   public void tearDown() throws Exception
   {
      super.tearDown();
      cache_.stop();
      cache_ = null;
   }

   public void testConcurrentPassivation() throws Exception
   {
      Fqn base = Fqn.fromString("/org/jboss/test/data/concurrent/passivation");

      // Create a bunch of nodes; more than the /org/jboss/test/data
      // region's maxNodes so we know eviction will kick in
      for (int i = 0; i < 35000; i++)
      {
         cache_.put(new Fqn(base, i / 100), i, "value");
      }

      // Loop for long enough to have 5 runs of the eviction thread
      long loopDone = System.currentTimeMillis() + (5 * wakeupIntervalMillis_);

      System.out.println("Initialised; Loop for " + (5 * wakeupIntervalMillis_) + " millis");

      while (System.currentTimeMillis() < loopDone)
      {
         // If any get returns null, that's a failure
         for (int i = 0; i < 35000; i++)
         {
            Fqn fqn = new Fqn(base, i / 100);
            assertNotNull("Get on Fqn " + fqn + " returned null", cache_.get(fqn));
         }
      }
   }

   public static Test suite()
   {
      return new TestSuite(org.jboss.cache.passivation.ConcurrentPassivationTest.class);
   }

   public static void main(String[] args)
   {
      junit.textui.TestRunner.run(org.jboss.cache.passivation.ConcurrentPassivationTest.suite());
   }
}
