/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */

package org.jboss.cache.passivation;

import junit.framework.TestCase;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.CacheImpl;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.factories.XmlConfigurationParser;
import org.jboss.cache.loader.DummyInMemoryCacheLoader;
import org.jboss.cache.misc.TestingUtil;
import org.jboss.cache.notifications.annotation.CacheListener;
import org.jboss.cache.notifications.annotation.NodeActivated;
import org.jboss.cache.notifications.annotation.NodeLoaded;
import org.jboss.cache.notifications.annotation.NodePassivated;
import org.jboss.cache.notifications.event.Event;
import org.jboss.cache.notifications.event.NodeEvent;

/**
 * @author Ben Wang, Feb 11, 2004
 */
public class LocalPassivationIntegrationTest extends TestCase
{
   CacheImpl cache_;
   protected final static Log log = LogFactory.getLog(LocalPassivationIntegrationTest.class);
   int wakeupIntervalMillis_ = 0;
   PassivationListener listener_;
   private static final int LISTENER_WAIT_TIME = 200; // needed since notifications are delivered asynchronously

   public LocalPassivationIntegrationTest(String s)
   {
      super(s);
      listener_ = new PassivationListener();
   }

   public void setUp() throws Exception
   {
      super.setUp();
      cache_ = (CacheImpl) DefaultCacheFactory.getInstance().createCache(false);
      initCaches(cache_);
      cache_.getConfiguration().setUseRegionBasedMarshalling(true);

      cache_.start();

      cache_.getNotifier().addCacheListener(listener_);
      listener_.resetCounter();

      wakeupIntervalMillis_ = cache_.getConfiguration().getEvictionConfig().getWakeupIntervalSeconds() * 1000;
      log("wakeupInterval is " + wakeupIntervalMillis_);
      if (wakeupIntervalMillis_ <= 0)
      {
         fail("testEviction(): eviction thread wake up interval is illegal " + wakeupIntervalMillis_);
      }
   }

   void initCaches(CacheImpl cache) throws Exception
   {
      cache.setConfiguration(new XmlConfigurationParser().parseFile("META-INF/local-passivation-service.xml"));// read in generic local xml
      cache.getConfiguration().setTransactionManagerLookupClass("org.jboss.cache.transaction.DummyTransactionManagerLookup");
      cache.getConfiguration().getCacheLoaderConfig().getFirstCacheLoaderConfig().setClassName(DummyInMemoryCacheLoader.class.getName());
   }

   public void tearDown() throws Exception
   {
      super.tearDown();
      cache_.stop();
   }

   /**
    */
   public void testActivationEvent() throws Exception
   {
      String rootStr = "/org/jboss/test/data/";
      String str = rootStr + "0";
      cache_.remove("/");
      listener_.resetCounter();

      cache_.put(str, str, str);

      TestingUtil.sleepThread(20000);
      assertFalse("UnversionedNode should not exist", cache_.exists(str, str));
      String val = (String) cache_.get(str, str);
      assertNotNull("DataNode should be activated ", val);
      TestingUtil.sleepThread(LISTENER_WAIT_TIME);
      assertEquals("Eviction counter ", 1, listener_.getCounter());
   }

   void log(String msg)
   {
      System.out.println("-- " + msg);
   }

   @CacheListener
   public class PassivationListener
   {
      int counter = 0;
      int loadedCounter = 0;

      public int getCounter()
      {
         return counter;
      }

      public void resetCounter()
      {
         counter = 0;
         loadedCounter = 0;
      }

      @NodeActivated
      public void nodeActivated(NodeEvent ne)
      {
         if (!ne.isPre())
         {
            counter++;
            System.out.println("nodeActivate(): counter: " + counter);
         }
      }

      @NodePassivated
      public void nodePassivated(NodeEvent ne)
      {
         if (ne.isPre())
         {
            System.out.println("nodePassivate(): " + ne.getFqn());
         }
      }

      @NodeLoaded
      public void nodeLoaded(Event e)
      {
         if (!e.isPre())
         {
            loadedCounter++;
         }
      }

   }
}
