/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */

package org.jboss.cache.pojo.rollback;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.aop.proxy.ClassProxy;
import org.jboss.cache.pojo.PojoCache;
import org.jboss.cache.pojo.PojoCacheFactory;
import org.jboss.cache.pojo.test.Person;
import org.jboss.cache.transaction.DummyTransactionManager;

import javax.transaction.TransactionManager;
import java.util.HashMap;
import java.util.Map;

/**
 * Additional basic tests
 *
 * @author Ben Wang
 */

public class MapTxUndoTest extends TestCase
{
   Log log_ = LogFactory.getLog(MapTxUndoTest.class);
   PojoCache cache_;
   TransactionManager tx_mgr;

   public MapTxUndoTest(String name)
   {
      super(name);
   }

   protected void setUp() throws Exception
   {
      super.setUp();
      log_.info("setUp() ....");
      String configFile = "META-INF/local-service.xml";
      boolean toStart = false;
      cache_ = PojoCacheFactory.createCache(configFile, toStart);
      cache_.start();
      tx_mgr = DummyTransactionManager.getInstance();

   }

   protected void tearDown() throws Exception
   {
      super.tearDown();
      cache_.stop();
   }

//   public void testDummy() {}

   public void testSimple() throws Exception
   {
      HashMap map = new HashMap();
      map.put("1", "test1");

      tx_mgr.begin();
      cache_.attach("/a", map);
      tx_mgr.getTransaction().rollback();
      assertFalse("Should not have cache interceptor ", isProxy(map));

      cache_.attach("/a", map);
   }

   public void testSimpleTxWithRollback1() throws Exception
   {
      log_.info("testSimpleTxWithRollback1() ....");
      Person test = new Person();
      test.setName("Ben");
      test.setAge(10);
      HashMap map = new HashMap();
      map.put("1", "English");
      test.setHobbies(map);

      tx_mgr.begin();
      cache_.attach("/a", test);
      tx_mgr.getTransaction().rollback();
      assertFalse("Should not have cache interceptor ", isProxy(test.getHobbies()));

      cache_.attach("/a", test);
   }

   private boolean isProxy(Object pojo)
   {
      if (pojo instanceof ClassProxy) return true;
      return false;
   }

   public void testSimpleTxWithRollback2() throws Exception
   {
      log_.info("testSimpleTxWithRollback1() ....");
      Person test = new Person();
      test.setName("Ben");
      test.setAge(10);
      HashMap map = new HashMap();
      map.put("1", "English");
      test.setHobbies(map);

      cache_.attach("/a", test);

      tx_mgr.begin();
      cache_.detach("/a");
      tx_mgr.getTransaction().rollback();

      assertTrue("Should still have cache interceptor ", isProxy(test.getHobbies()));
      cache_.detach("/a");
   }

   public void testSimpleTxWithRollback3() throws Exception
   {
      log_.info("testSimpleTxWithRollback1() ....");
      Person test = new Person();
      test.setName("Ben");
      test.setAge(10);
      HashMap map = new HashMap();
      map.put("1", "English");
      test.setHobbies(map);
      tx_mgr.begin();
      cache_.attach("/a", test);
      cache_.detach("/a");
      tx_mgr.getTransaction().rollback();

      assertFalse("Should not have cache interceptor ", isProxy(test.getHobbies()));
   }

   /**
    * Contributed by Niztan Niv
    *
    * @throws Exception
    */
   public void testNestedMapWithRollback() throws Exception
   {
      // create cached data objects
      Map obj1 = new HashMap();
      obj1.put("id", "1");
      cache_.attach("objs/1", obj1);
      obj1 = (Map) cache_.find("objs/1");

      Map obj2 = new HashMap();
      obj2.put("id", "2");
      cache_.attach("objs/2", obj2);
      obj2 = (Map) cache_.find("objs/2");

      // create cached collection of data objects
      Map indexMap = new HashMap();
      cache_.attach("objsIndex", indexMap);
      indexMap = (Map) cache_.find("objsIndex");

      // initialize collection by adding a data object
      final String KEY = "KEY";
      indexMap.put(KEY, obj1);

      Object beforeModify = indexMap.get(KEY);
      System.out.println("beforeModify: " + beforeModify + ", data object id: " + ((Map) beforeModify).get("id"));

      // modify the collection by replacing the first data object with the second
      // and then roll-back the transaction
      tx_mgr.begin();
      indexMap.put(KEY, obj2);
      tx_mgr.rollback();

      Object afterRollback = indexMap.get(KEY);
      System.out.println("afterRollback: " + afterRollback + ", data object id: " + ((Map) afterRollback).get("id"));

      // check if state of collection was restored
      assertEquals(beforeModify, afterRollback);
   }

   public void testPlainMapRollback() throws Exception
   {
      // create cached data objects
      Map obj1 = new HashMap();
      obj1.put("id", "1");
      cache_.attach("objs/1", obj1);
      obj1 = (Map) cache_.find("objs/1");

      tx_mgr.begin();
      obj1.put("id", "newId");
      tx_mgr.rollback();

      assertEquals("1", obj1.get("id"));
   }

   public static Test suite() throws Exception
   {
      return new TestSuite(MapTxUndoTest.class);
   }


   public static void main(String[] args) throws Exception
   {
      junit.textui.TestRunner.run(MapTxUndoTest.suite());
   }

}
