/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.cache;

import jasco.options.Options;
import jasco.runtime.CombinedHooks;
import jasco.runtime.CutpointAfterComparator;
import jasco.runtime.CutpointBeforeComparator;
import jasco.runtime.CutpointReplaceComparator;
import jasco.runtime.DoNotCache;
import jasco.runtime.JascoMethod;
import jasco.runtime.MethodJoinpoint;
import jasco.runtime.VectorSorter;
import jasco.runtime.aspect.HookContextProperty;
import jasco.runtime.aspect.HookHelper;
import jasco.runtime.aspect.IHook;
import jasco.runtime.aspect.PCutpointConstructorRegular;
import jasco.runtime.aspect.StatefulAspect;
import jasco.runtime.aspect.factory.IPredefinedAspectFactory;
import jasco.runtime.connector.DynamicHookElement;
import jasco.runtime.connector.HookElement;
import jasco.util.generators.JavaGenerator;
import jasco.util.generators.MethodGenerator;

import java.io.File;
import java.io.FileWriter;
import java.lang.reflect.Modifier;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import javassist.ClassPool;
import javassist.CtClass;
import javassist.CtField;
import javassist.CtMethod;
import javassist.CtNewConstructor;
import javassist.CtNewMethod;

  

public class CombinedHooksFactory  {

	private Hashtable codeCache = new Hashtable();
	private static long classId = 0;
	protected static final ClassPool classPool = ClassPool.getDefault();
	private static CombinedHooksFactory factory=new CombinedHooksFactory();
	
	private boolean clearLinkedHooks=false;

	
	public static final int NO_CONTEXT = 10;
	public static final int STATIC_CONTEXT = 11;
	public static final int DYNAMIC_CONTEXT = 12;
	
	private int currentHookId = 0; 
	private int currentCombStratId = 0;
	
	public static final CombinedHooks EMPTY_CONBINED_HOOKS = new CombinedHooks()  
	{
	public Object invoke(JascoMethod method) throws Exception  {return method.invokeOriginalJAsCoMethod();}
	public void init(Vector v)  {}};
	
	static  { 
		classPool.appendSystemPath();
	}
	protected static long getNextClassId()  {
	
		return ++classId;
	}
	
	protected void setClearLinkedHooks() {
		
		clearLinkedHooks=true;
	}
	
	public static CombinedHooksFactory getDefault()  {
	
		return factory;
	}
	
	protected String getHookName(HookElement hook)  {
		
			return "hook"+hook.getId();
		}
	
	protected String accessField(String field)  {
	
		return field;
	}
	
	protected String accessFieldHook(HookElement hook)  {
	
		return accessFieldHook(hook,false);
	}
	
	protected String accessFieldHook(HookElement hook, boolean forConstructor)  {
		
		if(forConstructor)
			return accessField(getHookName(hook));
		else return getHookName(hook); 
	}
	
	protected CombinedHooks checkForPredef(Vector hooks)  {
		
		if(hooks.size()!=1)
			return null;
		
		HookElement elm = (HookElement) hooks.elementAt(0);
		
		if(elm.getCutpoint() instanceof StatefulAspect || elm.getCutpoint() instanceof DoNotCache || elm instanceof DynamicHookElement)
			return null;
		
		boolean hasAfterThrowing = HookHelper.definesAdvice(elm.getCutpoint(),IHook.AFTER_THROWING);
		boolean hasAfterReturning = HookHelper.definesAdvice(elm.getCutpoint(),IHook.AFTER_RETURNING);
		if(hasAfterThrowing||hasAfterReturning) //this is not ever predefined -> return null
			return null;
		
		boolean hasAroundThrowing = HookHelper.definesAdvice(elm.getCutpoint(),IHook.AROUND_THROWING);
		boolean hasAroundReturning = HookHelper.definesAdvice(elm.getCutpoint(),IHook.AROUND_RETURNING);
		if(hasAroundThrowing||hasAroundReturning) //this is not ever predefined -> return null
			return null;
		
		if(this.shouldCheckCombStrats(hooks))
			return null;
		
		boolean hasBefore = elm.getBefore().getExecution();	
		boolean hasReplace = elm.getReplace().getExecution();
		boolean hasAfter = elm.getAfter().getExecution();
		
		if(hasBefore && hasReplace && hasAfter)
			return new OneReplaceBeforeAfter(hooks);
		else if(hasBefore && !hasReplace && hasAfter)
			return new OneBeforeAfter(hooks);
		else if(hasBefore && !hasReplace && !hasAfter)
			return new OneBefore(hooks);
		else if(!hasBefore && !hasReplace && hasAfter)
			return new OneAfter(hooks);
		else if(!hasBefore && hasReplace && !hasAfter)
			return new OneReplace(hooks);		
		else return null;	
	}
	
	/**
	 * @param hooks
	 * @return
	 */
	private boolean hasDynamicHookElements(Vector hooks) {
		Iterator  i = hooks.iterator();
		while(i.hasNext()) {
			if(i.next() instanceof DynamicHookElement)
				return true;
		}
		return false;
	}

	protected void addFieldImpl(String f, CtClass clz) throws Exception {
		
		CtField field = CtField.make(f,clz);
		clz.addField(field);
	}
	
	protected void addField(String f, CtClass clz) throws Exception {
		
		
		addFieldImpl("private final "+f,clz); 
	}
	
	private static int logid;
	
	protected void generateConstructorCode(String cname, java.util.Vector hooks, JascoMethod originalMethod, CtClass fieldClass, CachedAspectRep cachedAspectRep) throws java.lang.Exception {
				
		
		CtClass cutp = classPool.get("jasco.runtime.aspect.IHook");
		CtClass hash = classPool.get("java.util.Hashtable");
		CtClass cutparray = classPool.get("jasco.runtime.aspect.IHook[]");

		
		MethodGenerator constr = new MethodGenerator(cname,"");
		constr.addVariable("java.util.Vector","hooks");
		constr.addVariable("jasco.runtime.MethodJoinpoint","jp");
		constr.addModifier(Modifier.PUBLIC);
		
		String constrCode = "";

		for (int j = 0; j < hooks.size() ; ++j) {
			currentHookId++;
			HookElement cutpointElement = (HookElement)hooks.elementAt(j);
			cutpointElement.setId(currentHookId);
			
			IHook cp = cutpointElement.getCutpoint();
			
			addField(cp.getClass().getName()+" "+getHookName(cutpointElement)+";", fieldClass);
			constrCode+="jasco.runtime.aspect.IHook cp = "+"((jasco.runtime.connector.HookElement)hooks.elementAt("+j+")).getCutpoint(jp); //"+cp.getClass().getName()+JavaGenerator.NEWLINE;
			constrCode+=accessFieldHook(cutpointElement,true)+"= ("+cp.getClass().getName()+") cp;"+JavaGenerator.NEWLINE;
			
			if (cutpointElement.getCutpoint()._Jasco_getConstructorRep().doNotCache())  {
				addField("java.util.Hashtable "+getHookName(cutpointElement)+"hashtable;", fieldClass);
				
				int index = cutpointElement.getCutpoint()._Jasco_getConnector().indexOf(cutpointElement);
				constrCode+=accessFieldHook(cutpointElement,true)+"hashtable = "+" "+accessFieldHook(cutpointElement,true)+"._Jasco_getConnector().getMethodHashtable("+index+");"+JavaGenerator.NEWLINE;
			}
		}
		
		constrCode+=generateCombStratInit(hooks,fieldClass);
		
		constrCode+=generateFactoryInit(hooks,fieldClass);
		
		constr.setImplementation(constrCode);
		
		cachedAspectRep.setConstructorCode(constr);
	}
	
	/**
	 * @param hooks
	 * @param fieldClass
	 * @return
	 * @throws Exception
	 */
	private String generateFactoryInit(Vector hooks, CtClass fieldClass) throws Exception {
		String code="";
		for (int j = 0; j < hooks.size() ; ++j) {
			HookElement cutpointElement = (HookElement)hooks.elementAt(j);
			if(cutpointElement instanceof DynamicHookElement)
				code+= generateFactoryInit((DynamicHookElement)cutpointElement, fieldClass,j);
		}
		return code;
	}
	
	protected String getHookElementName(HookElement el) {
		
		return "hookElement"+el.getId();
	}
	
	protected String getFactoryName(HookElement el) {
		
		return "aspectFactory"+el.getId();
	}
	
	/**
	 * @param element
	 * @param fieldClass
	 * @return
	 * @throws Exception
	 */
	protected String generateFactoryInit(DynamicHookElement element, CtClass fieldClass, int index) throws Exception {
	
		String hname = getHookElementName(element);
		addField("jasco.runtime.connector.DynamicHookElement "+hname+";",fieldClass);
		String code =  accessField(hname)+"="+"(jasco.runtime.connector.DynamicHookElement)hooks.elementAt("+index+");"+JavaGenerator.NEWLINE;
		
		return code;
	}
	

	protected String generateFactoryAspectsForDynamicHookElements(DynamicHookElement el, String assignStartCode) {
		
		String type = el.getCutpoint().getClass().getName();
		String hookElementField = accessField(getHookElementName(el));
		String code=assignStartCode+" ("+type+") "+hookElementField+".getCutpoint(method);"+JavaGenerator.NEWLINE;
		return code;
		
	}

	private boolean shouldCheckCombStrats(Vector hooks) {
		
		return CombinationStrategyResolver.hasNonCachableCombStrats(hooks); 
	}
	
	private String getCombStratName() {
		
		return getCombStratName(this.currentCombStratId++);
	}
	
	private String getCombStratName(int i) {
		
		return "combstrat"+i;
	}

	private String generateCombStratInit(Vector hooks,CtClass fieldClass) throws Exception {
		
		Vector v = CombinationStrategyResolver.getNonCachableCombStrats(hooks);
		
		if(v.isEmpty())
			return "";
		
		String code = "java.util.Vector nonCachableV = jasco.runtime.cache.CombinationStrategyResolver#getNonCachableCombStrats(hooks);"+JavaGenerator.NEWLINE;
		
		Iterator i = v.iterator();
		while(i.hasNext()) {
			i.next();
			String combStratName = getCombStratName();
			addField("jasco.runtime.connector.CombinationStrategy "+combStratName+";", fieldClass);
			code+=accessField(combStratName)+" = (jasco.runtime.connector.CombinationStrategy) nonCachableV.elementAt(0);"+JavaGenerator.NEWLINE;
		}
		
		code+=accessField(getHookListName())+" = new jasco.runtime.connector.HookList(hooks);";
		
		addField("jasco.runtime.connector.HookList "+getHookListName()+";", fieldClass);
		
		return code;
	}
	
	private String generateCombStratExecution(Vector hooks) {
		
		Vector v = CombinationStrategyResolver.getNonCachableCombStrats(hooks);
		if(v.isEmpty())
			return "";
		
		String code=""+accessField(getHookListName())+".resetChanged();"+JavaGenerator.NEWLINE;
		int startId = currentCombStratId-v.size();
		for(int i=startId;i<currentCombStratId;++i) 
			code+=accessField(getCombStratName(i))+".validateCombinations("+accessField(getHookListName())+");"+JavaGenerator.NEWLINE;
			
		return code;
	}

	protected HookContextProperty computeContext(Vector hooks)  {
		
		HookContextProperty[] v = new HookContextProperty[hooks.size()];
		for (int i = 0; i < hooks.size() ; ++i) {
			HookElement cutpointElement = (HookElement)hooks.elementAt(i);
			IHook cp = cutpointElement.getCutpoint();
			v[i]=cp._Jasco_getHookContextPropertyForAdvice(HookContextProperty.GLOBAL);
		}
			 
		HookContextProperty result =  HookContextProperty.intersection(v);
		
		if(hasCustomNonCachableAspectFactories(hooks)) { //target and args have to init in jp for custom factories!
			result=new HookContextProperty(true,result.isDirectArgsArray(),result.isDirectCalledObject(),
					result.isDirectCalledMethod(),true,result.isProceedWithDefinedArgs());	
		}
		
		return result;
	}
	
	protected boolean hasCustomNonCachableAspectFactories(Vector hooks) {
		
		Iterator i = hooks.iterator();
		while(i.hasNext()) {
			HookElement next = (HookElement) i.next();
			if(!(next instanceof DynamicHookElement)) 
				continue;
			DynamicHookElement el = (DynamicHookElement) next;
			if(el.doNotCache()&&!(el.getAspectFactory() instanceof IPredefinedAspectFactory))
				return true;
		}
		return false;
	}
	
	protected boolean generateInvokeCode(java.util.Vector hooks, JascoMethod originalMethod, CtClass fieldClass, CachedAspectRep cachedAspectRep) throws java.lang.Exception {
	
		clearLinkedHooks=false;
		MethodGenerator invokeMethod = new MethodGenerator("invoke","java.lang.Object");
		invokeMethod.addVariable("jasco.runtime.JascoMethod","method");
		invokeMethod.addException("java.lang.Exception");
		invokeMethod.addModifier(Modifier.PUBLIC);
		
		String code = "";
		
		
		
		String startCode= "";
		
		//kan beter
		boolean noReplace = false;
		boolean noAfter = true;
		HookContextProperty requiredHooksContext = computeContext(hooks);

		Vector testHooks = new Vector();
		VectorSorter.sort(hooks,new CutpointReplaceComparator());
		Vector replaceVector = new Vector();
		Vector replaceVectorGeneric = new Vector();
		String firstReplaceHook = "";
		
		
		boolean executeCombStratTest = this.shouldCheckCombStrats(hooks);
		
		
		if(!checkWeavingNecessary(hooks))
			return false;
		
		for (int i = 0; i < hooks.size() ; ++i) {
			HookElement cutpointElement = (HookElement)hooks.elementAt(i);
			IHook cp = cutpointElement.getCutpoint();
			int id = cutpointElement.getId();
			 
			if(cutpointElement.getAfter().getExecution())
				noAfter=false;
			
			if (cutpointElement.getReplace().getExecution()&&HookHelper.definesAdvice(cutpointElement.getCutpoint(),IHook.REPLACE))
				replaceVector.add(cutpointElement);
			if(cutpointElement.getReplace().getExecution())
				replaceVectorGeneric.add(cutpointElement);
			
			code+=generateAspectInit(cutpointElement,cp.getClass().getName()+" "+getHookName(cutpointElement)+" = ");
			
			if (cp._Jasco_getConstructorRep().doNotCache())  {
				testHooks.add(cutpointElement);
				code+="boolean "+"hook"+id+"condition = "+accessFieldHook(cutpointElement)+"._Jasco_getConstructorRep().eval(method,"+accessFieldHook(cutpointElement,true)+"hashtable,false);"+JavaGenerator.NEWLINE;
				if(cp instanceof StatefulAspect) {
					code+="int "+"hook"+id+"State = "+accessFieldHook(cutpointElement)+"._jasco_setStatefulAspectState((("+PCutpointConstructorRegular.class.getName()+")"+accessFieldHook(cutpointElement)+"._Jasco_getConstructorRep()).getCurrentStateID());"+JavaGenerator.NEWLINE;
				}
			}
			
			if(HookHelper.definesAdvice(cp,IHook.IS_APPLICABLE)&&!testHooks.contains(cutpointElement)) {
				code+="boolean "+"hook"+id+"condition = "+accessFieldHook(cutpointElement)+".isApplicable("+generateAdviceArgs(IHook.IS_APPLICABLE, cutpointElement)+");"+JavaGenerator.NEWLINE;
				testHooks.add(cutpointElement);
			} else if(HookHelper.definesAdvice(cp,IHook.IS_APPLICABLE))
					code+="hook"+id+"condition = "+accessFieldHook(cutpointElement)+".isApplicable("+generateAdviceArgs(IHook.IS_APPLICABLE,cutpointElement)+") && hook"+id+"condition;"+JavaGenerator.NEWLINE;
		}
		
		if(executeCombStratTest)
			code+=generateCombStratExecution(hooks);
		
		if(executeCombStratTest) {
			for (int i = 0; i < hooks.size() ; ++i) {
				HookElement cutpointElement = (HookElement)hooks.elementAt(i);
				IHook cp = cutpointElement.getCutpoint();
				int id = cutpointElement.getId();
				String thisCondition = "jasco.runtime.cache.CombinationStrategyResolver"+"#shouldExecute("+accessFieldHook(cutpointElement)+", "+accessField(this.getHookListName())+")"; 
				if(!testHooks.contains(cutpointElement)) {
					code+= "boolean "+"hook"+id+"condition = "+thisCondition+";"+JavaGenerator.NEWLINE;;
					testHooks.add(cutpointElement);
				}
				else code+="hook"+id+"condition = "+"hook"+id+"condition &&"+thisCondition+";"+JavaGenerator.NEWLINE;
				
			}
		}
		
		if(replaceVector.size()==0)
			noReplace=true;
		else  {
			firstReplaceHook=accessFieldHook(((HookElement)replaceVector.elementAt(0)));
		}
		
		code = startCode+code;
		
		if (replaceVector.size()>0)  {
			//constrCode+="linkedHooks=new jasco.runtime.aspect.IHook["+(replaceVector.size()-1)+"];"+JavaGenerator.NEWLINE;
			addField("jasco.runtime.aspect.IHook[] "+getLinkedHooksName()+"=new jasco.runtime.aspect.IHook["+(replaceVector.size()-1)+"];",fieldClass);//=new jasco.runtime.aspect.IHook["+(replaceVector.size()-1)+"];",cachedAspect);
		}
		
		String replaceTestCode = "java.util.Vector linkedHooksVector = new java.util.Vector();"+JavaGenerator.NEWLINE;
		boolean testReplaces = false;
		//watch out to never invoke setLinkedHook on original method, otherwise in de soup!
		for (int i = 0; i < replaceVector.size() ; ++i) {
			HookElement cutpointElement = (HookElement)replaceVector.elementAt(i);
			int id = cutpointElement.getId();
			
			if(i>0)
				cachedAspectRep.appendToConstructorCode(accessField(getLinkedHooksName())+"["+(i-1)+"]="+accessFieldHook(cutpointElement)+";"+JavaGenerator.NEWLINE);
			
			if (testHooks.contains(cutpointElement))  {
				testReplaces=true;
				replaceTestCode+="if("+"hook"+id+"condition) "+JavaGenerator.NEWLINE;
				replaceTestCode+="linkedHooksVector.add("+accessFieldHook(cutpointElement)+");"+JavaGenerator.NEWLINE;
			}
			else replaceTestCode+="linkedHooksVector.add("+accessFieldHook(cutpointElement)+");"+JavaGenerator.NEWLINE;
		}
		
		this.computeInit(replaceVector,replaceVectorGeneric,requiredHooksContext,testReplaces);
		
		String aroundExType = computeAroundThrowingExceptionType(hooks);
		String aroundThrowingCatchCode="} catch("+aroundExType+" ex) {"+JavaGenerator.NEWLINE;
		boolean noAroundThrowing = true;
		
		String tempCatchCode = "";
		//	do around throwing
		for (int i = 0; i < replaceVectorGeneric.size() ; ++i) {
			
			HookElement cutpointElement = (HookElement)replaceVectorGeneric.elementAt(i);
			int id = cutpointElement.getId();
			if(!requiresInvocationThrowingReturning(IHook.AROUND_THROWING,cutpointElement.getCutpoint()))
				continue;
			
			if(!noAroundThrowing) {
				tempCatchCode="try{"+tempCatchCode+"continueBoolean=false;}catch("+aroundExType+" ex2) {ex=ex2;}";
				aroundThrowingCatchCode+=tempCatchCode;
				tempCatchCode="";
			}
			else {
				aroundThrowingCatchCode+="boolean continueBoolean=true;"+JavaGenerator.NEWLINE;
				noAroundThrowing=false;
			}
			
			String args= generateAdviceArgs(IHook.AROUND_THROWING,cutpointElement);
			if(!args.equals(""))
				args=", "+args;
			if (testHooks.contains(cutpointElement))  {
				tempCatchCode+="if(hook"+id+"condition) ";
			}
			tempCatchCode+="if(continueBoolean)";
			String definedOutput = defineOutput(originalMethod,requiredHooksContext);
			if(definedOutput!=null) {
				tempCatchCode+="output=";
			}
			tempCatchCode+=accessFieldHook(cutpointElement)+"."+IHook.AROUND_THROWING+"(ex "+args+");"+JavaGenerator.NEWLINE;
		}
		{
		aroundThrowingCatchCode+=tempCatchCode;
		
		}
		
		String aroundReturningCode = "";
		//		do around returning
		for (int i = replaceVectorGeneric.size()-1; i >=0 ; --i) {
			
			HookElement cutpointElement = (HookElement)replaceVectorGeneric.elementAt(i);
			int id = cutpointElement.getId();
			if(!requiresInvocationThrowingReturning(IHook.AROUND_RETURNING,cutpointElement.getCutpoint()))
				continue;
			
			String definedOutput = defineOutput(originalMethod,requiredHooksContext);
			if(definedOutput!=null) { //no output define with inline -> no around returning possible...
				if (testHooks.contains(cutpointElement))  {
					aroundReturningCode+="if(hook"+id+"condition) ";
				}
				String args= generateAdviceArgs(IHook.AROUND_RETURNING,cutpointElement);
				if(!args.equals(""))
					args=", "+args;
				aroundReturningCode+="output="+accessFieldHook(cutpointElement)+"."+IHook.AROUND_RETURNING+"("+this.castAdviceArgs(IHook.AROUND_RETURNING,"output")+args+");"+JavaGenerator.NEWLINE;
			}
		}
		
		cachedAspectRep.addInvokeInitCode(generateInvokeInitCode(replaceVector,requiredHooksContext,noReplace,testReplaces));
		
		VectorSorter.sort(hooks,new CutpointBeforeComparator());
		for (int i = 0; i < hooks.size() ; ++i) {
			HookElement cutpoint = (HookElement)hooks.elementAt(i);
			int id = cutpoint.getId();
			if (cutpoint.getBefore().getExecution())  {
				if (testHooks.contains(cutpoint))  {
					code+="if(hook"+id+"condition) ";
				}
				code+=accessFieldHook(cutpoint)+".before("+generateAdviceArgs(IHook.BEFORE,cutpoint)+");"+JavaGenerator.NEWLINE;
			}
		}
		
		String afterThrowingCatchCode="} catch("+computeAfterThrowingExceptionType(hooks)+" ex) {"+JavaGenerator.NEWLINE;
		boolean noAfterThrowing = true;
		VectorSorter.sort(hooks,new CutpointAfterComparator());
		for (int i = 0; i < hooks.size() ; ++i) {
			HookElement cutpoint = (HookElement)hooks.elementAt(i);
			int id = cutpoint.getId();
			if(cutpoint.getAfter().getExecution())  {
				
				if(!requiresInvocationThrowingReturning(IHook.AFTER_THROWING,cutpoint.getCutpoint()))
					continue;
				
				noAfterThrowing=false;
				if (testHooks.contains(cutpoint))  {
					afterThrowingCatchCode+="if(hook"+id+"condition) ";
				}
				String args= generateAdviceArgs(IHook.AFTER_THROWING,cutpoint);
				if(!args.equals(""))
					args=", "+args;
				afterThrowingCatchCode+=accessFieldHook(cutpoint)+".after_throwing(ex "+args+");"+JavaGenerator.NEWLINE;
			}	
		}
		boolean hasFinallyBlock = !noAfter||replaceVector.size()>1;
		boolean outputDefined =false;
		String tryCode="";
		if(!noAroundThrowing) {
			String definedOutput = defineOutput(originalMethod,requiredHooksContext);
			if(definedOutput!=null&&!outputDefined)  {
				tryCode+=definedOutput+JavaGenerator.NEWLINE; 
				outputDefined=true;
			}
			tryCode+="try{"+JavaGenerator.NEWLINE;
			aroundThrowingCatchCode+="}";
		}
		else aroundThrowingCatchCode="";
		
		if(!noAfterThrowing) {
			afterThrowingCatchCode+="throw(ex);}"+JavaGenerator.NEWLINE; 
			String definedOutput = defineOutput(originalMethod,requiredHooksContext);
			if(definedOutput!=null&&!outputDefined)  {
				tryCode+=definedOutput+JavaGenerator.NEWLINE; 
				outputDefined=true;
			}
			tryCode+="try{"+JavaGenerator.NEWLINE;
		} 
		else afterThrowingCatchCode="";
		
		if(noAroundThrowing&&noAfterThrowing) //geen try
			if(hasFinallyBlock){//but after->need finally block! of setLinkedHooks, dus ook finally blok nodig!
				String definedOutput = defineOutput(originalMethod,requiredHooksContext);
				if(definedOutput!=null&&!outputDefined)  {
					tryCode+=definedOutput+JavaGenerator.NEWLINE; 
					outputDefined=true;
				}
				tryCode+="try{"+JavaGenerator.NEWLINE;
			}
		
		code+=tryCode;
		
		if (noReplace)  {
			code+=generateOriginalMethodInvokeNoReplace(originalMethod,requiredHooksContext,outputDefined)+";"+JavaGenerator.NEWLINE;
		}
		else if (testReplaces)  {
			if(replaceVector.size()==1) {
				if(!outputDefined) {
					code+="Object output = null;"+JavaGenerator.NEWLINE;
				}
				
				code+="if("+firstReplaceHook+"condition) "+JavaGenerator.NEWLINE;
				code+=JavaGenerator.TAB+"output = "+firstReplaceHook+"."+IHook.REPLACE+"("+generateAdviceArgs(IHook.REPLACE,replaceVector,testReplaces)+");"+JavaGenerator.NEWLINE;
				code+="else ";
				code+="output = "+generateOriginalMethodInvoke(originalMethod)+";"+JavaGenerator.NEWLINE;
			}
			else  {
				code+=replaceTestCode;
				if(!outputDefined) {
					code+="Object output = null;"+JavaGenerator.NEWLINE;
				}
				code+="if(linkedHooksVector.size()>0) {"+JavaGenerator.NEWLINE;
				code+=JavaGenerator.TAB+"jasco.runtime.aspect.IHook firstReplaceHook = linkedHooksVector.elementAt(0);"+JavaGenerator.NEWLINE;
				code+=JavaGenerator.TAB+"linkedHooksVector.removeElementAt(0);"+JavaGenerator.NEWLINE;
				code+=JavaGenerator.TAB+"jasco.runtime.aspect.IHook[] linkedHooksTested = new jasco.runtime.aspect.IHook[linkedHooksVector.size()];"+JavaGenerator.NEWLINE;
				code+=JavaGenerator.TAB+"linkedHooksVector.copyInto(linkedHooksTested);"+JavaGenerator.NEWLINE;
				code+=JavaGenerator.TAB+"method.setLinkedHooks(linkedHooksTested);"+JavaGenerator.NEWLINE;
				code+=JavaGenerator.TAB+"output = firstReplaceHook."+IHook.REPLACE+"("+generateAdviceArgs(IHook.REPLACE,replaceVector,testReplaces)+");"+JavaGenerator.NEWLINE;
				clearLinkedHooks=true;
				code+="}"+JavaGenerator.NEWLINE;
				code+="else output = method.invokeOriginalJAsCoMethod();"+JavaGenerator.NEWLINE;	
			}
		}
		else code+=generateReplaceSequence(requiredHooksContext,replaceVector,testReplaces,fieldClass,outputDefined);
		
		code+=aroundReturningCode;
		
		//after returning:
		VectorSorter.sort(hooks,new CutpointAfterComparator());
		for (int i = 0; i < hooks.size() ; ++i) {
			HookElement cutpoint = (HookElement)hooks.elementAt(i);
			int id = cutpoint.getId();
			if(cutpoint.getAfter().getExecution())  {
				String definedOutput = defineOutput(originalMethod,requiredHooksContext);
				if(definedOutput!=null) { //no output define with inline -> no after returning possible...
					
					if(requiresInvocationThrowingReturning(IHook.AFTER_RETURNING,cutpoint.getCutpoint())) {
						if (testHooks.contains(cutpoint))  {
							code+="if(hook"+id+"condition) ";
						}
						String args= generateAdviceArgs(IHook.AFTER_RETURNING,cutpoint);
						if(!args.equals(""))
							args=", "+args;
						code+=accessFieldHook(cutpoint)+".after_returning("+this.castAdviceArgs(IHook.AFTER_RETURNING,"output")+args+");"+JavaGenerator.NEWLINE;
					}
					
				}
			}	
		}
		//now catch blocks
		code+=aroundThrowingCatchCode+afterThrowingCatchCode;
		
		if(hasFinallyBlock) {//add finally block
			if(noAroundThrowing&&noAfterThrowing)
				code+="}";//TRY NOG NIET CLOSED
			code+="finally {"+JavaGenerator.NEWLINE;
		}
		
//		clear linked hooks if necessary
		if(clearLinkedHooks)
			code+="method.clearLinkedHooks();"+JavaGenerator.NEWLINE;
		
		//still sorted through after!
		for (int i = 0; i < hooks.size() ; ++i) {
			HookElement cutpoint = (HookElement)hooks.elementAt(i);
			int id = cutpoint.getId();
			if(cutpoint.getAfter().getExecution())  {
				if(!HookHelper.definesAdvice(cutpoint.getCutpoint(),IHook.AFTER))
					continue;
				if (testHooks.contains(cutpoint))  {
					code+="if(hook"+id+"condition) ";
				}
				code+=accessFieldHook(cutpoint)+".after("+generateAdviceArgs(IHook.AFTER,cutpoint)+");"+JavaGenerator.NEWLINE;
			} 
		}
		
	
		for (int i = 0; i < hooks.size() ; ++i) {
			HookElement cutpoint = (HookElement)hooks.elementAt(i);
			int id = cutpoint.getId();
			code+=moreInvokeCode(cutpoint.getCutpoint(), accessFieldHook(cutpoint), "hook"+id+"condition");
	
		}
		
		for (int i = 0; i < hooks.size() ; ++i) {
			HookElement cutpoint = (HookElement)hooks.elementAt(i);
			int id = cutpoint.getId();
			IHook cp = cutpoint.getCutpoint();
			if(cp instanceof StatefulAspect) {
				code+=accessFieldHook(cutpoint)+"._jasco_setStatefulAspectState(hook"+id+"State);"+JavaGenerator.NEWLINE;
			}
		}
		
		if(hasFinallyBlock) //cloase finally bloack
			code+="}";
		
		code+=doReturn("output",noReplace,requiredHooksContext)+JavaGenerator.NEWLINE;
		
		invokeMethod.setImplementation(code);
		
		cachedAspectRep.setInvokeCode(invokeMethod);
		
		return true;
	}
	
	

	private boolean checkWeavingNecessary(Vector hooks) {
		for (int i = 0; i < hooks.size() ; ++i) {
			HookElement cutpointElement = (HookElement)hooks.elementAt(i);
			IHook cp = cutpointElement.getCutpoint();
			int id = cutpointElement.getId();
			if(cutpointElement.getBefore().getExecution()) {
				return true;
			}
			if(cutpointElement.getReplace().getExecution()) {
				if(HookHelper.definesAdvice(cp,IHook.REPLACE))
					return true;
				if(HookHelper.definesAdvice(cp,IHook.AROUND_RETURNING)&&requiresInvocationThrowingReturning(IHook.AROUND_RETURNING,cp))
					return true;
				if(HookHelper.definesAdvice(cp,IHook.AROUND_THROWING)&&requiresInvocationThrowingReturning(IHook.AROUND_THROWING,cp))
					return true;
			}	
			if(cutpointElement.getAfter().getExecution()) {
				if(HookHelper.definesAdvice(cp,IHook.AFTER))
					return true;
				if(HookHelper.definesAdvice(cp,IHook.AFTER_THROWING)&&requiresInvocationThrowingReturning(IHook.AFTER_THROWING,cp))
					return true;
				if(HookHelper.definesAdvice(cp,IHook.AFTER_RETURNING)&&requiresInvocationThrowingReturning(IHook.AFTER_RETURNING,cp))
					return true;
			}
		}
		return false;
	}

	/**
	 * @param cutpointElement
	 * @return
	 */
	private String generateAspectInit(HookElement cutpointElement, String assignCodeStart) {
		String code = "";
		if(cutpointElement instanceof DynamicHookElement)
			code+=generateFactoryAspectsForDynamicHookElements((DynamicHookElement)cutpointElement,assignCodeStart);
		if(code.equals(""))
			return assignCodeStart+=accessFieldHook(cutpointElement,true)+";"+JavaGenerator.NEWLINE;
		else return code;
	}
	

	/**
	 * @return
	 */
	private String getLinkedHooksName() {
		return "linkedHooks"+this.currentHookId;
	}
	
	private String getHookListName() {
		return "hookList"+this.currentHookId;
	}

	protected String moreInvokeCode(IHook hook, String accessHook, String conditionboolean) {
		
		return "";
	}
	
	protected final String computeAfterThrowingExceptionType(Vector hooks) {
		
		return computeThrowingExceptionType(hooks,IHook.AFTER_THROWING);
	}
	
	protected final String computeAroundThrowingExceptionType(Vector hooks) {
		
		return computeThrowingExceptionType(hooks,IHook.AROUND_THROWING);
	}
	
	protected String computeThrowingExceptionType(Vector hooks, String adviceType) {
		
		return "Exception";
	}
	
	public String generateAdviceArgs(String adviceType, HookElement applicableHook) {
		
		if(adviceType.equals(IHook.REPLACE))
			throw new IllegalArgumentException("this method cannot be invoked for around advice!");
		
		Vector v = new Vector();
		v.add(applicableHook);
		return generateAdviceArgs(adviceType,v,false);
	}
	
	/**
	 * only for returning/throwing advices, might be that there is no advice declared
	 * 
	 * @param adviceType
	 * @return
	 */
	public boolean requiresInvocationThrowingReturning(String adviceType, IHook theHook) {
		
		return HookHelper.definesAdvice(theHook,adviceType);
	} 
	
	public String generateAdviceArgs(String adviceType, Vector applicableHooks, boolean testReplaces) {
		
		return "method, method.getCalledObject(), method.getArgumentsArray()";
	}
	
	public String defineOutput(JascoMethod newMethod, HookContextProperty requiredHooksContext) throws Exception {
		
		return "Object output = null;";
	}
	
	protected String castAdviceArgs(String adviceType, String resultName) {
		
		
		return resultName;
			
	}
	
	protected String generateReplaceSequence(HookContextProperty requiredHooksContext, Vector replaceVector, boolean testReplaces, CtClass fieldClass, boolean outputDefined) throws Exception {
	
		String code="";
		String firstReplaceHook = accessFieldHook(((HookElement)replaceVector.elementAt(0)));
		
		if(replaceVector.size()==1) {
			if(!outputDefined)
				code+="Object ";
			code+="output = "+firstReplaceHook+"."+IHook.REPLACE+"("+generateAdviceArgs(IHook.REPLACE,replaceVector,testReplaces)+");"+JavaGenerator.NEWLINE;
		}
		else  {		
			code+="method.setLinkedHooks("+accessField(getLinkedHooksName())+");"+JavaGenerator.NEWLINE;
			if(!outputDefined)
				code+="Object ";
			code+="output = "+firstReplaceHook+"."+IHook.REPLACE+"("+generateAdviceArgs(IHook.REPLACE,replaceVector,testReplaces)+");"+JavaGenerator.NEWLINE;
			setClearLinkedHooks();
		} 
		
		return code;
	}
	
	protected String generateInvokeInitCode(Vector hooks, HookContextProperty requiredHooksContext, boolean noReplace, boolean testReplaces) throws Exception {
		
		return "";
	}
	 
	protected String generateOriginalMethodInvokeNoReplace(JascoMethod method, HookContextProperty requiredHooksContext, boolean outputDefined) throws Exception {
		
		String code="";
		if(!outputDefined)
			code+="Object ";
		code+="output = "+generateOriginalMethodInvoke(method);
			
		
		return code;
	}
	
	protected String generateOriginalMethodInvoke(JascoMethod method) throws Exception {
		
		return "method.invokeOriginalJAsCoMethod()";
	}
	
	protected String doReturn(String var, boolean noReplace, HookContextProperty requiredHooksContext) throws Exception   {
	
		return "return "+var+";";
	}
	
	//TODO change field accessess...
		
	public CombinedHooks generateCachedAspectJavassist(java.util.Vector thehooks, JascoMethod originalMethod) throws java.lang.Exception {
		
		
		if (Options.hasJuttaPredefinedCache())  {
			CombinedHooks res = checkForPredef(thehooks);
			if (res != null)  {
				if(Options.showDebugOutput())
					jasco.util.logging.Logger.getInstance().showOutput("JASCO DEBUG: using predefined cached behavior "+res.getClass().getName());
				return res;
			}
		}
		
		Vector hooks=(Vector)thehooks.clone();
		String cname = "CachedAspect"+getNextClassId();	
		CachedAspectRep rep = new CachedAspectRep(cname);
		CtClass cachedAspect = classPool.makeClass("jasco.runtime.cache."+cname);
		 
		CtClass combHooks = classPool.get("jasco.runtime.CombinedHooks");
		cachedAspect.addInterface(combHooks);
		
		generateConstructorCode(cname,hooks,originalMethod,cachedAspect,rep);	
		if(!generateInvokeCode(hooks,originalMethod,cachedAspect,rep)) //no aspect advices to be executed...
			return CombinedHooksFactory.EMPTY_CONBINED_HOOKS;
		
		if(Options.showDebugOutput()) {
			//System.out.println(Options.getTempDir());
			Options.getLogDir().mkdirs();
			FileWriter boe = new FileWriter(Options.getLogDir()+File.separator+"jutta"+(logid++)+".txt");
			boe.write(rep.getInvokeCode());boe.write(rep.getConstructorCode());
			boe.close();
		}
		
		cachedAspect.stopPruning(true);
		
		Class cachedClass = null;
		CombinedHooks cached = getCachedResult(rep.getInvokeCode(),rep.getConstructorCode());
		if (cached!=null)  {
			cachedClass=cached.getClass();	
		}
		else  {
			cachedAspect.addConstructor(CtNewConstructor.make(rep.getConstructorCode(),cachedAspect));
			CtMethod invokeMethodBytes = CtNewMethod.make(rep.getInvokeCode(),cachedAspect);
			cachedAspect.addMethod(invokeMethodBytes);
			cachedClass = cachedAspect.toClass();
		}
	
		cached = (CombinedHooks) cachedClass.getConstructor(new Class[]  {Vector.class, MethodJoinpoint.class}).newInstance(new Object[] {thehooks,null});
		addCachedResult(rep,cached);
		
		cachedAspect.defrost();
		return cached;
		
	} 
	
	
	protected void addCachedResult(CachedAspectRep rep, CombinedHooks hooks)  {
	
		codeCache.put(rep.getInvokeCode()+rep.getConstructorCode(),hooks); 
	}
	
	protected CombinedHooks getCachedResult(String icode, String ccode)  {
	
		CombinedHooks res = (CombinedHooks) codeCache.get(icode+ccode);
		return res;
	}	
	
	protected void computeInit(Vector replaceVector, Vector genericReplaceVector, HookContextProperty requiredHooksContext, boolean testReplaces) {
		
	}
}