package jasco.runtime.distribution;

import jasco.runtime.ConnectorRegistry;
import jasco.runtime.MethodJoinpoint;
import jasco.runtime.Advice.Type;
import jasco.runtime.aspect.IHook;
import jasco.runtime.connector.Connector;
import jasco.runtime.connector.HookElement;
import jasco.util.logging.Logger;

import java.lang.reflect.Proxy;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.concurrent.Callable;
import java.util.concurrent.FutureTask;


public class DistributedInvoker {
		
	private static DistributedInvoker _managerInstance = new DistributedInvoker();
	
	private DistributedInvoker() {
		super();
	}
	
	public static DistributedInvoker getInstance() {
		
		return _managerInstance;
	}
	
	/**
	 * What this method should do is invoke the
	 * DistributedInvoker.invokeSynchronous method on the correct
	 * host in a synchronous manner. The host can be fetched from the
	 * adviceExecution argument. The invoked method will take care of executing
	 * the advice etc... I will also make sure that the
	 * DistributedInvoker.invokeRemoteSynchronous method is executed
	 * from synchronous advices (have to alter compiler for that).
	 * 
	 * Thats all that needs to be changed to the distributed implementation. The
	 * joinpoints still need to be broadcasted for asynch advice execution. I
	 * will just make sure that synch advices are not executed on such
	 * joinpoints.
	 */
	public Object invokeRemoteSynchronous(DistributedAdviceExecution adviceExecution) {

		if(adviceExecution.getHost().isLocalHost()) //optimization. if this is local host, execute directly instead of via JGroups
			return invokeSynchronous(adviceExecution);

		else {
			//	Makes the remote synchronous call
			//return invokeSynchronous(adviceExecution);
			return DistributePointcutHandler.getInstance().invokeRemoteSynchronous(adviceExecution);
		}
	}
	
	public Object invokeSynchronous(DistributedAdviceExecution adviceExecution) {
		return new Invoker(adviceExecution).call();
	}
	
	public Object invokeRemoteASynchronous(DistributedAdviceExecution adviceExecution) {
		
		ObjectFutureWrapper wrapper;
		
		if(adviceExecution.getHost().isLocalHost()) //optimization. if this is local host, execute directly instead of via JGroups
			wrapper= invokeASynchronous(adviceExecution);

		else {
			//	Makes the remote synchronous call
			//return invokeSynchronous(adviceExecution);
			wrapper= DistributePointcutHandler.getInstance().invokeRemoteASynchronous(adviceExecution);
		}
		
		return generateTransparantProxy(wrapper,adviceExecution.getJP());
	}
	
	public Object generateTransparantProxy(ObjectFutureWrapper wrapper, MethodJoinpoint jp) {
		Class returnType = jp.getReturnType();
		if(!returnType.isInterface()) {
			Logger.getInstance().showError("jp return type should be an interface for asynchronous advice! "+jp.getSignature());
			return null;
		}
		
		return Proxy.newProxyInstance(jp.getClassLoader(), new Class[] {returnType}, new FutureWrapper(wrapper.getHost(),wrapper.getId()));
	}
	
	protected class Invoker implements Callable {
		
		private DistributedAdviceExecution adviceExecution; 
		
		public Invoker(DistributedAdviceExecution adviceExecution) {
			
			this.adviceExecution=adviceExecution;
		}
		
		public Object call() {
			
			IHook hook = fetchHook(adviceExecution.getConnectorClassName(),adviceExecution.getAspectInstanceName(),adviceExecution.getJP());
			
			try {
				return executeAdvice(hook,adviceExecution.getAdviceType(),adviceExecution.getJP());
			} catch (Exception e) {
				Logger.getInstance().showError("throwing exceptions for distr synch advice executions is not yet supported");
				e.printStackTrace();
				return null;
			}
		}
	}
	
	
	public ObjectFutureWrapper invokeASynchronous(DistributedAdviceExecution adviceExecution) {
		FutureTask future = new FutureTask(new Invoker(adviceExecution));
		Thread t = new Thread(future);
		t.start();
		return FutureManager.getInstance().registerFutureTask(future);
	}
	
	
	
	private Object executeAdvice(IHook hook, Type adviceType, DistributedJascoMethod jp) throws Exception {
		
		Object result = null;
		switch(adviceType) {
		
		case BEFORE:
			hook.before(jp,jp.getCalledObject(),jp.getArgumentsArray());
			break;
			
		case AFTER:
			hook.after(jp,jp.getCalledObject(),jp.getArgumentsArray());
			break;
			
		case AROUND:
			jp.clearLinkedHooks();
			result=hook.around(jp,jp.getCalledObject(),jp.getArgumentsArray());
			break;
			
		case ISAPPLICABLE:
			result=hook.isApplicable(jp,jp.getCalledObject(),jp.getArgumentsArray());
			break;
			
		default:
			throw new IllegalArgumentException("Advice type not yet supported for distributed synchronized advice execution: "+adviceType);
		
		}
		
		
		return result;
		
		/*
		boolean  exBefore=false,exAfter=false,exAround = false;
		
		switch(adviceType) {
		
		case BEFORE:
			exBefore=true;
			break;
			
		case AFTER:
			exAfter=true;
			break;
			
		case AROUND:
			exAround=true;
			break;
			
		case ISAPPLICABLE:
			Object result=hook.isApplicable(jp,jp.getCalledObject(),jp.getArgumentsArray());
			return result;
		}
		
		CutpointExecutionElement before = new CutpointExecutionElement(exBefore,0);
		CutpointExecutionElement after = new CutpointExecutionElement(exAfter,0);
		CutpointExecutionElement around = new CutpointExecutionElement(exAround,0);
		
		HookElement he = new HookElement(hook,before,after,around,hook._Jasco_getCutpointName());
		Vector res = new Vector();
		res.add(he);
		
		return JAsCoJPInterpreter.executeHooks(res,jp,false);*/
	}

	private IHook fetchHook(String connectorName, String aspectName, MethodJoinpoint jp) {
		
		Iterator i = ConnectorRegistry.getConnectors();
		while(i.hasNext()) {
			Connector c = (Connector) i.next();
			if(c.getClass().getName().equals(connectorName))
				return fetchHook(c,aspectName, jp);
		}
		
		throw new IllegalArgumentException("connector not found for "+connectorName);
	}

	private IHook fetchHook(Connector c, String aspectName, MethodJoinpoint jp) {
		
		Enumeration i = c.getHooks();
		while(i.hasMoreElements()) {
			HookElement hookEl = (HookElement) i.nextElement();
			IHook hook = hookEl.getCutpoint(jp);
			if(hook._Jasco_getCutpointName().equals(aspectName))
				return hook;
		}
		
		throw new IllegalArgumentException("hook not found for "+aspectName+" in connector "+c.getName());
	}

}
