/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.util.generators;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Iterator;
import java.util.Vector;

public class ClassGenerator extends JavaGenerator {
	
	public static final int ANNOTATION =  0x08000000;

	private GPackage thepackage = null;
    private String       superclass = "";
    private final Vector interfaces = new Vector();
    private final GImports imports = new GImports();
    private final Vector methods = new Vector();
    private final Vector variables = new Vector();
	private final Vector staticInits = new Vector();
	private final Vector plainJava = new Vector();

    public ClassGenerator() {
        this("unknown", new GPackage(""));
    }

    public ClassGenerator(String name, GPackage aPackage) {
        super(name);
        setPackage(aPackage);
    }
    
    public static final boolean isAnnotation(int modifiers) {
    	
    	return (modifiers & ANNOTATION)!=0;
    }
    
    public static final int setAnnotation(int modifiers) {
    	
    	modifiers|=ANNOTATION;
    	return modifiers;
    }
    
	public ClassGenerator(String name, String aPackage) {
        this(name,new GPackage(aPackage));
        
    }

    public ClassGenerator(String name) {
        this(name, new GPackage(""));
    }
    
    public String getFullName() {
    	
    	if(hasPackage())
    		return thepackage.getPackageName()+"."+getName();
    	else return getName();
    	
    }
    
	public boolean containsMethod(String name) {
		
		Iterator i = methods.iterator();
		while(i.hasNext()) {
			MethodGenerator method = (MethodGenerator) i.next();
			if(method.getName().equals(name))
				return true;
		}
		return false;
	}
	
	public boolean containsVariable(String name) {
		
		Iterator i = variables.iterator();
		while(i.hasNext()) {
			VariableGenerator method = (VariableGenerator) i.next();
			if(method.getName().equals(name))
				return true;
		}
		return false;
	}
	
	public void addStaticInitFirst(String code)  {
		
			staticInits.insertElementAt(code,0);
		}
	
	public void addStaticInit(String code)  {
	
		staticInits.add(code);
	}
	
	public Iterator getStaticInits()  {
	
		return staticInits.iterator();
	}
	
	public Iterator getPlainJava()  {
	
		return plainJava.iterator();
	}
	
	public Iterator getConstructors()  {
	
		Vector res = new Vector();
		Iterator i =getMethods();
		while (i.hasNext())  {
			MethodGenerator g = (MethodGenerator) i.next();
			if(g.getReturnType().equals(""))
				res.add(g);
		}	
		return res.iterator();
	}	
	
	public void addPlainJava(String code,String documentation)  {
		plainJava.add(code + documentate(documentation));
	}
	
	public void addPlainJava(String code)  {
		plainJava.add(code);
	}

    public GPackage getPackage() {
        return thepackage;
    }
	
	protected void setPackage(GPackage aPackage) {
        this.thepackage = aPackage;
    }
	
	public void setPackage(String aPackage) {
        this.thepackage = new GPackage(aPackage);
    }
	
	public String getPackageName() 
	{	
		if(getPackage()!=null)
			return getPackage().getPackageName();
		else return "";	
	}
	
	public String getPackageDocumentation()
	{
		return getPackage().getDocumentation();
	} 

	//added for documentation	
	private String documentate(String aDocumentation)
	{
		if (aDocumentation != "") 
		{
			return " //" + aDocumentation;
		}
		else
		{
			return "";
		}
	}
	
    public boolean hasPackage() {
		if(getPackageName()==null)
			return false;
        else return !getPackageName().equals("");
    } 

    public void addImportPackage(GImport aImport) {
		if (!(isImportPackage(aImport.getImportName())))
		{
        	imports.addImport(aImport);
		}  
    } 

    public Iterator getImportPackages() {
        return imports.iterator();
    } 

    public boolean isImportPackage(String pack) {
        return imports.contains(pack);
    } 

    public void addMethod(MethodGenerator method) {
        methods.add(method);
    } 
	
	public MethodGenerator getMethod(String varname) {
        Iterator i = getMethods();
		while (i.hasNext())  {
			MethodGenerator var = (MethodGenerator) i.next();
			if(var.getName().equals(varname))
				return var;
		}
		return null;	
    }

    public Iterator getMethods() {
        return methods.iterator();
    } 

    public void addVariable(VariableGenerator variable) {
        variables.add(variable);
    } 

    public Iterator getVariables() {
        return variables.iterator();
    } 
	
	public VariableGenerator getVariable(String varname) {
        Iterator i = getVariables();
		while (i.hasNext())  {
			VariableGenerator var = (VariableGenerator) i.next();
			if(var.getName().equals(varname))
				return var;
		}
		return null;	
    }

    public void setSuperClass(String name) {
        superclass = name;
    } 

    public String getSuperClass() {
        return superclass;
    } 

    public boolean hasSuperClass() {
        return !superclass.equals("");
    } 

    public void addInterface(String name) {
        if(!interfaces.contains(name))
			interfaces.add(name);
    } 

    public Iterator getInterfaces() {
        return interfaces.iterator();
    } 

    public String addImportAndGetClassName(String className) {
        int i = className.lastIndexOf(".");
        if (i == -1) return className;
        String pack = className.substring(0, i);
        addImportPackage(new GImport(pack + ".*",""));
        return className.substring(i + 1);
    } 
    public String getClassName(String className) {
        int i = className.lastIndexOf(".");
        if (i == -1) return className;
        return className.substring(i + 1);
    } 
    
    public void addInnerClass(ClassGenerator gen) {
    	
    	this.addPlainJava(gen.toString());
    }

    public String toString() {
        String code = "";
        if (hasPackage())
            code += "package " + getPackageName() + ";" + documentate(getPackageDocumentation()) + NEWLINE + NEWLINE;

        Iterator importi = getImportPackages();
        while (importi.hasNext()) {
            GImport next = (GImport) importi.next();
            code += "import ";
            if(next.isStatic())
            	code+="static ";
            code+= next.getImportName() + ";" + documentate(next.getDocumentation()) + NEWLINE;
        } 

        code += NEWLINE;
        code += super.toString();
        code+=getAnnotationsString()+JavaGenerator.NEWLINE;

        if (Modifier.isPublic(getModifiers()))
            code += "public ";
		
		if (Modifier.isPrivate(getModifiers()))
            code += "private ";	
		
		if (Modifier.isProtected(getModifiers()))
            code += "protected ";		

        if (Modifier.isAbstract(getModifiers()))
            code += "abstract ";

        if (Modifier.isFinal(getModifiers()))
            code += "final ";
		
		if (Modifier.isStatic(getModifiers()))
            code += "static ";		

        if (Modifier.isInterface(getModifiers()))
            code += "interface ";
        else if(isAnnotation(getModifiers()))
        	code += "@interface ";
        else code += "class ";

        code += getName() + " ";
        
        code+= getGenerics()+" ";

        if (hasSuperClass())
            code += "extends " + getSuperClass() + " ";

        String extend = "implements ";
        if (Modifier.isInterface(getModifiers()))
            extend = "extends ";

        Iterator inti = getInterfaces();
        if (inti.hasNext()) {
            String next = (String) inti.next();
            code += extend + next;

            while (inti.hasNext()) {
                next = (String) inti.next();
                code += ", " + next;
            } 
        } 

        code += " {" + NEWLINE + NEWLINE;

        Iterator vari = getVariables();
        while (vari.hasNext()) {
            String next = addTabsBeforeLines(((VariableGenerator) vari.next()).toString());
            code += next + NEWLINE;
        } 

        code += NEWLINE;
		
		Iterator stai = getStaticInits();
        while (stai.hasNext()) {
			String next = stai.next().toString();
			code += TAB+"static {"+NEWLINE+NEWLINE;
			code += addTabsBeforeLines(next);
			code += TAB + "}"+NEWLINE;
        } 

        code += NEWLINE;

        Iterator methodi = getMethods();
        while (methodi.hasNext()) {
            String next = addTabsBeforeLines(((MethodGenerator) methodi.next()).toString());
            code += next + NEWLINE + NEWLINE;
        }
		
		Iterator plj = getPlainJava();
        while (plj.hasNext()) {
			String next = plj.next().toString();
			code += addTabsBeforeLines(next);//+NEWLINE
        } 

        code += NEWLINE; 

        code += "}";

        return code;
    } 

    public void addMainMethod(String code) {
        MethodGenerator main = new MethodGenerator("main", "void");
        main.addModifier(Modifier.PUBLIC);
        main.addModifier(Modifier.STATIC);
        main.setImplementation(code);
        main.addVariable("String[]", "args");
        addMethod(main);
    } 

    // add a method with a fixed implementation based on the signature of an existing method
    public void addMethodToFile(Method method, String methodBody) {
        // Method heading
        String          methodName = method.getName();
        String          returnType = method.getReturnType().getName();
        MethodGenerator mymethod = new MethodGenerator(methodName, returnType);
        mymethod.addModifier(method.getModifiers());
        // Parameters
        Class[] parameterTypes = method.getParameterTypes();
        for (int k = 0; k < parameterTypes.length; k++) {
            mymethod.addVariable(parameterTypes[k].getName(), "par" + (k + 1));
        } 
        // Exceptions
        Class[] exceptionTypes = method.getExceptionTypes();
        for (int k = 0; k < exceptionTypes.length; k++) {
            mymethod.addException(exceptionTypes[k].getName());
        } 
        // Method Implementation
        mymethod.setImplementation(methodBody);
        addMethod(mymethod);
    } 

}


