package org.jgroups.protocols;

import jasco.runtime.distribution.causality.LocalVectorClock;

import org.jgroups.Address;
import org.jgroups.Event;
import org.jgroups.Message;
import org.jgroups.protocols.CAUSALTAGS.CausalHeader;
import org.jgroups.protocols.CAUSALTAGS.CausalNewViewHeader;


/**
 * LDBN -4-03-2008
 * This class handle the adition and deletion of hosts regarding the size of the vector clocks.
 * It won't increment the vector clock. The clock should be managed by the
 * sequence constructs. NOTE: we could include some message encoding to manage
 * the update of clock information???
 *
 * @author dnielben
 *
 */
public class AWEDLOCALCLOCKMANAGEMENT extends CAUSALTAGS {

	public AWEDLOCALCLOCKMANAGEMENT() {
		super();
		
		// This class will be used to have access to the current local
		// vector clock
		LocalVectorClock.getInstance().setManager(this);
	}
	
    /**
     * 
     * This message is a modified version of the parent class
     * It does not modify the clock information.
     * It only pass the message. adding the locl clock
     * information.
     * 
     * NOTE this is the message that is going to be sent
     * 
     * Process a downward event.
     * @param evt The event.
     */
    public Object down(Event evt) {
        try {
            // If not a MSG, just pass down.
            if (evt.getType()!=Event.MSG) {
                return down_prot.down(evt);
            }
            
            Message msg = (Message) evt.getArg();
            
            // If unicast, just pass down.
            if (msg.getDest()!=null && ! msg.getDest().isMulticastAddress()) {
                return down_prot.down(evt);
            }
    
            // Multicast MSG:
            // - if enabled, get the next time vector, add it and pass down;
            // - otherwise, add to the downward waiting queue.
            TransportedVectorTime tvt=null;
            
            synchronized(lock) {
                if (isEnabled()) {
                	// LDBN - 4-3-2008
                	// No clock update, only pointcuts can increase bector clock
                    // currentView.increment();
                    tvt=currentView.getTransportedVectorTime();
                    if (log.isDebugEnabled()) log.debug("Sent 1 down message @ "+currentView.timeVectorString());
                } else {
                    if (log.isDebugEnabled()) log.debug("Enqueued 1 down message...");
                    downwardWaitingQueue.add(evt);
                }
            }
            
            if (tvt!=null) {
                msg.putHeader(getName(), new CausalHeader(tvt));
                return down_prot.down(evt);
            }
        } catch (RuntimeException e) {
            if (debug) log.error("*** down: "+e.getMessage(), e);
            throw e;
        }

        return null;
    }
    
    // This method just pass the message to the applicationprivate
    //NOTE: This is the message that is received
    public Object upMsg(Event evt) {
        Message msg = (Message) evt.getArg();
        Address src=msg.getSrc();
        
        // Check for a causal new view header
        Object obj = msg.getHeader(CausalNewViewHeader.NAME);

        if (obj instanceof CausalNewViewHeader) {
            processNewViewSynchronization(src, (CausalNewViewHeader)obj, msg.getObject());           
            return null;
        }
        
        obj = msg.getHeader(getName());

        if (!(obj instanceof CausalHeader)) {
            if((msg.getDest() == null || msg.getDest().isMulticastAddress()) 
                    && log.isErrorEnabled()) log.error("NO CAUSAL.Header found");
            return up_prot.up(evt);
        }
        
        TransportedVectorTime messageVector = ((CausalHeader)obj).getVectorTime();
        
        synchronized (lock) {
            if (currentView==null||currentView.getView().getIndex(src)<0) {
                if (log.isDebugEnabled()) log.debug("Discarding "+obj+" from "+msg.getSrc());
                return null;
            }
            
            // LDBN - 4-3-2008
            // We pass the message and do not order the delivery of messages to
            // the application
                if (log.isDebugEnabled()) log.debug("passing up message "+msg+", headers are "+msg.printHeaders()+", local vector is "+currentView.timeVectorString());
                up_prot.up(evt);
                
                // LDBN - 4-3-2008
                // No clock update is made. Only Sockets communication can modify the local clock
                // currentView.max(messageVector);
            
        }
        return null;
    }
    
    public TransportedVectorTime getCurrentVectorTime(){
    	synchronized(lock){
    	   return currentView.getTransportedVectorTime();
    	}
    }
    
    public void computeCurrentVectorTime(TransportedVectorTime tvt){
    	synchronized(lock){
     	    currentView.max(tvt);
     	}
    }
}
