package org.cachebench.cachewrappers;

import net.sf.ehcache.CacheManager;
import net.sf.ehcache.Ehcache;
import net.sf.ehcache.Element;
import net.sf.ehcache.config.Configuration;
import net.sf.ehcache.config.ConfigurationFactory;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.cachebench.SerializableCacheWrapper;

import java.io.Serializable;
import java.net.URL;
import java.util.Map;
import java.util.StringTokenizer;


/**
 * An implementation of SerializableCacheWrapper that uses EHCache as an underlying implementation.
 *
 * Pass in a -Dbind.address=IP_ADDRESS
 *
 * @author Manik Surtani (manik@surtani.org)
 * @version $Id: EHCacheWrapper.java,v 1.6 2007/05/21 16:17:56 msurtani Exp $
 */
public class EHCacheWrapper implements SerializableCacheWrapper
{
   private CacheManager manager;
   private Ehcache cache;
   private Log logger = LogFactory.getLog("org.cachebench.cachewrappers.EHCacheWrapper");

   /* (non-Javadoc)
   * @see org.cachebench.CacheWrapper#init(java.util.Properties)
   */
   public void init(Map parameters) throws Exception
   {
      logger.debug("Initializing the cache with props " + parameters);
      URL url = getClass().getClassLoader().getResource((String) parameters.get("config"));
      logger.debug("Config URL = " + url);
      Configuration c = ConfigurationFactory.parseConfiguration(url);
      c.setSource("URL of " + url);

      // frig the manager to use a bind address as per sys props
      String bindAddress = System.getProperty("bind.address");

      if (bindAddress!= null)
      {
         String props = c.getCacheManagerPeerListenerFactoryConfiguration().getProperties();
         props = injectBindAddress(props, bindAddress);
         c.getCacheManagerPeerListenerFactoryConfiguration().setProperties(props);
      }

      manager = new CacheManager(c);
      
      logger.debug("Finish Initializing the cache");
   }

   private String injectBindAddress(String props, String bindAddress)
   {
      String newProps = "hostName=" + bindAddress;

      StringTokenizer st = new StringTokenizer(props, ",", false);
      while (st.hasMoreTokens())
      {
         boolean skip = false;
         String p = st.nextToken();
         if (p != null)
         {
            p.trim();
            String[] kv = p.split("=");
            if (kv[0] != null)
            {
               kv[0].trim();
               if (kv[0].equals("hostName"))
                  skip = true;
               else
                  newProps += ", " + kv[0];
            }
            if (kv[1] != null && !skip)
            {
               kv[1].trim();
               newProps += "=" + kv[1];
            }

         }
      }

      return newProps;
   }

   /* (non-Javadoc)
   * @see org.cachebench.CacheWrapper#setUp()
   */
   public void setUp() throws Exception
   {
      logger.info("Caches avbl:");
      for (String s : manager.getCacheNames()) logger.info("    * " + s);
      cache = manager.getCache("cache");
      logger.info("Using named cache " + cache);
      logger.info("Bounded peers: " + manager.getCachePeerListener().getBoundCachePeers());
      logger.info("Remote peers: " + manager.getCacheManagerPeerProvider().listRemoteCachePeers(cache));
   }

   /* (non-Javadoc)
   * @see org.cachebench.CacheWrapper#tearDown()
   */
   public void tearDown() throws Exception
   {
      manager.shutdown();
   }

   /* (non-Javadoc)
   * @see org.cachebench.SerializableCacheWrapper#putSerializable(java.io.Serializable, java.io.Serializable)
   */
   public void putSerializable(Serializable key, Serializable value) throws Exception
   {
      Element element = new Element(key, value);
      cache.put(element);
   }

   /* (non-Javadoc)
   * @see org.cachebench.SerializableCacheWrapper#getSerializable(java.io.Serializable)
   */
   public Object getSerializable(Serializable key) throws Exception
   {
      return cache.get(key);
   }

   public void empty() throws Exception
   {
      cache.removeAll();
   }

   /* (non-Javadoc)
   * @see org.cachebench.CacheWrapper#put(java.lang.Object, java.lang.Object)
   */
   public void put(Object key, Object value) throws Exception
   {
      throw new Exception("Put: Does not deal with non-Serializables, should not be called.");
   }

   /* (non-Javadoc)
   * @see org.cachebench.CacheWrapper#get(java.lang.Object)
   */
   public Object get(Object key) throws Exception
   {
      throw new Exception("Get: Does not deal with non-Serializables, should not be called.");
   }

   public int getNumMembers()
   {
      return manager.getCacheManagerPeerProvider().listRemoteCachePeers(cache).size();
   }

   public String getInfo()
   {
      return cache.getKeys().toString() + " remote peers: " + manager.getCachePeerListener().getBoundCachePeers();
   }

   public Object getReplicatedData(String key) throws Exception
   {
      return get(key);
   }
}
