/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.config.EvictionPolicyConfig;
import org.jboss.cache.config.EvictionRegionConfig;
import org.jboss.cache.eviction.EvictedEventNode;
import org.jboss.cache.eviction.EvictionPolicy;
import org.jboss.cache.eviction.NodeEventType;
import org.jboss.cache.util.Util;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

/**
 * Default implementation of a {@link Region}
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
public class RegionImpl implements Region
{
   private static final Log log = LogFactory.getLog(RegionImpl.class);

   private RegionManager regionManager;
   private Fqn fqn;
   private boolean active;
   private ClassLoader classLoader;
   private BlockingQueue<EvictedEventNode> nodeEventQueue = null;
   private int capacityWarnThreshold = 0;
   private EvictionRegionConfig configuration = new EvictionRegionConfig();
   private EvictionPolicy policy;

   /**
    * Constructs a marshalling region from an fqn and region manager.
    */
   public RegionImpl(Fqn fqn, RegionManager regionManager)
   {
      this.fqn = fqn;
      this.regionManager = regionManager;
      this.active = !regionManager.isDefaultInactive();
   }

   /**
    * Constructs an eviction region from a policy and configuration, defined by an fqn and region manager.
    */
   public RegionImpl(EvictionPolicy policy, EvictionRegionConfig config, Fqn fqn, RegionManager regionManager)
   {
      this(fqn, regionManager);
      this.configuration = config;
      this.policy = policy;
      createQueue();
   }

   public void registerContextClassLoader(ClassLoader classLoader)
   {
      this.classLoader = classLoader;
   }

   public void unregisterContextClassLoader()
   {
      this.classLoader = null;
   }

   public void activate()
   {
      regionManager.activate(fqn);
      active = true;
   }

   public void activateIfEmpty()
   {
      regionManager.activateIfEmpty(fqn);
      active = true;
   }

   public void deactivate()
   {
      regionManager.deactivate(fqn);
      active = false;
   }

   public boolean isActive()
   {
      return active;
   }

   public ClassLoader getClassLoader()
   {
      return classLoader;
   }

   public Fqn getFqn()
   {
      return fqn;
   }

   public void setActive(boolean b)
   {
      active = b;
   }

   // -------- eviction stuff -----

   public void markNodeCurrentlyInUse(Fqn fqn, long timeout)
   {
      EvictedEventNode markUse = new EvictedEventNode(fqn, NodeEventType.MARK_IN_USE_EVENT);
      markUse.setInUseTimeout(timeout);
      putNodeEvent(markUse);
   }

   public void unmarkNodeCurrentlyInUse(Fqn fqn)
   {
      EvictedEventNode markNoUse = new EvictedEventNode(fqn, NodeEventType.UNMARK_USE_EVENT);
      putNodeEvent(markNoUse);
   }

   public String toString()
   {
      return "RegionImpl{" +
              "fqn=" + fqn +
              "; classloader="+ classLoader +
              "; active=" + active +
              "; eviction=" + (getEvictionPolicy() != null) +
              "; timerThreadRegistered=" + (getEvictionPolicy() != null && regionManager.getEvictionTimerTask().isRegionRegisteredForProcessing(this)) +
              '}';
   }

   public int compareTo(Region other)
   {
      return getFqn().compareTo(other.getFqn());
   }

   public void putNodeEvent(EvictedEventNode event)
   {
      try
      {
         if (nodeEventQueue == null) createQueue();// in case the queue does not exist yet.
         if (nodeEventQueue.size() > capacityWarnThreshold)
         {
            log.warn("putNodeEvent(): eviction node event queue size is at 98% threshold value of capacity: " + configuration.getEventQueueSize() +
                    " Region: " + fqn +
                    " You will need to reduce the wakeUpIntervalSeconds parameter.");
         }

         nodeEventQueue.put(event);
      }
      catch (InterruptedException e)
      {
         log.debug("give up put", e);
      }
   }

   public EvictedEventNode takeLastEventNode()
   {
      try
      {
         return nodeEventQueue.poll(0, TimeUnit.SECONDS);
      }
      catch (InterruptedException e)
      {
         log.debug("trace", e);
      }
      return null;
   }

   public int nodeEventQueueSize()
   {
      return nodeEventQueue.size();
   }

   public void resetEvictionQueues()
   {
      nodeEventQueue.clear();
   }

   private synchronized void createQueue()
   {
      if (nodeEventQueue == null)
      {
         if (configuration == null)
         {
            throw new IllegalArgumentException("null eviction configuration");
         }
         int size = configuration.getEventQueueSize();
         capacityWarnThreshold = (98 * size) / 100 - 100;
         if (capacityWarnThreshold <= 0)
         {
            throw new RuntimeException("Capacity warn threshold used in eviction is smaller than 1.");
         }
         nodeEventQueue = new LinkedBlockingQueue<EvictedEventNode>(size);
      }
   }

   public EvictionRegionConfig getEvictionRegionConfig()
   {
      return this.configuration;
   }

   public EvictionPolicyConfig getEvictionPolicyConfig()
   {
      return configuration == null ? null : configuration.getEvictionPolicyConfig();
   }

   public EvictionPolicy getEvictionPolicy()
   {
      return policy;
   }

   public void setEvictionPolicy(EvictionPolicyConfig evictionPolicyConfig)
   {
      configuration.setEvictionPolicyConfig(evictionPolicyConfig);
      policy = createPolicy(evictionPolicyConfig.getEvictionPolicyClass());
      regionManager.getEvictionTimerTask().addRegionToProcess(this);
      if (nodeEventQueue == null) createQueue();
   }

   private EvictionPolicy createPolicy(String className)
   {
      if (className == null)
      {
         throw new IllegalArgumentException("null className");
      }
      try
      {
         if (log.isTraceEnabled()) log.trace("Instantiating " + className);
         EvictionPolicy ep = (EvictionPolicy) Util.loadClass(className).newInstance();
         ep.setCache(regionManager.getCache());
         return ep;
      }
      catch (Exception e)
      {
         log.fatal("Unable to instantiate eviction policy class " + className, e);
         return null;
      }
   }
}
