/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.lock.IdentityLock;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.marshall.MethodCallFactory;
import org.jboss.cache.marshall.MethodDeclarations;
import org.jboss.cache.optimistic.DataVersion;
import org.jboss.cache.transaction.GlobalTransaction;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Basic data node class.  Throws {@link UnsupportedOperationException} for version-specific methods like {@link #getVersion()} and
 * {@link #setVersion(org.jboss.cache.optimistic.DataVersion)}, defined in {@link org.jboss.cache.NodeSPI}.
 *
 * @author Manik Surtani (<a href="mailto:manik@jboss.org">manik@jboss.org</a>)
 * @since 2.0.0
 */
public class UnversionedNode<K, V> extends AbstractNode<K, V> implements NodeSPI<K, V>
{

   /**
    * Default output indent for printing.
    */
   private static final int INDENT = 4;

   /**
    * Debug log.
    */
   private static Log log = LogFactory.getLog(UnversionedNode.class);

   /**
    * True if all children have been loaded. This is set when CacheImpl.getChildrenNames() is called.
    */
   private boolean childrenLoaded = false;

   /**
    * True if data has been loaded from the cache loader.
    */
   private boolean dataLoaded = true;

   /**
    * Lock manager that manages locks to be acquired when accessing the node inside a transaction. Lazy set just in case
    * locking is not needed.
    */
   private transient IdentityLock lock_ = null;

   /**
    * A reference of the CacheImpl instance.
    */
   private transient CacheImpl<K, V> cache;

   /**
    * Map of general data keys to values.
    */
   private final Map<K, V> data = new HashMap<K, V>();

   /**
    * Constructs a new node with an FQN of Root.
    */
   public UnversionedNode()
   {
      this.fqn = Fqn.ROOT;
   }

   /**
    * Constructs a new node with a name, etc.
    *
    * @param mapSafe <code>true</code> if param <code>data</code> can safely be directly assigned to this object's
    *                {@link #data} field; <code>false</code> if param <code>data</code>'s contents should be copied into
    *                this object's {@link #data} field.
    */
   protected UnversionedNode(Object child_name, Fqn fqn, Map<K, V> data, boolean mapSafe, CacheSPI<K, V> cache)
   {
      init(child_name, fqn, cache);
      if (data != null)
      {
         this.data.putAll(data);// ? is this safe
      }
   }

   /**
    * Initializes with a name and FQN and cache.
    */
   private void init(Object child_name, Fqn fqn, CacheSPI<K, V> cache)
   {
      if (cache == null)
      {
         throw new IllegalArgumentException("no cache init for " + fqn);
      }
      this.cache = (CacheImpl) cache;
      this.fqn = fqn;
      if (!fqn.isRoot() && !child_name.equals(fqn.getLastElement()))
      {
         throw new IllegalArgumentException("Child " + child_name + " must be last part of " + fqn);
      }
   }

   /**
    * Returns a parent by checking the TreeMap by name.
    */
   public NodeSPI<K, V> getParent()
   {
      if (fqn.isRoot())
      {
         return null;
      }
      return cache.peek(fqn.getParent(), true);
   }

   private synchronized void initLock()
   {
      if (lock_ == null)
      {
         lock_ = new IdentityLock(cache.getConfiguration().getIsolationLevel(), this);
      }
   }

   private synchronized Map<Object, Node<K, V>> children()
   {
      if (children == null)
      {
         if (getFqn().isRoot())
         {
            children = new ConcurrentHashMap<Object, Node<K, V>>(64, .5f, 16);
         }
         else
         {
            // Less segments to save memory
            children = new ConcurrentHashMap<Object, Node<K, V>>(4, .75f, 4);
         }
      }
      return children;
   }

   public CacheSPI<K, V> getCache()
   {
      return cache;
   }

   public boolean isChildrenLoaded()
   {
      return childrenLoaded;
   }

   public void setChildrenLoaded(boolean flag)
   {
      childrenLoaded = flag;
   }

   public V get(K key)
   {
      return cache.get(getFqn(), key);
   }

   public V getDirect(K key)
   {
      return data == null ? null : data.get(key);
   }


   private boolean isReadLocked()
   {
      return lock_ != null && lock_.isReadLocked();
   }

   private boolean isWriteLocked()
   {
      return lock_ != null && lock_.isWriteLocked();
   }

   public IdentityLock getLock()
   {
      initLock();
      return lock_;
   }

   public Map<K, V> getData()
   {
      if (cache == null) return Collections.emptyMap();
      return cache.getData(getFqn());

/*
      Map<K, V> dMap = new HashMap<K, V>();
      for (K k : cache.getKeys(getFqn()))
      {
         dMap.put(k, cache.get(fqn, k));
      }
      return Collections.unmodifiableMap(dMap);
*/
   }

   public Map<K, V> getDataDirect()
   {
      if (data == null) return Collections.emptyMap();
      return Collections.unmodifiableMap(data);
   }

   public V put(K key, V value)
   {
      return cache.put(getFqn(), key, value);
   }

   public V putDirect(K key, V value)
   {
      return data.put(key, value);
   }

   public NodeSPI getOrCreateChild(Object child_name, GlobalTransaction gtx)
   {
      return getOrCreateChild(child_name, gtx, true);
   }

   private NodeSPI getOrCreateChild(Object child_name, GlobalTransaction gtx, boolean createIfNotExists)
   {

      NodeSPI child;
      if (child_name == null)
      {
         throw new IllegalArgumentException("null child name");
      }

      child = (NodeSPI) children().get(child_name);
      InvocationContext ctx = cache.getInvocationContext();
      if (createIfNotExists && child == null)
      {
         // construct the new child outside the synchronized block to avoid
         // spending any more time than necessary in the synchronized section
         Fqn child_fqn = new Fqn(this.fqn, child_name);
         NodeSPI newChild = (NodeSPI) cache.getConfiguration().getRuntimeConfig().getNodeFactory().createNode(child_name, this, null);
         if (newChild == null)
         {
            throw new IllegalStateException();
         }
         synchronized (this)
         {
            // check again to see if the child exists
            // after acquiring exclusive lock
            child = (NodeSPI) children().get(child_name);
            if (child == null)
            {
               cache.getNotifier().notifyNodeCreated(child_fqn, true, ctx);
               child = newChild;
               children.put(child_name, child);
               if (gtx != null)
               {
                  MethodCall undo_op = MethodCallFactory.create(MethodDeclarations.removeNodeMethodLocal, gtx,
                                                                child_fqn, false);
                  cache.addUndoOperation(gtx, undo_op);
                  // add the node name to the list maintained for the current tx
                  // (needed for abort/rollback of transaction)
                  // cache.addNode(gtx, child.getFqn());
               }
            }
         }

         // notify if we actually created a new child
         if (newChild == child)
         {
            if (log.isTraceEnabled())
            {
               log.trace("created child: fqn=" + child_fqn);
            }
            cache.getNotifier().notifyNodeCreated(child_fqn, false, ctx);
         }
      }
      return child;

   }

   public V remove(K key)
   {
      return cache.remove(getFqn(), key);
   }

   public V removeDirect(K key)
   {
      if (data == null) return null;
      return data.remove(key);
   }

   public void printDetails(StringBuffer sb, int indent)
   {
      printDetailsInMap(sb, indent);
   }

   /**
    * Returns a debug string.
    */
   @Override
   public String toString()
   {
      StringBuffer sb = new StringBuffer();
      sb.append(getClass().getSimpleName());
      if (deleted)
      {
         sb.append(" (deleted) [ ").append(fqn);
      }
      else
      {
         sb.append("[ ").append(fqn);
      }
      if (data != null)
      {
         synchronized (data)
         {
            sb.append(" data=").append(data.keySet());
         }
      }
      if (children != null && !children.isEmpty())
      {
         sb.append(" child=").append(getChildrenNamesDirect());
      }
      if (lock_ != null)
      {
         if (isReadLocked())
         {
            sb.append(" RL");
         }
         if (isWriteLocked())
         {
            sb.append(" WL");
         }
      }
      sb.append("]");
      return sb.toString();
   }

   public Node<K, V> addChild(Fqn f)
   {
      Fqn nf = new Fqn(getFqn(), f);
      cache.put(nf, null);
      return getChild(f);
   }

   public void addChildDirect(NodeSPI<K, V> child)
   {
      if (child.getFqn().getParent().equals(getFqn()))
      {
         synchronized (this)
         {
            children().put(child.getFqn().getLastElement(), child);
         }
      }
      else
         throw new CacheException("Attempting to add a child [" + child.getFqn() + "] to [" + getFqn() + "].  Can only add direct children.");
   }

   public NodeSPI<K, V> addChildDirect(Fqn f)
   {
      if (f.size() == 1)
      {
         GlobalTransaction gtx = cache.getInvocationContext().getGlobalTransaction();
         return getOrCreateChild(f.getLastElement(), gtx);
      }
      else
      {
         throw new UnsupportedOperationException("Cannot directly create children which aren't directly under the current node.");
      }

   }

   public void clearData()
   {
      cache.removeData(getFqn());
   }

   public void clearDataDirect()
   {
      if (data != null) data.clear();
   }

   public Node<K, V> getChild(Fqn fqn)
   {
      return cache.get(new Fqn(getFqn(), fqn));
   }

   public NodeSPI<K, V> getChildDirect(Fqn fqn)
   {
      if (fqn.size() == 1)
      {
         return getChildDirect(fqn.getLastElement());
      }
      else
      {
         NodeSPI currentNode = this;
         for (int i = 0; i < fqn.size(); i++)
         {
            Object nextChildName = fqn.get(i);
            currentNode = currentNode.getChildDirect(nextChildName);
            if (currentNode == null) return null;
         }
         return currentNode;
      }
   }

   public Set<Object> getChildrenNames()
   {
      return cache.getChildrenNames(getFqn());
   }

   public Set<Object> getChildrenNamesDirect()
   {
      return children == null ? Collections.emptySet() : new HashSet(children.keySet());
      //return childrenNames();
   }

   public Set<K> getKeys()
   {
      Set<K> keys = cache.getKeys(getFqn());
      return (Set<K>) (keys == null ? Collections.emptySet() : Collections.unmodifiableSet(keys));
   }

   public Set<K> getKeysDirect()
   {
      if (data == null)
      {
         return Collections.emptySet();
      }
      return Collections.unmodifiableSet(new HashSet<K>(data.keySet()));
   }

   public boolean hasChild(Fqn f)
   {
      return getChild(f) != null;
   }

   public boolean hasChild(Object o)
   {
      return getChild(o) != null;
   }

   public V putIfAbsent(K k, V v)
   {
      // make sure this is atomic.  Not hugely performant at the moment (should use the locking interceptors) but for now ...
      synchronized (this)
      {
         if (!getKeys().contains(k))
            return put(k, v);
         else
            return get(k);
      }
   }

   public V replace(K key, V value)
   {
      // make sure this is atomic.  Not hugely performant at the moment (should use the locking interceptors) but for now ...
      synchronized (this)
      {
         if (getKeys().contains(key))
         {
            return put(key, value);
         }
         else
            return null;
      }
   }

   public boolean replace(K key, V oldValue, V newValue)
   {
      // make sure this is atomic.  Not hugely performant at the moment (should use the locking interceptors) but for now ...
      synchronized (this)
      {
         if (oldValue.equals(get(key)))
         {
            put(key, newValue);
            return true;
         }
         else
            return false;
      }
   }

   public boolean removeChild(Fqn fqn)
   {
      return cache.removeNode(new Fqn(getFqn(), fqn));
   }

   public int dataSize()
   {
      return cache.getKeys(getFqn()).size();
   }

   public boolean removeChild(Object childName)
   {
      return removeChild(new Fqn(getFqn(), childName));
   }

   public boolean removeChildDirect(Object childName)
   {
      return children != null && children.remove(childName) != null;
   }

   public boolean removeChildDirect(Fqn f)
   {
      if (f.size() == 1)
      {
         return removeChildDirect(f.getLastElement());
      }
      else
      {
         NodeSPI child = getChildDirect(f);
         return child != null && child.getParent().removeChildDirect(f.getLastElement());
      }
   }

   public Map<Object, Node<K, V>> getChildrenMapDirect()
   {
      return children;
   }

   public void setChildrenMapDirect(Map<Object, Node<K, V>> children)
   {
      this.children().clear();
      this.children.putAll(children);
   }

   public void putAll(Map data)
   {
      cache.put(fqn, data);
   }

   public void replaceAll(Map data)
   {
      cache.put(fqn, data, true);
   }

   public void putAllDirect(Map<K, V> data)
   {
      if (data == null) return;
      this.data.putAll(data);
   }

   public void removeChildrenDirect()
   {
      if (children != null)
      {
         children.clear();
      }
      children = null;
   }

   public void print(StringBuffer sb, int indent)
   {
      printIndent(sb, indent);
      sb.append(Fqn.SEPARATOR).append(getName()).append(" ").append(getDataDirect().size());
      if (children != null)
      {
         for (Node node : children.values())
         {
            sb.append("\n");
            ((NodeSPI) node).print(sb, indent + INDENT);
         }
      }
   }

   // versioning

   public void setVersion(DataVersion version)
   {
      throw new UnsupportedOperationException("Versioning not supported");
   }

   public DataVersion getVersion()
   {
      throw new UnsupportedOperationException("Versioning not supported");
   }

   private void printIndent(StringBuffer sb, int indent)
   {
      if (sb != null)
      {
         for (int i = 0; i < indent; i++)
         {
            sb.append(" ");
         }
      }
   }

   public void addChild(Object child_name, Node<K, V> n)
   {
      if (child_name != null)
      {
         children().put(child_name, n);
      }
   }

   /**
    * Returns the name of this node.
    */
   private Object getName()
   {
      return fqn.getLastElement();
   }

   /**
    * Returns the name of this node.
    */
   public Fqn getFqn()
   {
      return fqn;
   }

   public void setFqn(Fqn fqn)
   {
      if (log.isTraceEnabled())
      {
         log.trace(getFqn() + " set FQN " + fqn);
      }
      this.fqn = fqn;

      if (children == null)
      {
         return;
      }

      // process children
      for (Map.Entry<Object, ? extends Node> me : children.entrySet())
      {
         NodeSPI n = (NodeSPI) me.getValue();
         Fqn cfqn = new Fqn(fqn, me.getKey());
         n.setFqn(cfqn);
      }
   }

   public Node<K, V> getChild(Object childName)
   {
      return cache.get(new Fqn(getFqn(), childName));
   }

   public NodeSPI<K, V> getChildDirect(Object childName)
   {
      if (childName == null) return null;
      return (NodeSPI<K, V>) (children == null ? null : children.get(childName));
   }

   public Set<Node<K, V>> getChildren()
   {
      if (cache == null) return Collections.emptySet();
      Set<Node<K, V>> children = new HashSet<Node<K, V>>();
      for (Object c : cache.getChildrenNames(getFqn()))
      {
         Node n = cache.get(new Fqn(getFqn(), c));
         if (n != null) children.add(n);
      }
      return Collections.unmodifiableSet(children);
   }

   public Set<NodeSPI<K, V>> getChildrenDirect()
   {
      // strip out deleted child nodes...
      if (children == null || children.size() == 0) return Collections.emptySet();

      Set<NodeSPI<K, V>> exclDeleted = new HashSet<NodeSPI<K, V>>();
      for (Node n : children.values())
      {
         NodeSPI spi = (NodeSPI) n;
         if (!spi.isDeleted()) exclDeleted.add(spi);
      }
      return Collections.unmodifiableSet(exclDeleted);
   }

   public boolean hasChildrenDirect()
   {
      return children != null && children.size() != 0;
   }

   public Set<NodeSPI<K, V>> getChildrenDirect(boolean includeMarkedForRemoval)
   {
      if (includeMarkedForRemoval)
      {
         if (children != null && !children.isEmpty())
         {
            return Collections.unmodifiableSet(new HashSet(children.values()));
         }
         else
         {
            return Collections.emptySet();
         }
      }
      else
      {
         return getChildrenDirect();
      }
   }

   /**
    * Adds details of the node into a map as strings.
    */
   private void printDetailsInMap(StringBuffer sb, int indent)
   {
      printIndent(sb, indent);
      indent += 2;// increse it
      if (!(getFqn()).isRoot())
      {
         sb.append(Fqn.SEPARATOR);
      }
      sb.append(getName());
      sb.append("  ");
      sb.append(data);
      if (children != null)
      {
         for (Node n : children.values())
         {
            sb.append("\n");
            ((NodeSPI) n).printDetails(sb, indent);
         }
      }
   }

   /**
    * Returns true if the data was loaded from the cache loader.
    */
   public boolean isDataLoaded()
   {
      return dataLoaded;
   }

   /**
    * Sets if the data was loaded from the cache loader.
    */
   public void setDataLoaded(boolean dataLoaded)
   {
      this.dataLoaded = dataLoaded;
   }

   public boolean isValid()
   {
      // TODO; implement this property, to detect if it has been evicted, removed by another thread, etc.  Method added for now as a dummy so it exists in the API
      return true;
   }
}
