package org.cachebench.tests;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.cachebench.CacheWrapper;
import org.cachebench.config.Configuration;
import org.cachebench.config.TestCase;
import org.cachebench.config.TestConfig;
import org.cachebench.tests.results.TestResult;

import java.util.Random;
import java.util.Date;

/**
 * Simulates work with a web session.
 * 
 * @author Mircea.Markus@jboss.com
 * @since 2.2
 */
public class SessionSimulatorTest implements CacheTest
{
   private static final Log log = LogFactory.getLog(SessionSimulatorTest.class);

   public static final String SESSION_PREFIX = "SESSION_";
   private static int LOG_AFTER_OPERATION_COUNT = 5000;
   Configuration configuration;
   private TestConfig thisTestConfig;
   private String sessionId;
   private CacheWrapper cacheWrapper;


   /**
    * total number of request to be made against this session: reads + writes
    */
   private int numberOfRequests;

   /**
    * for each session there will be created fixed number of attributes. On those attributes all the GETs and PUTs
    * are performed (for PUT is overwrite)
    */
   private int numberOfAttributes;

   /**
    * Each attribute will be a byte[] of this size
    */
   private int sizeOfAnAttribute;

   /**
    * Out of the total number of request, this define the frequency of writes (percentage)
    */
   private int writePercentage;

   public void setConfiguration(Configuration configuration)
   {
      this.configuration = configuration;
   }

   public TestResult doTest(String testName, CacheWrapper cache, String testCaseName, int sampleSize, int numThreads) throws Exception
   {
      this.cacheWrapper = cache;
      TestCase testCase = configuration.getTestCase(testCaseName);
      thisTestConfig = testCase.getTest(testName);
      readParams();
      initSession();
      return stressSession(testName, testCaseName);
   }

   private SessionSimulatorTestResult stressSession(String testName, String testCaseName) throws Exception
   {

      long totalBytesRead = 0;
      long totalBytesWritten = 0;
      long reads = 0;
      long writes = 0;
      int readPercentage = 100 - writePercentage;
      Random r = new Random();
      int randomAction;
      int randomAttribute;
      long start=System.currentTimeMillis();
      for (int i = 0; i < numberOfRequests; i++)
      {
         logRunCount(i);
         randomAction = r.nextInt(100);
         randomAttribute = r.nextInt(numberOfAttributes - 1);
         byte[] buf;
         if (randomAction < readPercentage)
         { // read
            try
            {
               buf = (byte[]) cacheWrapper.get(getSessionEntry(randomAttribute));
               totalBytesRead += buf.length;
               reads++;
            } catch (Exception e)
            {
               log.trace("Error appeared whilst reading from cache:" + e.getMessage());
            }
         } else
         {             // write
            buf = new byte[this.sizeOfAnAttribute];
            try
            {
               cacheWrapper.put(getSessionEntry(randomAttribute), buf);
               totalBytesWritten += buf.length;
               writes++;
            } catch (Exception e)
            {
               log.trace("Error appeared whilst writing to cache:" + e.getMessage());
            }
         }
      }
      long durration = System.currentTimeMillis() - start;
      SessionSimulatorTestResult result = new SessionSimulatorTestResult(reads, writes, durration, totalBytesRead, totalBytesWritten);
      result.setTestPassed(true);
      result.setTestName(testCaseName + getNodeIndex());
      result.setTestTime(new Date());
      result.setTestType(testName);
      log.trace("Returning result:" + result);
      return result;
   }

   private void logRunCount(int i)
   {
      if (i % LOG_AFTER_OPERATION_COUNT == 0)
      {
         log.info("SessionSimulatorTest performed " + i + " oprations");
      }
   }

   private void initSession() throws Exception
   {
      for (int i = 0; i < numberOfAttributes; i++)
      {
         cacheWrapper.put(getSessionEntry(i), new byte[sizeOfAnAttribute]);
      }
   }

   private String getSessionEntry(int i)
   {
      return sessionId + i;
   }

   private void readParams ()
   {
      sessionId = SESSION_PREFIX + configuration.getClusterConfig().getCurrentNodeIndex();
      log.debug("Session id is: " + sessionId);
      numberOfRequests = thisTestConfig.getIntValue("numberOfRequest");
      numberOfAttributes = thisTestConfig.getIntValue("numberOfAttributes");
      writePercentage = thisTestConfig.getIntValue("writePercentage");
      sizeOfAnAttribute = thisTestConfig.getIntValue("sizeOfAnAttribute");
      log.debug("recieved follosing params[ numberOfRequests=" + numberOfRequests + ", numberOfAttributes="
            + numberOfAttributes + ", writePercentage=" + writePercentage + ", sizeOfAnAttribute=" + sizeOfAnAttribute + " ]");
   }

   public String getNodeIndex()
   {
      return configuration.getClusterConfig().getCurrentNodeIndex() + "";
   }
}
