package org.cachebench.tests.simpletests;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.commons.math.stat.descriptive.DescriptiveStatistics;
import org.cachebench.CacheWrapper;
import org.cachebench.SerializableCacheWrapper;
import org.cachebench.tests.results.StatisticTestResult;
import org.cachebench.tests.CacheTest;
import org.cachebench.config.Configuration;

import java.io.Serializable;
import java.io.ByteArrayOutputStream;
import java.io.ObjectOutputStream;
import java.util.Date;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.atomic.AtomicLong;


/**
 *
 *
 * @author Manik Surtani (manik@surtani.org)
 *         (C) Manik Surtani, 2004
 */
public abstract class SimpleTest implements CacheTest
{
   protected Log log = LogFactory.getLog(this.getClass());
   protected AtomicLong numberOfBytesPut = new AtomicLong(0);
   protected long elapsedSecondsForAllPuts = 0;
   protected ExecutorService executor;
   private static final int EXECUTOR_SHUTDOWN_TIMEOUT_POLL_SECS = 60;
   protected Configuration configuration;
   protected final int LOG_FREQUENCY = 5000;

   protected StatisticTestResult performTestWithObjectType(String testCaseName, CacheWrapper cache, Class valueClass, String testName, int sampleSize, int numThreads) throws Exception
   {
      log.info("Number of threads " + numThreads);
      executor = Executors.newFixedThreadPool(numThreads);

      StatisticTestResult result = new StatisticTestResult();
      result.setTestName(testCaseName + getNodeIndex());
      result.setTestTime(new Date());
      result.setTestType(testName);

      log.info("Performing '" + sampleSize + "' PUTs");
      DescriptiveStatistics putStats = doPuts(cache, valueClass, sampleSize);
      executor = Executors.newFixedThreadPool(numThreads);
      log.info("Performing GETs");
      DescriptiveStatistics getStats = doGets(cache, sampleSize);

      result.setGetData(getStats);
      result.setPutData(putStats);
      result.setTestPassed(true); // The test is passed. The report would make use of this attribute.

      // calculate throughput, in transactions per second.
      // we only measure put operations for throughput.

      // calc tps.

      System.out.println("*** sum of time: " + elapsedSecondsForAllPuts);
      System.out.println("*** num puts occured: " + putStats.getN());

      try
      {
         result.setThroughputTransactionsPerSecond((int) (sampleSize / elapsedSecondsForAllPuts));
      }
      catch (ArithmeticException ae)
      {
         log.warn("Divide by 0 - elapsedSecondsForAllPuts = 0?");
         result.setThroughputTransactionsPerSecond(-999);
      }

      try
      {
         result.setThroughputBytesPerSecond((int) (numberOfBytesPut.longValue() / elapsedSecondsForAllPuts));
      }
      catch (ArithmeticException ae)
      {
         log.warn("Divide by 0 - elapsedSecondsForAllPuts = 0?");
         result.setThroughputBytesPerSecond(-999);
      }


      // set the number of members in the cluster
      result.setNumMembers(cache.getNumMembers());
      result.setNumThreads(numThreads);

      return result;
   }

   /**
    * @param cache      The Cachewrapper on which the bechmark is conducted.
    * @param sampleSize The size of the cache.
    * @return The Descriptive statistics of the cache benchmarking.
    */
   private DescriptiveStatistics doGets(final CacheWrapper cache, int sampleSize) throws Exception
   {
      log.debug("Inside doGets for : " + cache);
      final String key = "baseKey";
      final DescriptiveStatistics stats = DescriptiveStatistics.newInstance();
      final boolean useSerializable = cache instanceof SerializableCacheWrapper;

      for (int i = 0; i < sampleSize; i++)
      {
         final int cycleNumber = i;
         Runnable r = new Runnable()
         {
            public void run()
            {
               try
               {
                  if (!useSerializable)
                  {
                     // start your timer...
                     long startTime = System.currentTimeMillis();
                     cache.get(key + cycleNumber);
                     long statValue = (System.currentTimeMillis() - startTime);
                     stats.addValue(statValue);
                     log.debug("The Get stat : " + statValue);
                  }
                  else
                  {
                     SerializableCacheWrapper sCache = (SerializableCacheWrapper) cache;
                     long startTime = System.currentTimeMillis();
                     sCache.getSerializable(key + cycleNumber);
                     long statValue = (System.currentTimeMillis() - startTime);
                     stats.addValue(statValue);
                     log.debug("The Get stat : " + statValue);
                  }
                  logOperation(cycleNumber, "GETS");
               }
               catch (Exception e)
               {
                  // how should we handle this?  Log for now...
                  log.error("Operation failed!", e);
               }
            }
         };

         // submit task to be executed
         executor.execute(r);
      }

      // only leave once the task queue is empty!!
      blockTillTasksComplete();

      // return the raw data
      log.debug("Leaving doGets for : " + cache);
      return stats;
   }

   private void logOperation(int i, String s)
   {
      if ((i + 1) % LOG_FREQUENCY == 0)
      {
         log.info((i + 1) + " " + s + " were performed");
      }
   }

   /**
    * @param cache      The Cachewrapper on which the bechmark is conducted.
    * @param valueClass The value class instance which is going to be inserted into the Cache.
    * @param sampleSize The size of the cache.
    * @return The Descriptive statistics of the cache benchmarking.
    */
   private DescriptiveStatistics doPuts(final CacheWrapper cache, final Class valueClass, int sampleSize) throws Exception
   {
      log.debug("Inside doPuts for " + cache);
      final String key = "baseKey";
      final DescriptiveStatistics stats = DescriptiveStatistics.newInstance();
      final boolean useSerializable = cache instanceof SerializableCacheWrapper;
      numberOfBytesPut.set(0);
      elapsedSecondsForAllPuts = 0;
      long startElapsedTime = System.currentTimeMillis();

      for (int i = 0; i < sampleSize; i++)
      {
         final int cycleNumber = i;
         Runnable r = new Runnable()
         {
            public void run()
            {
               try
               {
                  // generate some value
                  Object value;
                  if (valueClass == null) value = null;
                  else if (valueClass.getName().equals("java.lang.String")) value = "value" + cycleNumber;
                  else if (valueClass.getName().equals("java.lang.Integer")) value = cycleNumber;
                  else value = valueClass.newInstance();

                  // even though some impls may use special marshalling to reduce the amount of data transmitted (such as JBoss Cache's
                  // CacheMarshaller) we still want to measure the serialized size of objects for this metric.

                  numberOfBytesPut.getAndAdd(calculateSerializedSize(value));

                  if (!useSerializable)
                  {
                     // start your timer...
                     long startTime = System.currentTimeMillis();
                     cache.put(key + cycleNumber, value);
                     long statValue = (System.currentTimeMillis() - startTime);
                     stats.addValue(statValue);
                  }
                  else
                  {
                     SerializableCacheWrapper sCache = (SerializableCacheWrapper) cache;
                     Serializable sValue = (Serializable) value;
                     long startTime = System.currentTimeMillis();
                     sCache.putSerializable(key + cycleNumber, sValue);
                     long statValue = (System.currentTimeMillis() - startTime);
                     stats.addValue(statValue);
                  }
                  logOperation(cycleNumber,"PUTS");
               }
               catch (Exception e)
               {
                  // how should we handle this?  Log for now...
                  log.error("Operation failed!", e);
               }
            }
         };
         // submit task to be executed
         executor.execute(r);
      }

      // only leave once the task queue is empty!!
      blockTillTasksComplete();
      elapsedSecondsForAllPuts = (System.currentTimeMillis() - startElapsedTime) / 1000;

      // return the raw data
      log.debug("Leaving doPuts for " + cache);
      return stats;
   }

   private void blockTillTasksComplete()
   {
      // now that just told us that all the tasks have been submitted.  Lets check that the executor has finished everything.
      executor.shutdown();
      while (!executor.isTerminated())
      {
         try
         {
            executor.awaitTermination(EXECUTOR_SHUTDOWN_TIMEOUT_POLL_SECS, TimeUnit.SECONDS);
         }
         catch (InterruptedException e)
         {
            // do nothing?
         }
      }
   }

   private long calculateSerializedSize(Object value)
   {
      ByteArrayOutputStream baos = null;
      ObjectOutputStream oos = null;
      try
      {
         baos = new ByteArrayOutputStream();
         oos = new ObjectOutputStream(baos);
         oos.writeObject(value);
         oos.close();
         baos.close();
         return baos.size();
      }
      catch (Exception e)
      {
         log.warn("Unable to calculate serialized size of object " + value, e);
         try
         {
            if (oos != null) oos.close();
            if (baos != null) baos.close();
         }
         catch (Exception e2)
         {
            log.warn("Unable to close streams", e2);
         }
      }
      return 0;
   }

   public String getNodeIndex()
   {
      try
      {
         return " - " + Integer.parseInt(System.getProperty("currentIndex"));
      } catch (Exception e) {
         return "";
      }
   }

   public void setConfiguration(Configuration configuration)
   {
      this.configuration = configuration;
   }
}
